/*
----------------------------------------------------------------------

	Clock art

	Alain Chebrou

	adc.c		The clock art code for clocks. ADC management

	When		Who	What
	10/05/21	ac	Creation

----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aalogmes.h"
#include	"aaprintf.h"

#include	"gpiobasic.h"
#include	"timbasic.h"
#include	"dmabasic.h"

#include	"stm32g0xx_ll_adc.h"
#include	"stm32g0xx_ll_bus.h"
#include	"stm32g0xx_ll_dma.h"

//#include	"string.h"
//#include	"stdlib.h"
//#include	"aautils.h"		// For memory dump

#include	"clock.h"

extern	const uint32_t	adcIndex [MOTOR_COUNT] ;
extern	uint16_t		adcValues [MOTOR_COUNT] ;
extern	uint32_t		adcDone ;


//--------------------------------------------------------------------------------
//	ADC and ADC DMA init

void	adcDmaCallback (uint32_t event, uint32_t dir, uintptr_t arg)
{
	(void) dir ;
	(void) arg ;

	// Check DMA transfer complete
	if (event == DMA_EV_COMPLETE)
	{
		// Call interruption treatment function
tst1_0 () ;
		aaLogMes ("DMA TC\n", 0, 0, 0, 0, 0) ;
		adcDone = 1 ;
	}

	// Check DMA transfer error
	if (event == DMA_EV_ERROR)
	{
		// Call interruption treatment function
		aaLogMes ("DMA Error\n", 0, 0, 0, 0, 0) ;
	}
}

//--------------------------------------------------------------------------------

void	adcDmaInit (uint32_t bufferSize, uint16_t * pBuffer)
{
	// Configure NVIC to enable DMA interruptions
	NVIC_SetPriority (DMA1_Channel1_IRQn, 1) ; // DMA IRQ lower priority than ADC IRQ ???
	NVIC_EnableIRQ (DMA1_Channel1_IRQn) ;

	LL_AHB1_GRP1_EnableClock (LL_AHB1_GRP1_PERIPH_DMA1) ;

	LL_DMA_ConfigTransfer (DMA1,
							LL_DMA_CHANNEL_1,
							LL_DMA_DIRECTION_PERIPH_TO_MEMORY |
							LL_DMA_MODE_CIRCULAR              |
							LL_DMA_PERIPH_NOINCREMENT         |
							LL_DMA_MEMORY_INCREMENT           |
							LL_DMA_PDATAALIGN_HALFWORD        |
							LL_DMA_MDATAALIGN_HALFWORD        |
							LL_DMA_PRIORITY_HIGH) ;

	// Select ADC as DMA transfer request
	LL_DMAMUX_SetRequestID (DMAMUX1, LL_DMAMUX_CHANNEL_0, LL_DMAMUX_REQ_ADC1) ;

	// Set DMA transfer addresses of source and destination
	LL_DMA_ConfigAddresses (DMA1,
	                         LL_DMA_CHANNEL_1,
	                         LL_ADC_DMA_GetRegAddr (ADC1, LL_ADC_DMA_REG_REGULAR_DATA),
	                         (uint32_t) pBuffer,
	                         LL_DMA_DIRECTION_PERIPH_TO_MEMORY);

	// Set DMA transfer size
	LL_DMA_SetDataLength (DMA1, LL_DMA_CHANNEL_1, bufferSize) ;

	// Set interrupt handler callback
	dmaSetCallbackRaw ((dma_t *) DMA1, LL_DMA_CHANNEL_1, adcDmaCallback, 0) ;

	// Enable DMA transfer complete interruption
	LL_DMA_EnableIT_TC (DMA1, LL_DMA_CHANNEL_1) ;

	// Enable DMA transfer error interruption
	LL_DMA_EnableIT_TE (DMA1, LL_DMA_CHANNEL_1) ;

	// Enable the DMA transfer
	LL_DMA_EnableChannel (DMA1, LL_DMA_CHANNEL_1) ;
}

//--------------------------------------------------------------------------------
//#define	ADC_RATIO	LL_ADC_OVS_RATIO_8
//#define	ADC_SHIFT	LL_ADC_OVS_SHIFT_RIGHT_3

#define	ADC_RATIO			LL_ADC_OVS_RATIO_4
#define	ADC_SHIFT			LL_ADC_OVS_SHIFT_RIGHT_2
#define	ADC_SAMPLINGTIME	LL_ADC_SAMPLINGTIME_19CYCLES_5

void	adcInit (void)
{
	// Enable ADC clock (core clock)
	LL_APB2_GRP1_EnableClock (LL_APB2_GRP1_PERIPH_ADC) ;
//	rccEnableClock (ADC1, 0) ;
	rccResetPeriph (ADC1, 0) ;

	// Set ADC clock (conversion clock)
    LL_ADC_SetClock (ADC1, LL_ADC_CLOCK_SYNC_PCLK_DIV4) ;

	// Set ADC data resolution
	LL_ADC_SetResolution			(ADC1, LL_ADC_RESOLUTION_12B) ;

	// Set ADC conversion data alignment
	LL_ADC_SetResolution			(ADC1, LL_ADC_DATA_ALIGN_RIGHT) ;

	// Set ADC low power mode
	LL_ADC_SetLowPowerMode			(ADC1, LL_ADC_LP_MODE_NONE) ;

	// Set the sampling time value common to all channels
    LL_ADC_SetSamplingTimeCommonChannels (ADC1, LL_ADC_SAMPLINGTIME_COMMON_1, ADC_SAMPLINGTIME) ;


	// Set ADC trigger source
	LL_ADC_REG_SetTriggerSource		(ADC1, LL_ADC_REG_TRIG_SOFTWARE) ;

	// Set ADC trigger polarity (unused for SW trigger)
	// LL_ADC_REG_SetTriggerEdge	(ADC1, LL_ADC_REG_TRIG_EXT_RISING) ;

	// Set ADC continuous mode
	LL_ADC_REG_SetContinuousMode	(ADC1, LL_ADC_REG_CONV_SINGLE) ;

	// Set DMA request generation mode
	LL_ADC_REG_SetDMATransfer (ADC1, LL_ADC_REG_DMA_TRANSFER_UNLIMITED);

    // Set ADC overrun behavior
	LL_ADC_REG_SetOverrun			(ADC1, LL_ADC_REG_OVR_DATA_OVERWRITTEN) ;


	LL_ADC_ClearFlag_CCRDY (ADC1) ;

	// Beware CCRDY is set only if CHSELRMOD value is changed
	if (LL_ADC_REG_GetSequencerConfigurable (ADC1) != LL_ADC_REG_SEQ_FIXED)
	{
		LL_ADC_REG_SetSequencerConfigurable (ADC1, LL_ADC_REG_SEQ_FIXED) ;
		while (LL_ADC_IsActiveFlag_CCRDY (ADC1) == 0)
		{
		}
		LL_ADC_ClearFlag_CCRDY (ADC1) ;
	}

/* only for fully configurable
	LL_ADC_REG_SetSequencerLength (ADC1, LL_ADC_REG_SEQ_SCAN_DISABLE) ;
	while (LL_ADC_IsActiveFlag_CCRDY (ADC1) == 0)
	{
	}
	LL_ADC_ClearFlag_CCRDY (ADC1) ;
*/
	// Configure the sequencer and over sampling
	LL_ADC_REG_SetSequencerDiscont		(ADC1, LL_ADC_REG_SEQ_DISCONT_DISABLE) ;
	LL_ADC_SetOverSamplingScope			(ADC1, LL_ADC_OVS_GRP_REGULAR_CONTINUED) ;
	LL_ADC_ConfigOverSamplingRatioShift	(ADC1, ADC_RATIO, ADC_SHIFT) ;
	LL_ADC_SetOverSamplingDiscont		(ADC1, LL_ADC_OVS_REG_CONT) ;
	LL_ADC_SetTriggerFrequencyMode		(ADC1, LL_ADC_CLOCK_FREQ_MODE_LOW) ;

	// Beware CCRDY is set only if SCANDIR value is changed
	if (LL_ADC_REG_GetSequencerScanDirection (ADC1) != LL_ADC_REG_SEQ_SCAN_DIR_FORWARD)
	{
		LL_ADC_REG_SetSequencerScanDirection (ADC1, LL_ADC_REG_SEQ_SCAN_DIR_FORWARD) ;
		while (LL_ADC_IsActiveFlag_CCRDY (ADC1) == 0)
		{
		}
		LL_ADC_ClearFlag_CCRDY (ADC1) ;
	}

	// Beware CCRDY is set if the CHSELR register is written
	LL_ADC_REG_SetSequencerChAdd (ADC1, LL_ADC_CHANNEL_0 | LL_ADC_CHANNEL_4  |
										LL_ADC_CHANNEL_5 | LL_ADC_CHANNEL_6  |
										LL_ADC_CHANNEL_7 | LL_ADC_CHANNEL_8  |
										LL_ADC_CHANNEL_9 | LL_ADC_CHANNEL_10) ;
	while (LL_ADC_IsActiveFlag_CCRDY (ADC1) == 0)
	{
	}
	LL_ADC_ClearFlag_CCRDY (ADC1) ;


	// Set ADC channels sampling time
	LL_ADC_SetChannelSamplingTime (ADC1, LL_ADC_CHANNEL_0, LL_ADC_SAMPLINGTIME_COMMON_1);
	LL_ADC_SetChannelSamplingTime (ADC1, LL_ADC_CHANNEL_4, LL_ADC_SAMPLINGTIME_COMMON_1);
	LL_ADC_SetChannelSamplingTime (ADC1, LL_ADC_CHANNEL_5, LL_ADC_SAMPLINGTIME_COMMON_1);
	LL_ADC_SetChannelSamplingTime (ADC1, LL_ADC_CHANNEL_6, LL_ADC_SAMPLINGTIME_COMMON_1);
	LL_ADC_SetChannelSamplingTime (ADC1, LL_ADC_CHANNEL_7, LL_ADC_SAMPLINGTIME_COMMON_1);
	LL_ADC_SetChannelSamplingTime (ADC1, LL_ADC_CHANNEL_8, LL_ADC_SAMPLINGTIME_COMMON_1);
	LL_ADC_SetChannelSamplingTime (ADC1, LL_ADC_CHANNEL_9, LL_ADC_SAMPLINGTIME_COMMON_1);
	LL_ADC_SetChannelSamplingTime (ADC1, LL_ADC_CHANNEL_10, LL_ADC_SAMPLINGTIME_COMMON_1);

	LL_ADC_DisableIT_EOC (ADC1) ;
	LL_ADC_DisableIT_EOS (ADC1) ;
}

//--------------------------------------------------------------------------------
// dmaMode is one of:
//	LL_ADC_REG_DMA_TRANSFER_NONE
//	LL_ADC_REG_DMA_TRANSFER_LIMITED		= one shot
//	LL_ADC_REG_DMA_TRANSFER_UNLIMITED	= circular

void	adcStart (void)
{
	uint32_t dmaMode ;

	LL_ADC_EnableInternalRegulator (ADC1) ;
	bspDelayUs (LL_ADC_DELAY_INTERNAL_REGUL_STAB_US + 1u) ;

	// Disable DMA for calibration
	dmaMode = LL_ADC_REG_GetDMATransfer (ADC1); ;
	LL_ADC_REG_SetDMATransfer (ADC1, LL_ADC_REG_DMA_TRANSFER_NONE) ;
	LL_ADC_StartCalibration (ADC1) ;
	while (LL_ADC_IsCalibrationOnGoing (ADC1) == 1u)
	{
	}
	LL_ADC_REG_SetDMATransfer (ADC1, dmaMode) ;		// Restore DMA mode

	// Delay after calibration and ADC enable (2 fADC clock)
	bspDelayUs (LL_ADC_DELAY_CALIB_ENABLE_ADC_CYCLES) ;		// Enough is fADC >= 1 MHz

	// Enable ADC
	LL_ADC_Enable (ADC1) ;
	while (LL_ADC_IsActiveFlag_ADRDY (ADC1) == 0)
	{
	}
//	LL_ADC_ClearFlag_ADRDY (ADC1) ;
}

//--------------------------------------------------------------------------------

void	adcTrig (void)
{
	if ((LL_ADC_IsEnabled(ADC1) == 1)               &&
		(LL_ADC_IsDisableOngoing(ADC1) == 0)        &&
		(LL_ADC_REG_IsConversionOngoing(ADC1) == 0))
	{
tst1_1 () ;
		LL_ADC_REG_StartConversion (ADC1) ;
	}
	else
	{
		AA_ASSERT (0) ;
	}
}

//--------------------------------------------------------------------------------
