/*
----------------------------------------------------------------------

	Clock art

	Alain Chebrou

	clock.c		The clock art code for clocks MCU

	When		Who	What
	01/02/21	ac	Creation
	23/05/21	ac	V1.1	1st release for production boards
	25/05/21	ac	V1.2	Remove assert in EXTI_CALLBACK(SPI_NSS_EXTI_LINE)
	27/05/21	ac	V1.3	Invert CMD_ACK open drain output

----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"	// For critical section
#include	"aalogmes.h"
#include	"aaprintf.h"

#include	"gpiobasic.h"
#include	"timbasic.h"
#include	"dmabasic.h"
#include	"spibasic.h"
#include	"extibasic.h"

#include	"stm32g0xx_ll_adc.h"
#include	"stm32g0xx_ll_bus.h"
#include	"stm32g0xx_ll_dma.h"

#include	"string.h"
#include	"stdlib.h"
#include	"aautils.h"		// For memory dump

#include	"clock.h"

#define		CLOCKART_VERSION	"1.4"

uint32_t	isMotorRunning	(void) ;


//--------------------------------------------------------------------------------
//	Pins descriptors for gpioConfigurePin()
//	LEDs and Button are defined in bspboard.h

const gpioPinDesc_t	clockPinDesc [] =
{
	{ PORT_CMDACK,	PIN_CMDACK,		0,	AA_GPIO_MODE_OUTPUT_OD },

	{ PORT_TST0,	PIN_TST0,		0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_LOW },
	{ PORT_TST1,	PIN_TST1,		0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_LOW },

	{ PORT_MRST,	PIN_MRST,		0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_MEDIUM },
	{ PORT_MDIR,	PIN_MDIR,		0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_MEDIUM },
	{ PORT_M0,		PIN_M0,			0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_MEDIUM },
	{ PORT_M1,		PIN_M1,			0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_MEDIUM },
	{ PORT_M2,		PIN_M2,			0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_MEDIUM },
	{ PORT_M3,		PIN_M3,			0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_MEDIUM },
	{ PORT_M4,		PIN_M4,			0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_MEDIUM },
	{ PORT_M5,		PIN_M5,			0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_MEDIUM },
	{ PORT_M6,		PIN_M6,			0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_MEDIUM },
	{ PORT_M7,		PIN_M7,			0,	AA_GPIO_MODE_OUTPUT_PP | AA_GPIO_SPEED_MEDIUM },

	{ PORT_ADDR0,	PIN_ADDR0,		0,	AA_GPIO_MODE_INPUT_UP },	// Board address
	{ PORT_ADDR1,	PIN_ADDR1,		0,	AA_GPIO_MODE_INPUT_UP },
	{ PORT_ADDR2,	PIN_ADDR2,		0,	AA_GPIO_MODE_INPUT_UP },
	{ PORT_ADDR3,	PIN_ADDR3,		0,	AA_GPIO_MODE_INPUT_UP },

	{ PORT_HL0,		PIN_HL0,		0,	AA_GPIO_MODE_ADC },			// Hall sensor
	{ PORT_HS0,		PIN_HS0,		0,	AA_GPIO_MODE_ADC },
	{ PORT_HL1,		PIN_HL1,		0,	AA_GPIO_MODE_ADC },
	{ PORT_HS1,		PIN_HS1,		0,	AA_GPIO_MODE_ADC },
	{ PORT_HL2,		PIN_HL2,		0,	AA_GPIO_MODE_ADC },
	{ PORT_HS2,		PIN_HS2,		0,	AA_GPIO_MODE_ADC },
	{ PORT_HL3,		PIN_HL3,		0,	AA_GPIO_MODE_ADC },
	{ PORT_HS3,		PIN_HS3,		0,	AA_GPIO_MODE_ADC },

	// SPI pins are configured by SPI library
	// UART pins are defined in uartbasic.c

} ;

//--------------------------------------------------------------------------------
// Generate motor pulses on these pins

typedef struct
{
	GPIO_TypeDef	* pPort;
	uint32_t		pinMask ;

} motorGpio_t ;

const motorGpio_t	motorGpio [MOTOR_COUNT] =
{
	{ BSP_IO_GPIO(PORT_M0 - 'A'), 1 << PIN_M0 },
	{ BSP_IO_GPIO(PORT_M1 - 'A'), 1 << PIN_M1 },
	{ BSP_IO_GPIO(PORT_M2 - 'A'), 1 << PIN_M2 },
	{ BSP_IO_GPIO(PORT_M3 - 'A'), 1 << PIN_M3 },
	{ BSP_IO_GPIO(PORT_M4 - 'A'), 1 << PIN_M4 },
	{ BSP_IO_GPIO(PORT_M5 - 'A'), 1 << PIN_M5 },
	{ BSP_IO_GPIO(PORT_M6 - 'A'), 1 << PIN_M6 },
	{ BSP_IO_GPIO(PORT_M7 - 'A'), 1 << PIN_M7 },
} ;

// The index of each motor in adcValues array
const uint32_t	adcIndex [MOTOR_COUNT] =
{
	#if (PROTO_BOARD == 0)
		0, 1, 6, 7, 4, 5, 2, 3
	#else
		0, 1, 2, 3, 4, 5, 6, 7
	#endif
} ;

//--------------------------------------------------------------------------------
//	Motor environment

typedef struct
{
	uint32_t	enable ;		// 0 or 1
	uint32_t	command ;		// MCMD_xxx
	uint32_t	direction ;		// DIR_CW or DIR_CCW
	uint32_t	count ;			// Count of pulses to generate
	uint32_t	stepSize ;		// Time between pulses
	uint32_t	stepCount ;		// Current count in stepSize
	uint32_t	pos ;			// Motor position
	uint32_t	temp ;			// Temporary value

} motorDesc_t ;

volatile motorDesc_t	motorDesc [MOTOR_COUNT] ;

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------

uint32_t		mcuAddress ;	// From hardware straps

uint32_t		stepSize	= 2u ;
uint32_t		pulseCount	= PULSE_PER_TURN * 2  ;		// PULSE_PER_TURN / 4u ;
uint32_t		direction	= 0u ;
uint32_t		motorAck ;

const timDesc_t	* pTimDesc ;
aaTaskId_t		motorTaskId ;

// Mode for motor task
#define			MODE_RUN		0u		// Mode normal
#define			MODE_RAZ		1u		// Mode to set motor position to origin
uint32_t		motorTaskMode;

// ADC buffer
uint16_t		adcValues [MOTOR_COUNT] ;
uint32_t		adcDone ;

//--------------------------------------------------------------------------------
// UART command buffer management

#define		CMDBUFFER_SIZE		128u

typedef struct
{
	char		cmdBuffer [CMDBUFFER_SIZE] ;
	char		* pCmdBuffer ;
	uint32_t	count ;

} cmdBuffer_t ;

cmdBuffer_t	cmdBuffer ;

void	cmdBufferInit (cmdBuffer_t * pBuff)
{
	pBuff->pCmdBuffer = pBuff->cmdBuffer;
	pBuff->count = 0 ;
}

//	Receive a command

uint32_t	cmdBufferGet (cmdBuffer_t * pBuff)
{
	char cc ;

	while (aaCheckGetChar () != 0)
	{
		cc = aaGetChar () ;
		aaPutChar (cc) ;
		if (cc == '\r')
		{
			* pBuff->pCmdBuffer = 0 ;	// End Of String
			aaPutChar ('\n') ;
			return 1 ;					// Command received
		}
		if (pBuff->pCmdBuffer < & pBuff->cmdBuffer[CMDBUFFER_SIZE])
		{
			* pBuff->pCmdBuffer++ = cc ;
			pBuff->count++ ;
		}
	}
	return 0 ;		// Command not received
}

//--------------------------------------------------------------------------------
//	SPI message buffer

#define		SPI_DESC		10u			// Master SPI descriptor ID in spiXXxx.h

const spiDesc_t 	* pSpiDesc  ;
spiHandle_t 		spiHandle ;
uint32_t			spiMessageEnded;

#define	MSG_BUFFSIZE	1024
uint16_t	msgBuffer [MSG_BUFFSIZE] ;
uint32_t	msgIndex ;			// Current position in msgBuffer ;

//--------------------------------------------------------------------------------
//	Command arrays for motorTask

#define		MCMD_FREE		0u		// This buffer is free if cmd field is MCMD_FREE
#define		MCMD_MOVE		1u		// Doing a move
#define		MCMD_DELAY		2u		// Waiting for a count of ms
#define		MCMD_ACK		3u		// Set acknowledge output (anded for all motors)
#define		MCMD_WAIT		4u		// Wait until all motors are stopped

#define		MCMD_BUFFCOUNT	8u						// Must be a power of 2
#define		MCMD_BUFMASK	(MCMD_BUFFCOUNT-1u)

typedef struct
{
	uint32_t	command ;
	uint32_t	direction ;
	uint32_t	count ;
	uint32_t	stepSize ;

} mCmd_t ;

typedef struct
{
	uint32_t	iRead ;		// For motorTask
	uint32_t	iWrite ;	// For message manager
	uint32_t	position ;	// Position of the hand when all commands in the buffer will be executed
							// (pulses 0-4319)

	mCmd_t		mCmd [MCMD_BUFFCOUNT] ;

} mBuffer_t ;

mBuffer_t	mBuffer [MOTOR_COUNT] ;

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------

static void	motorAckSet (uint32_t iMotor)
{
	aaIrqStatus_t	intState ;

	intState = bspSaveAndDisableIrq () ;
	motorAck |= 1u << iMotor ;
	if (motorAck == ((1u << MOTOR_COUNT) - 1u))
	{
		// Set physical output if SPI is ready to receive: not in packet processing
		if (isSpiReady () == 1u)
		{
			cmdAckOn() ;
		}
	}
	bspRestoreIrq (intState) ;
}

void	spiAckSet (void)
{
	aaIrqStatus_t	intState ;

	intState = bspSaveAndDisableIrq () ;
	if (motorAck == ((1u << MOTOR_COUNT) - 1u))
	{
		// Set physical output if all the motors have acknowledged
		cmdAckOn() ;
	}
	bspRestoreIrq (intState) ;
}

void	motorAckReset (void)
{
	aaIrqStatus_t	intState ;

	intState = bspSaveAndDisableIrq () ;
	motorAck = 0u ;
	cmdAckOff () ;
	bspRestoreIrq (intState) ;
}

//--------------------------------------------------------------------------------
// Define the motor task stack, to avoid use of dynamic memory
#define	MOTOR_STACK_SIZE	256u
static	bspStackType_t		motorStack [MOTOR_STACK_SIZE] ;

static	void	motorTask (uintptr_t arg)
{
	uint32_t				ii ;
	uint32_t				bWaitEnd ;
	volatile motorDesc_t	* pDesc ;

	(void) arg ;

	while (1)
	{
		// Wait for timer UI
		aaSignalWait (1, NULL, AA_SIGNAL_OR, AA_INFINITE) ;

		// Check for MODE_RAZ: set motor position to origin
		if (motorTaskMode == MODE_RAZ)
		{
			uint32_t	hallSensor ;
			pDesc = motorDesc ;
			for (ii = 0 ; ii < MOTOR_COUNT ; ii++, pDesc++)
			{
				if (pDesc->enable != 0)
				{
					pDesc->stepCount++ ;
					if (pDesc->stepCount == pDesc->stepSize)
					{
						// Set direction, then a setup time to pulse > 100ns
						if (pDesc->direction == 0)
						{
							motorDirCCW () ;
						}
						else
						{
							motorDirCW () ;
						}

						// Check Hall sensor: record the lower value
						hallSensor = adcValues [adcIndex[ii]] ;
						if (pDesc->temp > hallSensor)
						{
							pDesc->temp = hallSensor ;
							pDesc->pos  = 0 ;
						}
						else
						{
							pDesc->pos++ ;
						}

						// Generate a pulse
						motorGpio [ii].pPort->BSRR = motorGpio [ii].pinMask ;
						bspDelayUs (10);
						motorGpio [ii].pPort->BRR = motorGpio [ii].pinMask ;

						pDesc->stepCount = 0 ;
						if (--pDesc->count == 0)
						{
							pDesc->enable = 0 ;
						}
					}
				}
			}
			adcTrig () ;	// Start new Hall sensor measurement
			continue ;
		}


		// MODE_RUN
		bWaitEnd = 1 ^ isMotorRunning () ;	// 1 if all motors stopped

		pDesc = motorDesc ;
		for (ii = 0 ; ii < MOTOR_COUNT ; ii++, pDesc++)
		{
			while (1)
			{
				if (pDesc->enable == 0)
				{
					// Not executing a command: check if a new command is available
					mBuffer_t	* pBuffer = & mBuffer [ii] ;
					mCmd_t		* pCmd = & pBuffer->mCmd [pBuffer->iRead] ;

					if (pCmd->command == MCMD_MOVE)
					{
						// Copy the move command parameters to the motor descriptor
						pDesc->command		= pCmd->command ;
						pDesc->direction	= pCmd->direction ;
						pDesc->count		= pCmd->count ;
						pDesc->stepSize		= pCmd->stepSize ;
						pDesc->stepCount	= 0 ;
						pDesc->enable		= 1 ;

						pCmd->command		= MCMD_FREE ;
						pBuffer->iRead		= (pBuffer->iRead + 1) & MCMD_BUFMASK ;
					}
					else if (pCmd->command == MCMD_DELAY)
					{
						// Copy the delay command parameters to the motor descriptor
						pDesc->count		= pCmd->count ;
						if (pCmd->count == 0)
						{
							pDesc->count	= 1 ;		// To avoid 0 delay
						}
						pDesc->command		= pCmd->command ;
						pDesc->stepSize		= SPEED_HIGHEST ;
						pDesc->stepCount	= 0 ;
						pDesc->enable		= 1 ;

						pCmd->command		= MCMD_FREE ;
						pBuffer->iRead		= (pBuffer->iRead + 1) & MCMD_BUFMASK ;
					}
					else if (pCmd->command == MCMD_ACK)
					{
						// Set acknowledge for this motor
						motorAckSet (ii) ;
						pCmd->command		= MCMD_FREE ;
						pBuffer->iRead		= (pBuffer->iRead + 1) & MCMD_BUFMASK ;
						continue ;		// Check if there is another command
					}
					else if (pCmd->command == MCMD_WAIT)
					{
						// Wait until all motors are stopped
						pDesc->command		= pCmd->command ;
						pDesc->enable		= 1 ;
						pCmd->command		= MCMD_FREE ;
						pBuffer->iRead		= (pBuffer->iRead + 1) & MCMD_BUFMASK ;
					}
					else if (pCmd->command == MCMD_FREE)
					{
						// Nothing to do
					}
					else
					{
						AA_ASSERT (0) ;
					}
				}
				break ;
			}

			if (pDesc->enable != 0)
			{
				if (pDesc->command == MCMD_MOVE)
				{
					pDesc->stepCount++ ;
					if (pDesc->stepCount == pDesc->stepSize)
					{
						// Set direction, then a setup time to pulse > 100ns
						if (pDesc->direction == DIR_CCW)
						{
							motorDirCCW () ;
							pDesc->pos-- ;
							if (pDesc->pos == 0)
							{
								pDesc->pos = PULSE_PER_TURN - 1 ;
							}
						}
						else
						{
							motorDirCW () ;
							pDesc->pos++ ;
							if (pDesc->pos == PULSE_PER_TURN)
							{
								pDesc->pos = 0 ;
							}
						}

						// Generate a pulse
						motorGpio [ii].pPort->BSRR = motorGpio [ii].pinMask ;
						bspDelayUs (10);
						motorGpio [ii].pPort->BRR = motorGpio [ii].pinMask ;

						pDesc->stepCount = 0 ;
						if (--pDesc->count == 0)
						{
							pDesc->enable = 0 ;
							led1_Off	() ;
						}
					}
				}
				else if (pDesc->command == MCMD_DELAY)
				{
					pDesc->stepCount++ ;
					if (pDesc->stepCount == pDesc->stepSize)
					{
						pDesc->stepCount = 0 ;
						if (--pDesc->count == 0)
						{
							pDesc->enable = 0 ;
						}
					}
				}
				else if (pDesc->command == MCMD_WAIT)
				{
					// Wait until all motors are stopped
					if (bWaitEnd != 0)
					{
						// End of command
						pDesc->enable = 0 ;
					}
				}
				else
				{
					// Unknown command, Nothing to do
					AA_ASSERT (0) ;
				}
			}
		}
	}
}

//--------------------------------------------------------------------------------

static	void tim7Callback (const timDesc_t * pTimDesc_, uintptr_t arg, uint32_t cause)
{
	(void) pTimDesc_ ;
	(void) arg ;
	(void) cause ;

	// Awake the motor task
	aaSignalSend (motorTaskId, 1) ;

	// UIF flag is cleared by interrupt handler
}

//--------------------------------------------------------------------------------
//	Initialize resources

void	clockInit (void)
{
	uint32_t	ii ;
	uint32_t	jj ;
	mBuffer_t	* pBuffer ;

	// Configure GPIO
	for (ii = 0u ; ii < sizeof (clockPinDesc) / sizeof (gpioPinDesc_t) ; ii++)
	{
		gpioConfigurePin (& clockPinDesc [ii]) ;
	}

	// Output GPIO default value
	led0_Off	() ;
	led1_Off	() ;
	tst0_0		() ;
	tst1_0		() ;
	cmdAckOff	() ;

	BSP_IO_GPIO(PORT_M0-'A')->BRR = (1 << PIN_M0) ;
	BSP_IO_GPIO(PORT_M1-'A')->BRR = (1 << PIN_M1) ;
	BSP_IO_GPIO(PORT_M2-'A')->BRR = (1 << PIN_M2) ;
	BSP_IO_GPIO(PORT_M3-'A')->BRR = (1 << PIN_M3) ;
	BSP_IO_GPIO(PORT_M4-'A')->BRR = (1 << PIN_M4) ;
	BSP_IO_GPIO(PORT_M5-'A')->BRR = (1 << PIN_M5) ;
	BSP_IO_GPIO(PORT_M6-'A')->BRR = (1 << PIN_M6) ;
	BSP_IO_GPIO(PORT_M7-'A')->BRR = (1 << PIN_M7) ;
	motorDirCCW	  () ;
	motorResetOff () ;

	// Read MCU address
	mcuAddress  =  (BSP_IO_GPIO (PORT_ADDR0-'A')->IDR >> PIN_ADDR0) & 1u ;
	mcuAddress |= ((BSP_IO_GPIO (PORT_ADDR1-'A')->IDR >> PIN_ADDR1) & 1u) << 1 ;
	mcuAddress |= ((BSP_IO_GPIO (PORT_ADDR2-'A')->IDR >> PIN_ADDR2) & 1u) << 2 ;
	mcuAddress |= ((BSP_IO_GPIO (PORT_ADDR3-'A')->IDR >> PIN_ADDR3) & 1u) << 3 ;

	aaPrintf ("Clock Art V%s  @%d   mode:%s\n", CLOCKART_VERSION, mcuAddress, (PROTO_BOARD == 1) ? "Proto" : "final") ;

	//	Initialize slave SPI
	spiInit () ;

	// initialize motor descriptors
	for (ii = 0 ; ii < MOTOR_COUNT ; ii++)
	{
		motorDesc [ii].enable = 0 ;
	}

	// Initialize mBuffer (motorTask commands buffer)
	pBuffer = mBuffer ;
	for (ii = 0 ; ii < MOTOR_COUNT ; ii++, pBuffer++)
	{
		pBuffer->iRead = 0 ;
		pBuffer->iWrite = 0 ;
		for (jj = 0 ; jj < MCMD_BUFFCOUNT ; jj++)
		{
			pBuffer->mCmd [jj].command = MCMD_FREE ;
		}
	}

	// Create the clock task
	aaTaskCreate (
			AA_PRIO_MAX,		// Task priority
			"tMotors",			// Task name
			motorTask,			// Entry point,
			0,					// Entry point parameter
			motorStack,			// Stack pointer
			MOTOR_STACK_SIZE,	// Stack size
			0,					// Flags
			& motorTaskId) ;			// Created task id

	// Initialize the timer. Time base with UPDATE interrupt (UIF)
	if (timGetDesc (MOTOR_TIMER_ID, & pTimDesc) == TIM_EFAIL)
	{
		aaPrintf ("Unknown timer descriptor id: %u\n", MOTOR_TIMER_ID) ;
	}
	else
	{
		timTimeBaseInit (pTimDesc, 0u, MOTOR_TIMER_FREQ, TIM_INIT_ARR_PRELOADEN) ;
		timConfigureIntr (pTimDesc, MOTOR_TIMER_PRIO, tim7Callback, 0, TIM_INTR_UPDATE) ;
		timStart (pTimDesc) ;
	}
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	Add a delay command to the motor buffer

void	addDelay (uint32_t iMotor, uint32_t msCount)
{
	mBuffer_t	* pBuffer = & mBuffer [iMotor] ;
	mCmd_t		* pCmd = & pBuffer->mCmd [pBuffer->iWrite] ;

	if (msCount != 0)
	{
		if (pCmd->command != MCMD_FREE)
		{
			// Command buffer full
			AA_ASSERT (0) ;
			return ;
		}

		pCmd->count		= msCount ;
		pCmd->command	= MCMD_DELAY ;		// To put last
		pBuffer->iWrite = (pBuffer->iWrite + 1) & MCMD_BUFMASK ;
	}
}

//--------------------------------------------------------------------------------
//	Add a Ack command to the motor buffer

void	addMotorAck (uint32_t iMotor)
{
	mBuffer_t	* pBuffer = & mBuffer [iMotor] ;
	mCmd_t		* pCmd = & pBuffer->mCmd [pBuffer->iWrite] ;

	if (pCmd->command != MCMD_FREE)
	{
		// Command buffer full
		AA_ASSERT (0) ;
		return ;
	}

	pCmd->command	= MCMD_ACK ;		// To put last
	pBuffer->iWrite = (pBuffer->iWrite + 1) & MCMD_BUFMASK ;
}

//--------------------------------------------------------------------------------
//	Add a relative move command to the motor buffer:
//	Move 'pulseCount' pulses in the 'dir' direction (DIR_CW or DIR_CCW only)
//	The duration of the move is 'msCount' milli seconds

void	addMove (uint32_t iMotor, uint32_t dir, uint32_t pulseCount_, uint32_t msCount)
{
	uint32_t	stepSize_ ;
	uint32_t	tickCount ;
	uint32_t	N ;
	mBuffer_t	* pBuffer = & mBuffer [iMotor] ;
	mCmd_t		* pCmd = & pBuffer->mCmd [pBuffer->iWrite] ;

	if (pCmd->command != MCMD_FREE  ||  dir == DIR_SHORT  ||  dir == DIR_LONG)
	{
		// Command buffer full, or invalid parameter
		AA_ASSERT (0) ;
		return ;
	}

	if (pulseCount_ == 0)
	{
		return ;	// No move required
	}

	tickCount	= msCount * (MOTOR_TIMER_FREQ / 1000) ;
	stepSize_	= tickCount / pulseCount_ ;

	if (stepSize_ < (MOTOR_TIMER_FREQ / 1000))
	{
		// msCount too short, use the fastest speed
		stepSize_ = SPEED_HIGHEST ;
		tickCount	= pulseCount_ * (MOTOR_TIMER_FREQ / 1000) ;
	}

	// Insert a command for the move at high speed
	N				= (pulseCount_ * stepSize_) + pulseCount_ - tickCount ;
	pCmd->count		= N ;
	pCmd->stepSize	= stepSize_ ;
	pCmd->direction	= dir ;

	pCmd->command	= MCMD_MOVE ;		// To put last
	pBuffer->iWrite = (pBuffer->iWrite + 1) & MCMD_BUFMASK ;

	if (N != pulseCount_)
	{
		// Insert a command for the move at slow speed
		pCmd = & pBuffer->mCmd [pBuffer->iWrite] ;
		if (pCmd->command != MCMD_FREE)
		{
			// Command buffer full
			AA_ASSERT (0) ;
			return ;
		}

		pCmd->count		= pulseCount_ - N ;
		pCmd->stepSize	= stepSize_ + 1 ;
		pCmd->direction	= dir ;

		pCmd->command	= MCMD_MOVE ;		// To put last
		pBuffer->iWrite = (pBuffer->iWrite + 1) & MCMD_BUFMASK ;
	}

	// Add the move to know the future position
	if (dir == DIR_CW)
	{
		pBuffer->position += pulseCount_ ;
	}
	else
	{
		pBuffer->position += (PULSE_PER_TURN - (pulseCount_ % PULSE_PER_TURN)) ;
	}
	pBuffer->position %= PULSE_PER_TURN ;

}

//--------------------------------------------------------------------------------
//	Move to the absolute position 'position' in degree (0-359)
//	The 'dir' direction can be DIR_CW, DIR_CCW, DIR_SHORT or DIR_LONG: choose the fastest or slowest direction to reach the desired position.
//	The duration of the move is 'msCount' milli seconds

void	addMoveAbs (uint32_t iMotor, uint32_t dir, uint32_t position, uint32_t msCount)
{
	int32_t		cwAngle ;
	int32_t		ccwAngle ;
	uint32_t	direction_ ;
	int32_t		handPos = (int32_t) (mBuffer [iMotor].position / PULSE_PER_DEG) ;
// TODO : Compute pulse position, not degree position

	AA_ASSERT (position < 360u) ;

	cwAngle = (int32_t) position - handPos ;
	if (cwAngle == 0)
	{
		return ;	// Already at the position
	}

	if (cwAngle < 0)
	{
		cwAngle += 360 ;
	}

	ccwAngle = handPos - (int32_t) position ;
	if (ccwAngle < 0)
	{
		ccwAngle += 360 ;
	}

	direction_ = DIR_CW ;
	if (dir == DIR_SHORT)
	{
		if (ccwAngle < cwAngle)
		{
			direction_ = DIR_CCW ;
			cwAngle = ccwAngle ;
		}
	}
	else if (dir == DIR_LONG)
	{
		if (ccwAngle > cwAngle)
		{
			direction_ = DIR_CCW ;
			cwAngle = ccwAngle ;
		}
	}
	else if (dir == DIR_CCW)
	{
		direction_ = DIR_CCW ;
		cwAngle = ccwAngle ;
	}
	else
	{
		// Nothing to do
	}

	if (cwAngle != 0)
	{
		addMoveAngle (iMotor, direction_, cwAngle, msCount) ;
// TODO : use addMove()
	}
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	Returns 1 if at least one of the motors is moving: enabled and not waiting

uint32_t	isMotorRunning ()
{
	uint32_t	ii ;

	for (ii = 0 ; ii < MOTOR_COUNT ; ii++)
	{
		if ((motorDesc [ii].enable == 1)  &&  (motorDesc [ii].command != MCMD_WAIT))
		{
			return 1;
		}
	}
	return 0 ;
}

//--------------------------------------------------------------------------------
//	Set all hands to origin position
//	This is a board level command (not a motor command)

void	goToOrigin (void)
{
	uint32_t				ii ;
	volatile motorDesc_t	* pDesc ;

	// Check if all motors are disabled (stopped)
	while (isMotorRunning () == 1)
	{
		aaTaskDelay (1) ;
	}

	// Get current sensor value
	adcDone = 0 ;
	adcTrig () ;
	while (adcDone == 0)
	{
		aaTaskDelay (1) ;
	}

	// Start all motors
	motorTaskMode = MODE_RAZ ;
	pDesc = motorDesc ;
	for (ii = 0 ; ii < MOTOR_COUNT ; ii++, pDesc++)
	{
		pDesc->direction	= DIR_CW ;
		pDesc->count		= PULSE_PER_TURN ;
		pDesc->stepSize		= SPEED_HIGHEST ;		// Max speed
		pDesc->stepCount	= 0 ;
		pDesc->pos			= 0 ;
		pDesc->temp			= 4096 ;	// Higher than the highest ADC value

		pDesc->enable		= 1 ;		// Start
	}

	// Wait for motor end
	while (1)
	{
		aaTaskDelay (2) ;
		if (isMotorRunning () == 0)
		{
			break ;
		}
	}
	// Out of RAZ mode
	motorTaskMode = MODE_RUN ;

	// Go to origin
	{
		uint32_t				temp ;

		pDesc = motorDesc ;
		for (ii = 0 ; ii < MOTOR_COUNT ; ii++, pDesc++)
		{
//aaPrintf ("m %u min %u  pos %u\n", ii, pDesc->temp, pDesc->pos) ;

#if (PROTO_BOARD == 0)
			temp = PULSE_PER_TURN - pDesc->pos ;
			addMove (ii, DIR_CW, temp, temp) ;
#else
			// On the proto board the zero in not in front of Hall sensors
			if (pDesc->pos < ((PULSE_PER_TURN * 3u) / 4u))
			{
				temp = ((PULSE_PER_TURN * 3u) / 4u) - pDesc->pos ;	// 3240
				addMove (ii, DIR_CW, temp, temp) ;
//				pDesc->count = ((PULSE_PER_TURN * 3u) / 4u) - pDesc->pos ;	// 3240
			}
			else
			{
				temp = ((PULSE_PER_TURN * 7u) / 4u) - pDesc->pos ;	// 7560
				addMove (ii, DIR_CW, temp, temp) ;
//				pDesc->count = ((PULSE_PER_TURN * 7u) / 4u) - pDesc->pos ;	// 7560
			}
#endif
		}

		// Wait for motor end
		while (1)
		{
			aaTaskDelay (2) ;
			if (isMotorRunning () == 0)
			{
				break ;
			}
		}
	}

	// Initialize position
	pDesc = motorDesc ;
	for (ii = 0 ; ii < MOTOR_COUNT ; ii++, pDesc++)
	{
/*		if ((ii & 1) == 0)
		{
			pDesc->pos = 0 ;
		}
		else
		{
			// Because short hand go CCW on proto board ?
			pDesc->pos = PULSE_PER_TURN / 2u ;
		}
*/
		pDesc->pos = 0 ;
		mBuffer [ii].position = pDesc->pos ;
	}
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	Application entry point

void	clock ()
{
	char		* pCmd ;
	char		* savePtr ;
	char		* pArg1 ;
	char		* pArg2 ;
	char		* pArg3 ;
 	uint32_t	arg1 ;
 	uint32_t	arg2 ;
 	uint32_t	arg3 ;

 	(void) arg1;
 	(void) arg2 ;

	clockInit () ;

	adcDmaInit (MOTOR_COUNT, adcValues) ;
	adcInit () ;
	adcStart () ;

	// Wait for commands
	cmdBufferInit (& cmdBuffer) ;
	spiReceive () ;

	while (1)
	{
 		if (bspInput (BSP_BUTTON0) == 0)
		{
			// For test: move all motors

			for (arg1 = 0 ; arg1 < MOTOR_COUNT ; arg1++)
			{
				addDelay (arg1, arg1 * 250) ;
				addMoveAngle  (arg1, direction, 360, 360*12) ;
			}
			direction ^= 1 ;	// Invert direction

			while (bspInput (BSP_BUTTON0) == 0)
			{
			}
		}

		if (cmdBufferGet (& cmdBuffer) == 0)
		{
			spiCheck () ;
			aaTaskDelay (2) ;
			continue ;
		}

		// Command received
		pCmd = strtok_r (cmdBuffer.cmdBuffer, " \t", & savePtr) ;
		arg1 = 0 ;
		arg2 = 0 ;
		arg3 = 0 ;
		pArg2 = NULL ;
		pArg1 = strtok_r (NULL, " \t", & savePtr) ;
		if (pArg1 != NULL)
		{
			arg1 = strtoul (pArg1, (char **) NULL, 10);

			pArg2 = strtok_r (NULL, " \t", & savePtr) ;
			if (pArg2 != NULL)
			{
				arg2 = strtoul (pArg2, (char **) NULL, 10);

				pArg3 = strtok_r (NULL, " \t", & savePtr) ;
				if (pArg3 != NULL)
				{
					arg3 = strtoul (pArg3, (char **) NULL, 10);
				}
			}
		}

		if (0 == strcmp ("?", pCmd))
		{
			aaPrintf ("Clock Art V%s  @%d   mode:%s\n", CLOCKART_VERSION, mcuAddress, (PROTO_BOARD == 1) ? "Proto" : "final") ;
			aaPrintf ("h           Hall sensors\n") ;
			aaPrintf ("s           stepsize     %u\n", stepSize) ;
			aaPrintf ("c           pulseCount   %u\n", pulseCount) ;
			aaPrintf ("d           direction    %u\n", direction) ;
			aaPrintf ("+|-         dir CW|CCW   %u %s\n", direction, direction == DIR_CW ? "CW" : "CCW") ;
			aaPrintf ("0|1|..|7    (motor)\n") ;
			aaPrintf ("o           (origin)\n") ;
			aaPrintf ("z           (reset motor driver)\n") ;
			aaPrintf ("dp          Dump hands position\n") ;
			aaPrintf ("mp p ms     (Move motor 0) pulse ms)\n") ;
			aaPrintf ("md d ms     (Move motor 0) degree ms)\n") ;
			aaPrintf ("ma d p ms   (Move motor 0) dir pos ms)\n") ;
		}

		else if (pCmd [0] >= '0'  &&  pCmd [0] < '8')
		{
			mBuffer_t	* pBuffer = & mBuffer [pCmd [0] - '0'] ;
			mCmd_t		* pCmd_ = & pBuffer->mCmd [pBuffer->iWrite] ;

			if (pCmd_->command == MCMD_FREE)
			{
				pCmd_->direction	= direction ;
				pCmd_->count		= pulseCount ;
				pCmd_->stepSize	= stepSize ;

				pCmd_->command = MCMD_MOVE ;		// To put last
				pBuffer->iWrite = (pBuffer->iWrite + 1) & MCMD_BUFMASK ;
			}
			else
			{
				aaPrintf ("No free command\n") ;
			}
		}

		else if (0 == strcmp ("h", pCmd))
		{
			// Get analog values of Hall sensors
			adcDone = 0 ;
			adcTrig () ;
			while (adcDone == 0)
			{
				aaTaskDelay (1) ;
			}
			aaPrintf ("%4u %4u %4u %4u ", adcValues [0], adcValues [1], adcValues [2], adcValues [3]) ;
			aaPrintf ("%4u %4u %4u %4u\n", adcValues [4], adcValues [5], adcValues [6], adcValues [7]) ;
		}

		else if (0 == strcmp ("+", pCmd))
		{
			direction = 1 ;		// CW
		}

		else if (0 == strcmp ("-", pCmd))
		{
			direction = 0 ;		// CCW
		}

		else if (0 == strcmp ("s", pCmd))
		{
			stepSize = arg1 ;
		}

		else if (0 == strcmp ("c", pCmd))
		{
			pulseCount = arg1 ;
		}

		else if (0 == strcmp ("o", pCmd))
		{
			// Set hands to origin
			goToOrigin () ;
		}

		else if (0 == strcmp ("z", pCmd))
		{
			motorResetOn () ;
			bspDelayUs (1) ;
			motorResetOff () ;
		}

		else if (0 == strcmp ("mp", pCmd))
		{
			// Add a move command
			addMove (0, direction, arg1, arg2) ;
		}

		else if (0 == strcmp ("md", pCmd))
		{
			// Add a move command
			addMoveAngle (0, direction, arg1, arg2) ;
		}

		else if (0 == strcmp ("ma", pCmd))
		{
			// Add a move command
			addMoveAbs (0, arg1, arg2, arg3) ;
		}

		else if (0 == strcmp ("dp", pCmd))
		{
			// Dump hands position
			uint32_t	ii ;
			for (ii = 0 ; ii < MOTOR_COUNT ; ii++)
			{
				aaPrintf ("%u: %4u %3u\n", mBuffer[ii].position, mBuffer[ii].position / PULSE_PER_DEG) ;
			}
		}

		// Early debug only
		else if (0 == strcmp ("a", pCmd))
		{
			mcuAddress  =  (BSP_IO_GPIO (PORT_ADDR0-'A')->IDR >> PIN_ADDR0) & 1u ;
			mcuAddress |= ((BSP_IO_GPIO (PORT_ADDR1-'A')->IDR >> PIN_ADDR1) & 1u) << 1 ;
			mcuAddress |= ((BSP_IO_GPIO (PORT_ADDR2-'A')->IDR >> PIN_ADDR2) & 1u) << 2 ;
			mcuAddress |= ((BSP_IO_GPIO (PORT_ADDR3-'A')->IDR >> PIN_ADDR3) & 1u) << 3 ;

			aaPrintf ("Clock Art V%s  @%d\n", CLOCKART_VERSION, mcuAddress) ;
		}
		else if (0 == strcmp ("i0", pCmd))
		{
			led1_Off		() ;
			tst0_0			() ;
			tst1_0			() ;
			cmdAckOff		() ;
			motorResetOff	() ;
			motorDirCCW		() ;
		}

		else if (0 == strcmp ("i1", pCmd))
		{
			led1_On			() ;
			tst0_1			() ;
			tst1_1			() ;
			cmdAckOn		() ;
			motorResetOn	() ;
			motorDirCW		() ;
		}

		cmdBufferInit (& cmdBuffer) ;
	}
}

//--------------------------------------------------------------------------------
