/*
----------------------------------------------------------------------

	Clock art

	Alain Chebrou

	packet.c	Packet/messages management

	When		Who	What
	10/05/21	ac	Creation

----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aalogmes.h"
#include	"aaprintf.h"

#include	"gpiobasic.h"
#include	"spibasic.h"
#include	"extibasic.h"

#include	"aaUtils.h"		// aaDump

#include	"clock.h"
#include	"packet.h"

static	void	packetProcess (void) ;

#define	WITH_DUMP	0

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------

#define	SPI_DESC			10u					// Slave SPI descriptor ID in spiXXxx.h

static	const spiDesc_t 	* pSpiDesc ;
static	spiHandle_t 		spiHandle ;
static	uint32_t			spiConfigLast ;		// Flags of last spiSlaveConfigure() call
static	uint32_t			bReceived ;
static	uint32_t			bSpiReady ;			// 1 when SPI is ready to receive

static	uint16_t			pktBuffer [PACKET_SIZE] ;

//----------------------------------------------------------------------
// Slave NSS management: interrupt on rising edge
// The Slave SPI NSS is defined in the SPI descriptor: consistency must be ensured

// You must enable the required EXTI interrupt handler in extibasic.h
#define	SPI_NSS_EXTI_PORT	'A'
#define	SPI_NSS_EXTI_LINE	15

BSP_ATTR_ALLWAYS_INLINE  void	nssExtiInit (void)
{
	// Initialize the NSS input pin
	gpioPinDesc_t	pinDesc = {SPI_NSS_EXTI_PORT, SPI_NSS_EXTI_LINE, 0, AA_GPIO_MODE_INPUT_UP } ;

	extiInitGpio (& pinDesc, BSP_IRQPRIOMIN_PLUS(1), EXTI_FLAG_INTR | EXTI_FLAG_RISING) ;
}

// Remember that F4/L4/H7 does not provide the direction of the signal edge
// G0 manages both edges independently

void EXTI_CALLBACK(SPI_NSS_EXTI_LINE) (uint32_t flags)
{
	if ((flags & EXTI_FLAG_RISING) != 0)
	{
//		aaLogMes ("R\n", 0, 0, 0, 0, 0) ;
		bReceived = 1 ;		// Report that a package has arrived.
		bSpiReady = 0 ;
	}
}

//--------------------------------------------------------------------------------
//	Prepare SPI configuration

void	spiInit (void)
{
	// Find the SPI descriptor with index SPI_DESC
	if (spiGetDesc (SPI_DESC, & pSpiDesc) == SPI_EFAIL)
	{
		aaPrintf ("spiGetDesc error\n") ;
		AA_ASSERT (0) ;
	}
	spiConfigLast = 1u ;	// Not configured

	// Reset packet content
	bReceived = 0 ;		// No packet received
	bSpiReady = 0 ;		// SPI not ready to receive

	// Initialize interrupts on NSS edges
	nssExtiInit () ;
}

//----------------------------------------------------------------------

uint32_t	isSpiReady (void)
{
	return bSpiReady ;
}

//----------------------------------------------------------------------
//	Configure SPI only if not already configured
//	This allows to configure only once for many exchanges
//	Use SPI_NSS_USER, so the SPI descriptor should describe NSS pin accordingly

static	void	spiConfigure (const spiDesc_t * pDesc, uint32_t flags, spiHandle_t * pHdl)
{
	spiStatus_t			sts = SPI_ENONE ;

	if (flags != spiConfigLast)
	{
		aaPrintf ("Configure SPI: %08X\n", flags) ;
		spiConfigLast = flags ;
		sts = spiSlaveConfigure (pDesc, SPI_BITS_16 | SPI_NSS_USER | SPI_MODE_0 | flags, pHdl) ;
		if (sts != SPI_ENONE)
		{
			aaPrintf ("spiSlaveConfigure error: %d\n", sts) ;
			spiConfigLast = 1u ;	// Not configured
			AA_ASSERT (0) ;
		}
	}
	else
	{
		// Configuration not changed:
		// Reset and configure again the SPI because there is no way to flush the TX FIFO... Sigh...
		spiConfigAgain (* pHdl) ;
	}
}

//----------------------------------------------------------------------
// Prepare to receive a packet

void	spiReceive ()
{
	spiConfigure (pSpiDesc, SPI_DIR_RX_ONLY, & spiHandle) ;

//aaMEMSET (pktBuffer, 0, sizeof (pktBuffer)) ;
	pktBuffer [0] = 0u ;		// To avoid false receive when the master is reseted
	bReceived     = 0u ;		// No packet received
	spiSlaveExchangeDma (spiHandle, NULL, pktBuffer, PACKET_SIZE) ;
	bSpiReady     = 1u ;
	spiAckSet () ;
}

//----------------------------------------------------------------------
// If a packet has been received, process it

void	spiCheck ()
{
	if (bReceived != 0u)
	{
bspDelayUs (10) ;
		spiSlaveAbort (spiHandle) ;
		packetProcess () ;
		spiReceive() ;
	}
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	Functions to dump / process a packet

static	void	msgDefShort (msg_t * pMsg)
{
	(void) pMsg ;
	AA_ASSERT (0) ;		// It's an error to come here
}

//--------------------------------------------------------------------------------

static	void	msgDelay (msg_t * pMsg)
{
	uint32_t	ii ;
	uint32_t	delay ;

	#if (WITH_DUMP == 1)
		aaPrintf ("MSG_DELAY     %u\n", pMsg->msgHdr.cmd & ~MSG_SHORT_M) ;
	#endif

	delay = pMsg->msgHdr.cmd & ~MSG_SHORT_M ;
	if (pMsg->msgHdr.motor == MSG_MOTOR_ALL)
	{
		for (ii = 0 ; ii < MOTOR_COUNT ; ii++)
		{
			addDelay (ii, delay) ;
		}
	}
	else
	{
		addDelay (pMsg->msgHdr.motor, delay) ;
	}

}


//--------------------------------------------------------------------------------
//	Direction must be FLAG_MOVE_CW or FLAG_MOVE_CCW

static	void	msgMovePulseRel (msg_t * pMsg)
{
	uint32_t	dir ;
	uint32_t	ii ;

	#if (WITH_DUMP == 1)
		aaPrintf ("MSG_MOVE_PULSE_REL  %3u pls %4u ms %s\n", pMsg->msgHdr.cmd & FLAG_MOVE_PLS_M, pMsg->data [0],
				((pMsg->msgHdr.cmd & FLAG_MOVE_PLS_DIR_M) == FLAG_MOVE_PLS_CW) ? "CW" :
				((pMsg->msgHdr.cmd & FLAG_MOVE_PLS_DIR_M) == FLAG_MOVE_PLS_CCW) ? "CCW" :
				((pMsg->msgHdr.cmd & FLAG_MOVE_PLS_DIR_M) == FLAG_MOVE_PLS_SHORT) ? "SHORT ***" : "LONG ***") ;
	#endif

	dir =	((pMsg->msgHdr.cmd & FLAG_MOVE_PLS_DIR_M) == FLAG_MOVE_PLS_CW) ? DIR_CW : DIR_CCW ;

	if (pMsg->msgHdr.motor == MSG_MOTOR_ALL)
	{
		for (ii = 0 ; ii < MOTOR_COUNT ; ii++)
		{
			addMove (ii,
					dir,
					pMsg->msgHdr.cmd & FLAG_MOVE_PLS_M,
					pMsg->data [0]) ;
		}
	}
	else
	{
		addMove (pMsg->msgHdr.motor,
				dir,
				pMsg->msgHdr.cmd & FLAG_MOVE_PLS_M,
				pMsg->data [0]) ;
	}
}

//--------------------------------------------------------------------------------

void	(* const fMsgShort []) (msg_t * pMsg)  =
{
	msgDelay,
	msgMovePulseRel,
	msgDefShort,
	msgDefShort
} ;

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------

static	void	msgDefLong (msg_t * pMsg)
{
	(void) pMsg ;
	AA_ASSERT (0) ;		// It's an error to come here
}

//--------------------------------------------------------------------------------

static	void	msgZero (msg_t * pMsg)
{
	(void) pMsg ;

	#if (WITH_DUMP == 1)
		aaPrintf ("MSG_ZERO\n") ;
	#endif
	goToOrigin () ;
}


//--------------------------------------------------------------------------------

static	void	msgMoveAbs (msg_t * pMsg)
{
	uint32_t	dir ;
	uint32_t	ii ;

	#if (WITH_DUMP == 1)
		aaPrintf ("MSG_MOVE_ABS  %3u deg %4u ms %s\n", pMsg->msgHdr.cmd & FLAG_MOVE_ANGLE_M, pMsg->data [0],
				((pMsg->msgHdr.cmd & FLAG_MOVE_DIR_M) == FLAG_MOVE_CW) ? "CW" :
				((pMsg->msgHdr.cmd & FLAG_MOVE_DIR_M) == FLAG_MOVE_CCW) ? "CCW" :
				((pMsg->msgHdr.cmd & FLAG_MOVE_DIR_M) == FLAG_MOVE_SHORT) ? "SHORT" : "LONG") ;
	#endif

	dir =	((pMsg->msgHdr.cmd & FLAG_MOVE_DIR_M) == FLAG_MOVE_CW) ? DIR_CW :
			((pMsg->msgHdr.cmd & FLAG_MOVE_DIR_M) == FLAG_MOVE_CCW) ? DIR_CCW :
			((pMsg->msgHdr.cmd & FLAG_MOVE_DIR_M) == FLAG_MOVE_SHORT) ? DIR_SHORT : DIR_LONG ;

	if (pMsg->msgHdr.motor == MSG_MOTOR_ALL)
	{
		for (ii = 0 ; ii < MOTOR_COUNT ; ii++)
		{
			addMoveAbs (ii,
						dir,
						pMsg->msgHdr.cmd & FLAG_MOVE_ANGLE_M,
						pMsg->data [0]) ;
		}
	}
	else
	{
		addMoveAbs (pMsg->msgHdr.motor,
					dir,
					pMsg->msgHdr.cmd & FLAG_MOVE_ANGLE_M,
					pMsg->data [0]) ;
	}
}

//--------------------------------------------------------------------------------
//	Direction must be FLAG_MOVE_CW or FLAG_MOVE_CCW

static	void	msgMoveRel (msg_t * pMsg)
{
	uint32_t	dir ;
	uint32_t	ii ;

	#if (WITH_DUMP == 1)
		aaPrintf ("MSG_MOVE_REL  %3u deg %4u ms %s\n", pMsg->msgHdr.cmd & FLAG_MOVE_ANGLE_M, pMsg->data [0],
				((pMsg->msgHdr.cmd & FLAG_MOVE_DIR_M) == FLAG_MOVE_CW) ? "CW" :
				((pMsg->msgHdr.cmd & FLAG_MOVE_DIR_M) == FLAG_MOVE_CCW) ? "CCW" :
				((pMsg->msgHdr.cmd & FLAG_MOVE_DIR_M) == FLAG_MOVE_SHORT) ? "SHORT ***" : "LONG ***") ;
	#endif

	dir =	((pMsg->msgHdr.cmd & FLAG_MOVE_DIR_M) == FLAG_MOVE_CW) ? DIR_CW : DIR_CCW ;

	if (pMsg->msgHdr.motor == MSG_MOTOR_ALL)
	{
		for (ii = 0 ; ii < MOTOR_COUNT ; ii++)
		{
			addMoveAngle (ii,
						dir,
						pMsg->msgHdr.cmd & FLAG_MOVE_ANGLE_M,
						pMsg->data [0]) ;
		}
	}
	else
	{
		addMoveAngle (pMsg->msgHdr.motor,
					dir,
					pMsg->msgHdr.cmd & FLAG_MOVE_ANGLE_M,
					pMsg->data [0]) ;
	}
}


//--------------------------------------------------------------------------------
// The LED message is for the controller, not a motor

static	void	msgLed (msg_t * pMsg)
{
	#if (WITH_DUMP == 1)
		aaPrintf ("MSG_LED       %u\n", pMsg->msgHdr.cmd & ~MSG_LONG_M) ;
	#endif

	if ((pMsg->msgHdr.cmd & ~MSG_LONG_M) == 0)
	{
		bspOutput (BSP_LED1, 0) ;
	}
	else
	{
		bspOutput (BSP_LED1, 1) ;
	}
}


//--------------------------------------------------------------------------------
//	Add a Ack command to the motor buffer
//	If the argument is not 0 => add a delay in ms unit

static	void	msgAck (msg_t * pMsg)
{
	uint32_t	ii ;
	uint32_t	delay ;

	#if (WITH_DUMP == 1)
		aaPrintf ("MSG_ACK       %u\n", pMsg->msgHdr.cmd & ~MSG_LONG_M) ;
	#endif

	delay = pMsg->msgHdr.cmd & ~MSG_LONG_M ;
	if (pMsg->msgHdr.motor == MSG_MOTOR_ALL)
	{
		for (ii = 0 ; ii < MOTOR_COUNT ; ii++)
		{
			if (delay != 0u)
			{
				addDelay (ii, delay) ;
			}
			addMotorAck (ii) ;
		}
	}
	else
	{
		if (delay != 0u)
		{
			addDelay (pMsg->msgHdr.motor, delay) ;
		}
		addMotorAck (pMsg->msgHdr.motor) ;
	}
}

//--------------------------------------------------------------------------------
//	Controller message: reset the ack output

static	void	msgResetAck (msg_t * pMsg)
{
	(void) pMsg ;

	#if (WITH_DUMP == 1)
		aaPrintf ("MSG_RESET_ACK\n") ;
	#endif

	motorAckReset () ;
}

//--------------------------------------------------------------------------------

void	(* const fMsgLong []) (msg_t * pMsg)  =
{
		msgZero,		// 0x0000
		msgMoveAbs,		// 0x0800
		msgMoveRel,		// 0x1000
		msgLed,			// 0x1800
		msgAck,			// 0x2000
		msgDefLong,		// 0x2800
		msgDefLong,		// 0x3000
		msgDefLong,		// 0x3800
		msgDefLong,		// 0x4000
		msgDefLong,		// 0x4800
		msgDefLong,		// 0x5000
		msgDefLong,		// 0x5800
		msgDefLong,		// 0x6000
		msgDefLong,		// 0x6800
		msgDefLong,		// 0x7000
		msgResetAck		// 0x7800
} ;

//--------------------------------------------------------------------------------
//----------------------------------------------------------------------

static	void	packetProcess (void)
{
	pktHdr_t		* pPktHdr = (pktHdr_t *) pktBuffer ;
	msg_t			* pMsg ;
	msg_t			* pLast ;
	uint32_t		ii, sum, cmd ;

	// Check packet
	if (pPktHdr->synchro != PKT_SYNCHRO)
	{
		aaPrintf ("Synchro error %04X\n", pPktHdr->synchro) ;
		return ;
	}

	sum = 0 ;
	for (ii = 0 ; ii < pPktHdr->size ; ii++)
	{
		sum += pktBuffer [ii] ;
	}
	if ((sum & 0x0FFFF) != 0)
	{
		aaPrintf ("Sum error %04X\n", sum & 0x0FFFFu) ;
aaDump ((uint8_t *) pktBuffer, pPktHdr->size*2) ;
		return ;
	}

	// Process messages
	pMsg  = (msg_t *) (& pktBuffer [sizeof (pktHdr_t) / sizeof (uint16_t) ]) ;
	pLast = (msg_t *) (& pktBuffer [pPktHdr->size - 1]) ;
	while (pMsg != pLast)
	{
		#if (WITH_DUMP == 1)
			aaPrintf ("%08X  %04X %u   Board %2u Motor %u %s ", pMsg,
					pMsg->msgHdr.cmd, pMsg->msgHdr.size,
					pMsg->msgHdr.board, pMsg->msgHdr.motor,
					(pMsg->msgHdr.motor == 8) ? "(All)" : "") ;
		#endif

		if (pMsg->msgHdr.board == mcuAddress)
		{
			// The message is for this board
			cmd = pMsg->msgHdr.cmd ;
			if ((cmd & MSG_LONGBIT) == 0)
			{
				// Short command
				(* fMsgShort [cmd >> 13]) (pMsg) ;
			}
			else
			{
				// Long command
				cmd &= ~MSG_LONGBIT ;
				(* fMsgLong [cmd >> 11]) (pMsg) ;
			}
		}
		else
		{
			#if (WITH_DUMP == 1)
				aaPutChar ('\n') ;
			#endif
		}

		// Next message
		pMsg = (msg_t *) & (((uint16_t *) pMsg) [pMsg->msgHdr.size]) ;
	}
}

//--------------------------------------------------------------------------------
