/*
----------------------------------------------------------------------
	
	AdAstra - Real Time Kernel

	Alain Chebrou

	userInitTask.c	The first task

	When		Who	What
	25/11/13	ac	Creation

----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aalogmes.h"
#include	"aaprintf.h"
#include	"aautils.h"		// For displaytaskInfo()

#include	"clock.h"

//--------------------------------------------------------------------------------
//	Demo : Blink LEDs and test user button
//	Use BSP_LED0 and BSP_BUTTON0 because most boards have at least a user LED and a user button

// Define the LEDs task stack, to avoid use of dynamic memory
#define	LED_STACK_SIZE	64u
static	bspStackType_t	ledStack [LED_STACK_SIZE] ;

void	ledsTask (uintptr_t arg)
{
	(void) arg ;

	while (1)
	{
		// Blink the LED at 1Hz
		bspOutput (BSP_LED0, 1) ;
		aaTaskDelay (bspGetTickRate () / 2) ;

		// The LED is on as long as the button is pressed
		if (bspInput (BSP_BUTTON0) == 1)
		{
			bspOutput (BSP_LED0, 0) ;
			aaTaskDelay (bspGetTickRate () / 2) ;
		}
	}
}

//----------------------------------------------------------------------
//	Test of G0 USART

void	uartPutString (uartHandle_t hUsart, const char * str)
{
	const char * pStr = str ;

	while (* pStr != 0)
	{
		uartPutChar (hUsart, * pStr++) ;
	}
}

void	uartGetString (uartHandle_t hUsart)
{
	char	cc ;

	while (uartCheckGetChar (hUsart) != 0)
	{
		cc = uartGetChar (hUsart) ;
		aaPutChar (cc) ;
	}
}

#define	TUART	101

void	tUsart (void)
{
	uartHandle_t	hUsart ;
	aaError_t		err ;
	char			cc ;

	// -------------------------------------------- Test without interrupt
	err = uartInit (TUART, 19200, UART_FLAG_NOIT, & hUsart) ;
	AA_ASSERT (err == AA_ENONE) ;

	while (1)
	{
		cc = aaGetChar () ;
		if (cc == 'n')
			break ;

		uartPutChar (hUsart, 'A') ;
		aaTaskDelay (10) ;
		if (uartCheckGetChar (hUsart) != 0)
		{
			cc = uartGetChar (hUsart) ;
			aaPrintf ("OK: %c\n", cc) ;
		}
		else
		{
			aaPrintf ("Nothing\n", cc) ;
		}
	}
	uartClose (hUsart) ;

	// -------------------------------------------- Test with interrupt
	err = uartInit (TUART, 19200, 0, & hUsart) ;
	AA_ASSERT (err == AA_ENONE) ;

	while (1)
	{
		cc = aaGetChar () ;
		if (cc == 'n')
			break ;

		uartPutString (hUsart, "123456789\r\n") ;
		aaTaskDelay (10) ;
		if (uartCheckGetChar (hUsart) != 0)
		{
			uartGetString (hUsart) ;
		}
		else
		{
			aaPrintf ("Nothing\n", cc) ;
		}
	}
	uartClose (hUsart) ;
}

void	tDelay (void)
{
	uint32_t	start1, start2, delta ;

	aaTaskDelay (2) ;

	start1 = start2 = bspTsGet () ;
	aaTaskDelay (2) ;
	delta = bspTsDelta (& start2) ;
	aaPrintf ("Delta = %u %u %u\n", start1, start2, delta) ;
}

//----------------------------------------------------------------------
//	The first user task

//	The C++ entry, for demo purpose.
//	You can start your C++ code from there
void cppEntry (uintptr_t arg) ;

void	clock (void) ;

void	userInitTask (uintptr_t arg)
{
	// Remove all the lines that follow, and replace them with your application
	// If your application is entirely C++ then call cppEntry()
	// cppEntry (arg) ;

	(void) arg ;		// Unused arg

	// Create a task to blink LEDS
	aaTaskCreate (
			BSP_IRQPRIOMIN_PLUS(1),		// Task priority
			"tLeds",			// Task name
			ledsTask,			// Entry point,
			0,					// Entry point parameter
			ledStack,			// Stack pointer
			LED_STACK_SIZE,		// Stack size
			0,					// Flags
			NULL) ;				// Created task id

	clock () ;
	// No return ;

	// Use console: echo received char, assume AA_WITH_CONSOLE is 1 (console enabled)
	{
		char		cc ;

		while (1)
		{
			aaPrintf ("User task, t to execute the test\n") ;
			cc = aaGetChar () ;
			aaPutChar (cc) ;
			aaPutChar ('\n') ;

			if (cc == 'q')
			{
				break ;	// End of test, kill this task
			}

/*			if (cc == '+')
			{
				// To start the C++ application: call cppEntry()
				cppEntry (arg) ;
			}
*/
			if (cc == 't')
			{
				// Go to the test menu defined in test.c
//				tests () ;
				continue ;
			}

			#if (AA_WITH_MALLOC_TLSF == 1)
			{
				if (cc == 'm')
				{
					// Check dynamic memory partition
					{
						aaMemPoolCheck (1) ;
					}
					continue ;
				}
			}
			#endif

			// Display amount of free space
			#if (AA_WITH_MALLOC_BLOC == 1)
			{
				if (cc == 'b')
				{
					aaPrintf ("Malloc Block free space %d\n", aaMallocSize ()) ;
				}
			}
			#endif

			if (cc == 'u')
			{
				// Display task informations
				// To enable stack check, add AA_FLAG_STACKCHECK flag to each task creation
				displaytaskInfo () ;
				continue ;
			}

			if (cc == 'c')
			{
				// Clear task counters
				aaTaskStatClear () ;
				continue ;
			}

			// Use aaLogMes() to see if the UART is enabled (if DEBUG)
			aaLogMes ("log! %c\n", cc, 0, 0, 0, 0) ;
		}
	}

	aaPrintf ("Exit user task\n") ;
	aaTaskDelete (AA_SELFTASKID) ;	// Useless, example of use.
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	Manage kernel notifications

void	aaUserNotify (uint32_t	event, uintptr_t arg)
{
	switch (event)
	{
		case AA_NOTIFY_STACKOVFL:	// Stack overflow
			AA_ASSERT (0) ;
			break ;

		case AA_NOTIFY_STACKTHR:	// Stack threshold level is reached
			AA_ASSERT (0) ;
			break ;

		default:
			break ;
	}
	(void) arg ;
}

//--------------------------------------------------------------------------------
