/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	dmabasic.c	DMA management for L4+ and G0

	When		Who	What
	23/04/18	ac	Creation
	27/02/20	ac	Port to STM32H7xx
	13/04/21	ac	Port to STM32G0xx
	20/04/21	ac	Add dmaSetCallbackRaw()
					Unify G0 / L4+

	G0 DMA2 not tested

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2021 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"

#include	"gpiobasic.h"
#include	"dmabasic.h"

//----------------------------------------------------------------------

#if (defined STM32G0)
	#define		DMA2_OFFSET		7		// DMA1 7 channels, DMA2 5 channels if available
#elif
	#define		DMA2_OFFSET		7		// DMA1 7 channels, DMA2 7 channels
#endif

//	Array to get interrupt number of DMA stream
static	const IRQn_Type	dmaStreamIrqIndex [] =
{
	#if (defined STM32G0)
			DMA1_Channel1_IRQn,
			DMA1_Channel2_3_IRQn,
			DMA1_Channel2_3_IRQn,
		#if (! defined DMA2)
			DMA1_Ch4_7_DMAMUX1_OVR_IRQn,
			DMA1_Ch4_7_DMAMUX1_OVR_IRQn,
			DMA1_Ch4_7_DMAMUX1_OVR_IRQn,
			DMA1_Ch4_7_DMAMUX1_OVR_IRQn
		#else
			DMA1_Ch4_7_DMA2_Ch1_5_DMAMUX1_OVR_IRQn,
			DMA1_Ch4_7_DMA2_Ch1_5_DMAMUX1_OVR_IRQn,
			DMA1_Ch4_7_DMA2_Ch1_5_DMAMUX1_OVR_IRQn,
			DMA1_Ch4_7_DMA2_Ch1_5_DMAMUX1_OVR_IRQn,

			DMA1_Ch4_7_DMA2_Ch1_5_DMAMUX1_OVR_IRQn,
			DMA1_Ch4_7_DMA2_Ch1_5_DMAMUX1_OVR_IRQn,
			DMA1_Ch4_7_DMA2_Ch1_5_DMAMUX1_OVR_IRQn,
			DMA1_Ch4_7_DMA2_Ch1_5_DMAMUX1_OVR_IRQn,
			DMA1_Ch4_7_DMA2_Ch1_5_DMAMUX1_OVR_IRQn,
		#endif
	#endif	// G0

	#if (defined STM32L4)
		DMA1_Channel1_IRQn,			// DMA1 1-7
		DMA1_Channel2_IRQn,
		DMA1_Channel3_IRQn,
		DMA1_Channel4_IRQn,
		DMA1_Channel5_IRQn,
		DMA1_Channel6_IRQn,
		DMA1_Channel7_IRQn,

		DMA2_Channel1_IRQn,			// DMA2 1-7
		DMA2_Channel2_IRQn,
		DMA2_Channel3_IRQn,
		DMA2_Channel4_IRQn,
		DMA2_Channel5_IRQn,
		DMA2_Channel6_IRQn,
		DMA2_Channel7_IRQn,
	#endif	// L4
} ;

//----------------------------------------------------------------------

// DMA callback descriptor
typedef struct
{
	dmaCallback_t	fn ;	// Function to call, NULL if no function
	uintptr_t		arg ;	// User argument

} dmaCallbackDesc_t;

static	void dmaDefCallback (uint32_t event, uint32_t dir, uintptr_t arg) ;

static	dmaCallbackDesc_t	dmaCallbackArray [] =
{
	#if (defined STM32G0)
		{ dmaDefCallback, 0u },			// DMA1 1-7
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		#if (defined DMA2)
			{ dmaDefCallback, 0u },		// DMA2 1-5
			{ dmaDefCallback, 0u },
			{ dmaDefCallback, 0u },
			{ dmaDefCallback, 0u },
			{ dmaDefCallback, 0u },
		#endif
	#endif	// G0

	#if (defined STM32L4)
		{ dmaDefCallback, 0u },			// DMA1 1-7
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },

		{ dmaDefCallback, 0u },			// DMA2 1-7
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
		{ dmaDefCallback, 0u },
	#endif	// L4
} ;

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Default DMA callback

static	void dmaDefCallback (uint32_t event, uint32_t dir, uintptr_t arg)
{
	(void) event ;
	(void) dir ;
	(void) arg ;
}

//----------------------------------------------------------------------
//	Enable DMA peripheral clock

void		dmaEnableClock	(dma_t * pDma)
{
	#if (defined STM32G0)
		#if (defined DMA2)
			rccEnableClock (pDma, (pDma == (dma_t *) DMA1) ? RCC_AHBENR_DMA1EN : RCC_AHBENR_DMA2EN) ;
		#else
			rccEnableClock (pDma, RCC_AHBENR_DMA1EN) ;
		#endif

		// No clock enable for DMAMUX on G0
	#endif	// G0

	#if (defined STM32L4)
		rccEnableClock (pDma, (pDma == (dma_t *) DMA1) ? RCC_AHB1ENR_DMA1EN : RCC_AHB1ENR_DMA2EN) ;
		rccEnableClock (DMAMUX1, RCC_AHB1ENR_DMAMUX1EN) ;
	#endif	// L4
}

//----------------------------------------------------------------------
//	Return DMA_ISR flags for streamNum right aligned
//	Return a combination of: DMA_ISR_GIF1, DMA_ISR_TCIF1, DMA_ISR_HTIF1, DMA_ISR_TEIF1

uint32_t	dmaGetFlags		(dma_t * pDma, uint32_t streamNum)
{
	uint32_t	isr ;

	isr = pDma->ISR ;
	return ((isr >> (streamNum << 2)) & 0x0Fu) ;	// Each stream have 4 flags
}

//----------------------------------------------------------------------
//	Allows to clear DMA_ISR flags for streamNum. Bits are right aligned
//	Use a combination of: DMA_FLAG_FEIF, DMA_FLAG_DMEIF, DMA_FLAG_TEIF, DMA_FLAG_HTIF, DMA_FLAG_TCIF

void	dmaClearFlags	(dma_t * pDma, uint32_t streamNum, uint32_t flags)
{
	pDma->IFCR = (flags & 0x0Fu) << (streamNum << 2u) ;
}

//----------------------------------------------------------------------
// Configure and enable DMA interrupt

void	dmaSetIrq (const dmaDesc_t * pDmaDesc)
{
	uint32_t	irqIdx ;
	IRQn_Type	irqNum ;

	irqIdx = ((pDmaDesc->dmaNum == 1u) ? 0 : DMA2_OFFSET) + DMA_GETSTREAM (pDmaDesc) ;
	irqNum = dmaStreamIrqIndex [irqIdx] ;

	NVIC_SetPriority	(irqNum, pDmaDesc->irqPrio);
	NVIC_EnableIRQ		(irqNum) ;
}

//----------------------------------------------------------------------
// Disable DMA interrupt

void	dmaDisableIrq (const dmaDesc_t * pDmaDesc)
{
	uint32_t	irqIndex ;
	IRQn_Type	irqNum ;

	irqIndex = ((pDmaDesc->dmaNum == 1u) ? 0 : DMA2_OFFSET) + DMA_GETSTREAM (pDmaDesc) ;
	irqNum   = dmaStreamIrqIndex [irqIndex] ;
	
	NVIC_DisableIRQ (irqNum) ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Set DMA callback using DMA descriptor

void	dmaSetCallback (const dmaDesc_t * pDmaDesc, dmaCallback_t fn, uintptr_t arg)
{
	uint32_t	dmaIndex = ((pDmaDesc->dmaNum == 1u) ? 0 : DMA2_OFFSET) + DMA_GETSTREAM (pDmaDesc) ;

	if (fn == NULL)
	{
		// No user callback, set default callback
		dmaCallbackArray [dmaIndex].fn  = dmaDefCallback ;
	}
	else
	{
		dmaCallbackArray [dmaIndex].fn  = fn ;
		dmaCallbackArray [dmaIndex].arg = arg ;
	}
}

//----------------------------------------------------------------------
//	Set DMA callback using DMA pointer and stream number

void	dmaSetCallbackRaw (dma_t * pDma, uint32_t streamNum, dmaCallback_t fn, uintptr_t arg)
{
	uint32_t	dmaIndex = (((DMA_TypeDef *) pDma == DMA1) ? 0 : DMA2_OFFSET) + streamNum ;

	if (fn == NULL)
	{
		// No user callback, set default callback
		dmaCallbackArray [dmaIndex].fn  = dmaDefCallback ;
	}
	else
	{
		dmaCallbackArray [dmaIndex].fn  = fn ;
		dmaCallbackArray [dmaIndex].arg = arg ;
	}
}

//----------------------------------------------------------------------
//	dmaNum: 1..2
//	streamNum: 0..n

static	void	DMA_Stream_IRQHandler (uint32_t dmaNum, uint32_t streamNum)
{
	dma_t				* pDma = DMA_GETDMA (dmaNum) ;
	dmaCallbackDesc_t	* pCb = & dmaCallbackArray [((dmaNum == 1u) ? 0 : DMA2_OFFSET) + streamNum] ;
	uint32_t			reg ;
	uint32_t			maskBase, mask ;
	dmaStream_t			* pStream ;

	aaIntEnter () ;

	pStream = & pDma->stream [streamNum] ;

	// Read interrupt status register
	// Keep only interrupts flags for enabled interrupts for this stream
	reg = pDma->ISR ;
	reg &= (pStream->CCR & 0x0Eu) << (streamNum << 2u) ;

	// maskBase: mask of LSB interrupt flag for the stream
	// To get other flags mask: left shift maskBase using stream0 pos values
	maskBase = 1u << (streamNum << 2u) ;

	// Check transfer complete
	mask = maskBase << DMA_ISR_TCIF1_Pos ;
	if ((reg & mask) != 0u)
	{
		// Clear flag
		pDma->IFCR = mask ;

		(* pCb->fn) (DMA_EV_COMPLETE, dmaNum << 16u | streamNum, pCb->arg) ;
	}

	// Check half transfer complete
	mask =  maskBase << DMA_ISR_HTIF1_Pos ;
	if ((reg & mask) != 0u)
	{
		// Clear flag
		pDma->IFCR = mask ;

		(* pCb->fn) (DMA_EV_HCOMPLETE, dmaNum << 16u | streamNum, pCb->arg) ;
	}

	// Check transfer error
	mask = maskBase <<  DMA_ISR_TEIF1_Pos ;
	if ((reg & mask) != 0u)
	{
		// Clear flag
		pDma->IFCR = mask ;

		(* pCb->fn) (DMA_EV_ERROR, dmaNum << 16u | streamNum, pCb->arg) ;
	}

	aaIntExit () ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	STM32G0 DMA1 and DMA2 IRQ handlers

#if (defined STM32G0)

void DMA1_Channel1_IRQHandler (void) ;
void DMA1_Channel1_IRQHandler (void)
{
	DMA_Stream_IRQHandler (1u, 0u) ;
}

//----------------------------------------------------------------------

void DMA1_Channel2_3_IRQHandler (void) ;
void DMA1_Channel2_3_IRQHandler (void)
{
	dma_t		* pDma = (dma_t *) DMA1 ;
	uint32_t	isr = pDma->ISR ;
	uint32_t	flags ;

	flags = pDma->stream [1].CCR ;
	flags &= (isr >> DMA_ISR_GIF2_Pos) & 0x0Eu ;
	if (flags != 0)
	{
		DMA_Stream_IRQHandler (1u, 1u) ;
	}

	flags = pDma->stream [2].CCR ;
	flags &= (isr >> DMA_ISR_GIF3_Pos) & 0x0Eu ;
	if (flags != 0)
	{
		DMA_Stream_IRQHandler (1u, 2u) ;
	}
}

//----------------------------------------------------------------------
#if (! defined DMA2)
void DMA1_Ch4_7_DMAMUX1_OVR_IRQHandler (void) ;
void DMA1_Ch4_7_DMAMUX1_OVR_IRQHandler (void)
#else
void DMA1_Ch4_7_DMA2_Ch1_5_DMAMUX1_OVR_IRQHandler (void) ;
void DMA1_Ch4_7_DMA2_Ch1_5_DMAMUX1_OVR_IRQHandler (void)
#endif
{
	dma_t		* pDma = (dma_t *) DMA1 ;
	uint32_t	isr = pDma->ISR ;
	uint32_t	flags ;

	flags = pDma->stream [3].CCR ;
	flags &= (isr >> DMA_ISR_GIF4_Pos) & 0x0Eu ;
	if (flags != 0)
	{
		DMA_Stream_IRQHandler (1u, 3u) ;
	}

	flags = pDma->stream [4].CCR ;
	flags &= (pDma->ISR >> DMA_ISR_GIF5_Pos) & 0x0Eu ;
	if (flags != 0)
	{
		DMA_Stream_IRQHandler (1u, 4u) ;
	}

	flags = pDma->stream [5].CCR ;
	flags &= (pDma->ISR >> DMA_ISR_GIF6_Pos) & 0x0Eu ;
	if (flags != 0)
	{
		DMA_Stream_IRQHandler (1u, 5u) ;
	}

	flags = pDma->stream [6].CCR ;
	flags &= (pDma->ISR >> DMA_ISR_GIF7_Pos) & 0x0Eu ;
	if (flags != 0)
	{
		DMA_Stream_IRQHandler (1u, 6u) ;
	}

#if (defined DMA2)
	pDma = (dma_t *) DMA2 ;
	isr = pDma->ISR ;

	flags = pDma->stream [0].CCR ;
	flags &= (isr >> DMA_ISR_GIF1_Pos) & 0x0Eu ;
	if (flags != 0)
	{
		DMA_Stream_IRQHandler (2u, 0u) ;
	}

	flags = pDma->stream [1].CCR ;
	flags &= (isr >> DMA_ISR_GIF2_Pos) & 0x0Eu ;
	if (flags != 0)
	{
		DMA_Stream_IRQHandler (2u, 1u) ;
	}

	flags = pDma->stream [2].CCR ;
	flags &= (isr >> DMA_ISR_GIF3_Pos) & 0x0Eu ;
	if (flags != 0)
	{
		DMA_Stream_IRQHandler (2u, 2u) ;
	}

	flags = pDma->stream [3].CCR ;
	flags &= (isr >> DMA_ISR_GIF4_Pos) & 0x0Eu ;
	if (flags != 0)
	{
		DMA_Stream_IRQHandler (2u, 3u) ;
	}

	flags = pDma->stream [4].CCR ;
	flags &= (isr >> DMA_ISR_GIF5_Pos) & 0x0Eu ;
	if (flags != 0)
	{
		DMA_Stream_IRQHandler (2u, 4u) ;
	}
#endif
}

#endif	// G0

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	STM32L4 DMA1 and DMA2 IRQ handlers

#if (defined STM32L4)
void DMA1_Channel1_IRQHandler (void) ;
void DMA1_Channel1_IRQHandler (void)
{
	DMA_Stream_IRQHandler (1u, 0u) ;
}

void DMA1_Channel2_IRQHandler (void) ;
void DMA1_Channel2_IRQHandler (void)
{
	DMA_Stream_IRQHandler (1u, 1u) ;
}

void DMA1_Channel3_IRQHandler (void) ;
void DMA1_Channel3_IRQHandler (void)
{
	DMA_Stream_IRQHandler (1u, 2u) ;
}

void DMA1_Channel4_IRQHandler (void) ;
void DMA1_Channel4_IRQHandler (void)
{
	DMA_Stream_IRQHandler (1u, 3u) ;
}

void DMA1_Channel5_IRQHandler (void) ;
void DMA1_Channel5_IRQHandler (void)
{
	DMA_Stream_IRQHandler (1u, 4u) ;
}

void DMA1_Channel6_IRQHandler (void) ;
void DMA1_Channel6_IRQHandler (void)
{
	DMA_Stream_IRQHandler (1u, 5u) ;
}

void DMA1_Channel7_IRQHandler (void) ;
void DMA1_Channel7_IRQHandler (void)
{
	DMA_Stream_IRQHandler (1u, 6u) ;
}

//----------------------------------------------------------------------
//	DMA2 IRQ handlers

void DMA2_Channel1_IRQHandler (void) ;
void DMA2_Channel1_IRQHandler (void)
{
	DMA_Stream_IRQHandler (2u, 0u) ;
}

void DMA2_Channel2_IRQHandler (void) ;
void DMA2_Channel2_IRQHandler (void)
{
	DMA_Stream_IRQHandler (2u, 1u) ;
}

void DMA2_Channel3_IRQHandler (void) ;
void DMA2_Channel3_IRQHandler (void)
{
	DMA_Stream_IRQHandler (2u, 2u) ;
}

void DMA2_Channel4_IRQHandler (void) ;
void DMA2_Channel4_IRQHandler (void)
{
	DMA_Stream_IRQHandler (2u, 3u) ;
}

void DMA2_Channel5_IRQHandler (void) ;
void DMA2_Channel5_IRQHandler (void)
{
	DMA_Stream_IRQHandler (2u, 4u) ;
}

void DMA2_Channel6_IRQHandler (void) ;
void DMA2_Channel6_IRQHandler (void)
{
	DMA_Stream_IRQHandler (2u, 5u) ;
}

void DMA2_Channel7_IRQHandler (void) ;
void DMA2_Channel7_IRQHandler (void)
{
	DMA_Stream_IRQHandler (2u, 6u) ;
}

#endif	// L4

//----------------------------------------------------------------------
