/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	dmabasic.c	DMA management for L4+ and G0

	When		Who	What
	23/04/18	ac	Creation
	27/02/20	ac	Port to STM32H7xx
	13/04/21	ac	Port to STM32G0xx
	20/04/21	ac	Add dmaSetCallbackRaw()
					Unify G0 / L4+

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2021 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#if ! defined	DMABASIC_H_
#define DMABASIC_H_
//----------------------------------------------------------------------

#if (defined STM32G0)
	#include	"stm32g0xx_ll_dma.h"
#elif (defined STM32L4)
	#include	"stm32l4xx_ll_dma.h"
#elif
	#error "MCU not supported"
#endif

//----------------------------------------------------------------------
// No cache for L4 and G0, declaration for H7 compatibility only

// BEWARE: If DMA is used for I/O, the library must manage the data cache.
// So align buffers address on multiple of cache line size (32 bytes)
// Buffers size should be a multiple of cache line size (32 bytes). Mandatory for RX buffers.

// Macro to get variable aligned on 32-bytes,needed for cache maintenance purpose

# if ! defined ALIGN_32BYTES
	#if defined   (__GNUC__)		// GNU Compiler
		#define ALIGN_32BYTES(buf)  buf __attribute__ ((aligned (32)))
	#elif defined (__ICCARM__)		// IAR Compiler
		#define ALIGN_32BYTES(buf) _Pragma("data_alignment=32") buf
	#elif defined   (__CC_ARM)		// ARM Compiler
		#define ALIGN_32BYTES(buf) __align(32) buf
	#endif
#endif

// Macro to round the size of an array to a upper multiple of 32
#define	SIZE_32BYTES(count,type)	(((sizeof (type) * (count)) + 31u) & ~0x0000001Fu)

//----------------------------------------------------------------------

// Get pointer to DMA device from DMA number
#if (defined DMA2)
		#define	DMA_GETDMA(dmaNum)		((dma_t *)(((dmaNum) == 1u) ? DMA1 : DMA2))
#else
		#define	DMA_GETDMA(dmaNum)		((dma_t *)DMA1)
#endif
#define	DMA_GETSTREAM(pDesc)			((pDesc)->streamNum)

#define	DMA_NONE	0
#define	DMA_DMA1	1
#define	DMA_DMA2	2

typedef struct
{
	uint8_t		dmaNum ;		//	DMA number 1 or 2, DMA_NONE => unused descriptor
	uint8_t		streamNum ;		//	Stream number 0..n depending on MCU
	uint8_t		dir ;			//	Direction, e.g. LL_DMA_DIRECTION_PERIPH_TO_MEMORY
	uint8_t		irqPrio ;		//	Interrupt priority

} dmaDesc_t ;

// Event of DMA IRQ callback function
#define	DMA_EV_COMPLETE		1		// Transfer complete
#define	DMA_EV_HCOMPLETE	2		// Transfer half complete
#define	DMA_EV_ERROR		3		// Transfer error
#define	DMA_EV_GLOBAL		4		// Global interrupt

// Callback parameter source is 'dmaNum << 16 | streamNum' :
//		High word : DMA number (1 or 2)
//		Low word  : stream number (0..6)

// Prototype of DMA IRQ callback function
typedef void (* dmaCallback_t) (uint32_t event, uint32_t source, uintptr_t arg) ;

//----------------------------------------------------------------------
//	Some definitions of DMA structures easier to use than those of LL

// STM32G0: DMAMUX1 have up to 12 channels, 4 request generator
// STM32L4: DMAMUX1 have 14 channels, DMAMUX2 have 8 channels, 4 request generator
// STM32H7: DMAMUX1 have 16 channels, DMAMUX2 have 8 channels, 8 request generator

typedef struct
{
	__IO uint32_t	CCR  [16] ;		// 0x000 Request line multiplexer channel x configuration register
	__IO uint32_t	res1 [16] ;		//       Reserved
	__IO uint32_t	CSR ;			// 0x080 Request line multiplexer interrupt channel status register
	__IO uint32_t	CFR ;			// 0x084 Request line multiplexer interrupt clear flag register
	__IO uint32_t	res2 [30] ;		//       Reserved
	__IO uint32_t	RGCR [8] ;		// 0x100 Request generator channel x configuration register
	__IO uint32_t	res3 [8] ;		//       Reserved
	__IO uint32_t	RGSR ;			// 0x140 DMA Request Generator Status Register
	__IO uint32_t	RGCFR ;			// 0x144 Request generator interrupt status register

} dmaMux_t ;

typedef struct
{
	__IO uint32_t 	CCR ;         	// DMA channel x configuration register
	__IO uint32_t 	CNDTR ;       	// DMA channel x number of data register
	__IO uint32_t 	CPAR ;        	// DMA channel x peripheral address register
	__IO uint32_t 	CMAR ;        	// DMA channel x memory address register
	__IO uint32_t 	res ;			// Not defined by ST files...

} dmaStream_t;

// This is the structure pointed to by DMA1 or DMA2
typedef struct
{
	__IO uint32_t	ISR ;			// DMA interrupt status register
	__IO uint32_t	IFCR ;			// DMA interrupt flag clear register
	dmaStream_t		stream [7] ;

} dma_t ;

//----------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

void		dmaEnableClock		(dma_t * pDma) ;
void		dmaSetIrq			(const dmaDesc_t * pDmaDesc) ;
void		dmaDisableIrq		(const dmaDesc_t * pDmaDesc) ;
void		dmaSetCallback		(const dmaDesc_t * pDmaDesc, dmaCallback_t fn, uintptr_t arg) ;
void		dmaSetCallbackRaw	(dma_t * pDma, uint32_t streamNum, dmaCallback_t fn, uintptr_t arg) ;

// Flags to use with dmaGetFlags() and dmaClearFlags()
#define DMA_FLAG_GIF		((uint32_t) 0x00000001u)
#define DMA_FLAG_TCIF		((uint32_t) 0x00000002u)
#define DMA_FLAG_HTIF		((uint32_t) 0x00000004u)
#define DMA_FLAG_TEIF		((uint32_t) 0x00000008u)
#define DMA_FLAG_ALLIF		((uint32_t) 0x0000000Fu)

uint32_t	dmaGetFlags		(dma_t * pDma, uint32_t streamNum) ;
void		dmaClearFlags	(dma_t * pDma, uint32_t streamNum, uint32_t flags) ;

// Get a pointer to a stream registers
__STATIC_INLINE dmaStream_t * dmaStreamAddress (dma_t * pDma, uint32_t streamNum)
{
	return & pDma->stream [streamNum] ;
}

// Set stream to reset state
__STATIC_INLINE void		dmaResetStream	(dma_t * pDma, uint32_t streamNum)
{
	dmaStream_t * pStream = dmaStreamAddress (pDma, streamNum) ;

	pStream->CCR	= 0 ;
	pStream->CNDTR	= 0 ;
	pStream->CPAR	= 0 ;
	pStream->CMAR	= 0 ;
	dmaClearFlags (pDma, streamNum, DMA_FLAG_ALLIF) ;
}

#ifdef __cplusplus
}
#endif

//----------------------------------------------------------------------
#endif	// DMABASIC_H_
