/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	extibasic.c		GPIO and software interrupts management

	When		Who	What
	02/09/18	ac	Creation
	22/01/20	ac	Some cosmetic changes
				ac	extiResetInterrupt() changed to extiResetLine()
	30/04/21	ac	Port to Go
					Add individual lines callback

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2021 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"

#include	"extibasic.h"

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Asynch Interrupt/Event Controller (EXTI) registers

typedef struct
{
	__IO	uint32_t RTSR ;				/*!< EXTI Rising Trigger Selection Register 1,        Address offset:   0x00 */
	__IO	uint32_t FTSR ;				/*!< EXTI Falling Trigger Selection Register 1,       Address offset:   0x04 */
	__IO	uint32_t SWIER ;			/*!< EXTI Software Interrupt event Register 1,        Address offset:   0x08 */
	__IO	uint32_t RPR ;				/*!< EXTI Rising Pending Register 1,                  Address offset:   0x0C */
	__IO	uint32_t FPR ;				/*!< EXTI Falling Pending Register 1,                 Address offset:   0x10 */
    		uint32_t res1 [3] ;			/*!< Reserved 1,                                                0x14 -- 0x1C */

} extiCfg_t ;

typedef struct
{
	__IO	uint32_t	IMR ;			// EXTI Interrupt Mask Register
	__IO	uint32_t	EMR ;			// EXTI Event Mask Register
			uint32_t	res1 [2] ;

} extiMr_t ;	// Interrupt/Event mask registers

typedef struct
{
			extiCfg_t	cfg [2] ;		// 0x00	Lines 0..31 and 32..63
			uint32_t	res1 [8] ;
	__IO	uint32_t	EXTICR[4] ;  	// 0x60 -- 0x6C */
			uint32_t	res2 [4] ;
			extiMr_t	mr [2] ;		// 0x80
} exti_t ;

//----------------------------------------------------------------------

static inline extiCfg_t *	getExtiCfg (uint32_t lineNumber)
{
	if (lineNumber < 32)
	{
		return (extiCfg_t *) & ((exti_t *) EXTI)->cfg [0] ;
	}
	else
	{
		return (extiCfg_t *) & ((exti_t *) EXTI)->cfg [1] ;
	}

}

//----------------------------------------------------------------------

static inline extiMr_t *	getExtiMr (uint32_t lineNumber)
{
	if (lineNumber < 32)
	{
		return (extiMr_t *) & ((exti_t *) EXTI)->mr [0] ;
	}
	else
	{
		return (extiMr_t *) & ((exti_t *) EXTI)->mr [1] ;
	}
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Get the interrupt request number from line number (Configurable lines)
//	The lines without associated interrupt returns 0 (Direct lines)

#if defined(STM32G0C1xx) || defined(STM32G0B1xx)
	#define EXTI_LINE_COUNT		37u
#elif defined(STM32G0B0xx)
	#define EXTI_LINE_COUNT		37u
#elif defined(STM32G081xx) || defined(STM32G071xx)
	#define EXTI_LINE_COUNT		34u
#elif defined(STM32G070xx)
	#define EXTI_LINE_COUNT		34u
#elif defined(STM32G041xx) || defined(STM32G031xx)
	#define EXTI_LINE_COUNT		32u
#else
	#define EXTI_LINE_COUNT		32u
#endif


static	IRQn_Type	lineToIrqn_ (uint32_t lineNumber)
{
	IRQn_Type num = 0 ;

	if (lineNumber < 16)
	{
		// IO port line
		if (lineNumber < 2)
		{
			num = EXTI0_1_IRQn ;
		}
		else if (lineNumber < 4)
		{
			num = EXTI2_3_IRQn ;
		}
		else
		{
			num = EXTI4_15_IRQn ;
		}
	}
	else
	{
		switch (lineNumber)
		{
			case 16:
//			case 34:
				num = 1 ;	// PVD_IRQn, PVD_VDDIO2_IRQn ;
				break ;

			case 17:
			case 18:
			case 20:
				#if defined COMP1
					num = ADC1_COMP_IRQn ;
				#endif
				break ;
/*
			case 19:
			case 21:
				num = RTC_TAMP_IRQn ;
				break ;

			case 22:
				num = 24 ;	// I2C2_IRQn, I2C2_3_IRQn
				break ;

			case 23:
				num = I2C1_IRQn ;
				break ;

			case 25:
				num = USART1_IRQn ;
				break ;

			case 26:
				num = 28;  	// USART2_IRQn, USART2_LPUART2_IRQn ;
				break ;

			case 27:
				num = CEC_IRQn ;
				break ;

			case 28:
				num = 29 ;	// USART3_4_LPUART1_IRQn, USART3_4_5_6_LPUART1_IRQn
				break ;

			case 32:
			case 33:
				num = 8; 	// UCPD1_2_IRQn, USB_UCPD1_2_IRQn ;
				break ;
*/
			default:
				break ;
		}
	}

	AA_ASSERT (num != 0) ;
	return num ;
}

//----------------------------------------------------------------------
//	Configure registers for interrupt or event

static extiError_t	extiSetRegisters_ (uint32_t lineNumber, uint32_t priority, uint32_t flags)
{
	extiCfg_t	* pCfg = getExtiCfg (lineNumber) ;
	extiMr_t	* pMr  = getExtiMr (lineNumber);
	uint32_t	pinMask = 1u << (lineNumber & 0x1F) ;

	pMr->IMR &= ~pinMask ;
	pMr->EMR &= ~pinMask ;

	// Configure event
	if ((flags & EXTI_FLAG_EVENT) != 0u)
	{
		pMr->EMR |= pinMask ;
	}

	// Configure trigger active edge
	pCfg->FTSR &= ~pinMask ;
	if ((flags & EXTI_FLAG_FALLING) != 0u)
	{
		pCfg->FTSR |= pinMask ;		// Falling
	}

	pCfg->RTSR &= ~pinMask;
	if ((flags & EXTI_FLAG_RISING) != 0u)
	{
		pCfg->RTSR |= pinMask ;		// Rising
	}

	// Add to NVIC if interrupt is requested
	if ((flags & EXTI_FLAG_INTR) != 0u)
	{
		IRQn_Type irqNum = lineToIrqn_ (lineNumber) ;

		pMr->IMR |= pinMask ;
		if (irqNum != 0)
		{
			NVIC_EnableIRQ   (irqNum) ;
			NVIC_SetPriority (irqNum, priority) ;
		}
	}
	return EXTI_ENONE ;
}

//----------------------------------------------------------------------
//	Initialize an interrupt or an event for a GPIO pin
//	Pin: 0..15
//	If the flag EXTI_FLAG_GPIO_NOCONF is not present, the pin is configured using pPinDesc
//	Allowed flags: EXTI_FLAG_INTR, EXTI_FLAG_EVENT, EXTI_FLAG_RISING, EXTI_FLAG_FALLING

extiError_t	extiInitGpio	(const gpioPinDesc_t * pPinDesc, uint32_t priority, uint32_t flags)
{
	uint32_t		reg ;

	if (pPinDesc->pin > 15u)
	{
		AA_ASSERT (0) ;
		return EXTI_EFAIL ;
	}

	// Set proper GPIO port in EXTICRn
	reg = EXTI->EXTICR [pPinDesc->pin >> 2u] ;
	reg &= ~(0x00FFu << (8u * (pPinDesc->pin & 0x03u))) ;
	reg |= (uint32_t) (((pPinDesc->port - (uint8_t) 'A') << (8u * (pPinDesc->pin & 0x03u)))) ;
	EXTI->EXTICR [pPinDesc->pin >> 2u] = reg ;

	// Configure pin
	if ((flags & EXTI_FLAG_GPIO_NOCONF) == 0)
	{
		gpioConfigurePin (pPinDesc) ;
	}

	return extiSetRegisters_ (pPinDesc->pin, priority, flags) ;
}

//----------------------------------------------------------------------
//	Initialize an interrupt or an event on a line 16..EXTI_LINE_COUNT-1
//	Allowed flags: EXTI_FLAG_INTR, EXTI_FLAG_EVENT, EXTI_FLAG_RISING, EXTI_FLAG_FALLING

extiError_t	extiInitLine		(uint32_t lineNumber, uint32_t priority, uint32_t flags)
{
	if (lineNumber < 16u  ||  lineNumber >= EXTI_LINE_COUNT)
	{
		AA_ASSERT (0) ;
		return EXTI_EFAIL ;
	}

	return extiSetRegisters_ (lineNumber, priority, flags) ;
}

//----------------------------------------------------------------------
//	Initialize an exti line 0..EXTI_LINE_COUNT-1 for software interrupt
//	This doesn't allows to configure a software event

extiError_t	extiInitSwi		(uint32_t lineNumber, uint32_t priority)
{
	if (lineNumber >= EXTI_LINE_COUNT)
	{
		AA_ASSERT (0) ;
		return EXTI_EFAIL ;
	}

	// Force trigger on rising edge
	return extiSetRegisters_ (lineNumber, priority, EXTI_FLAG_INTR | EXTI_FLAG_RISING) ;
}

//----------------------------------------------------------------------

void		extiResetLine	(uint32_t lineNumber)
{
	extiCfg_t	* pCfg ;
	extiMr_t	* pMr ;
	uint32_t	pinMask = 1u << (lineNumber & 0x1Fu) ;

	if (lineNumber < EXTI_LINE_COUNT)
	{
		pCfg = getExtiCfg (lineNumber) ;
		pMr  = getExtiMr (lineNumber);

		pMr->IMR	&= ~pinMask ;
		pMr->EMR	&= ~pinMask ;
		pCfg->FTSR	&= ~pinMask ;
		pCfg->RTSR	&= ~pinMask ;
	}

	if (lineNumber < 16)
	{
		// Clear port in EXTICRn. In reality this selects the port A
		EXTI->EXTICR [lineNumber >> 2u] &= ~(0x00FFu << (8u * (lineNumber & 0x03u))) ;
	}
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
// Interrupt handlers

static	void	extiIrq_0_31_Handler (uint32_t line, extiCallaback_t pCb)
{
	uint32_t	flags = 0 ;
	uint32_t	mask = 1 << line ;

	// Rising event
	if ((EXTI->RPR1 & mask) != 0u)
	{
		// Clear bit
		EXTI->RPR1 = mask ;
		flags |= EXTI_FLAG_RISING ;
	}

	// Falling event
	if ((EXTI->FPR1 & mask) != 0u)
	{
		// Clear bit
		EXTI->FPR1 = mask ;
		flags |= EXTI_FLAG_FALLING ;
	}

	// Call the user callback
	if (flags != 0)
	{
		(* pCb) (flags)  ;
	}
}

//----------------------------------------------------------------------

#if (WITH_EXTI_0_1 == 1)
void EXTI0_1_IRQHandler (void) ;
void EXTI0_1_IRQHandler (void)
{
	uint32_t	reg ;

	aaIntEnter () ;
	reg = EXTI->RPR1 | EXTI->FPR1 ;
	if ((reg & EXTI_RPR1_RPIF0) != 0)
	{
		extiIrq_0_31_Handler (0, extiCallback0) ;
	}
	if ((reg & EXTI_RPR1_RPIF1) != 0)
	{
		extiIrq_0_31_Handler (1, extiCallback1) ;
	}
	aaIntExit () ;
}
#endif

//----------------------------------------------------------------------

#if (WITH_EXTI_2_3 == 1)
void EXTI2_3_IRQHandler (void) ;
void EXTI2_3_IRQHandler (void)
{
	uint32_t	reg ;

	aaIntEnter () ;
	reg = EXTI->RPR1 | EXTI->FPR1 ;
	if ((reg & EXTI_RPR1_RPIF2) != 0)
	{
		extiIrq_0_31_Handler (2, extiCallback2) ;
	}
	if ((reg & EXTI_RPR1_RPIF3) != 0)
	{
		extiIrq_0_31_Handler (3, extiCallback3) ;
	}
	aaIntExit () ;
}
#endif

//----------------------------------------------------------------------

#if (WITH_EXTI_4_15 == 1)
void EXTI4_15_IRQHandler (void) ;
void EXTI4_15_IRQHandler (void)
{
	uint32_t	reg ;

	aaIntEnter () ;

	reg = (EXTI->RPR1 | EXTI->FPR1) >> 4 ;
	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (4, extiCallback4) ;
	}
	reg >>= 1 ;

	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (5, extiCallback5) ;
	}
	reg >>= 1 ;

	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (6, extiCallback6) ;
	}
	reg >>= 1 ;

	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (7, extiCallback7) ;
	}
	reg >>= 1 ;

	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (8, extiCallback8) ;
	}
	reg >>= 1 ;

	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (9, extiCallback9) ;
	}
	reg >>= 1 ;

	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (10, extiCallback10) ;
	}
	reg >>= 1 ;

	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (11, extiCallback11) ;
	}
	reg >>= 1 ;

	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (12, extiCallback12) ;
	}
	reg >>= 1 ;

	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (13, extiCallback13) ;
	}
	reg >>= 1 ;

	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (14, extiCallback14) ;
	}
	reg >>= 1 ;

	if ((reg & 1) != 0)
	{
		extiIrq_0_31_Handler (15, extiCallback15) ;
	}

	aaIntExit () ;
}
#endif

//----------------------------------------------------------------------
// EXTI 16 and 34

#if (WITH_EXTI_PVD == 1)
#if (defined EXTI_RPR2_RPIF34)
	void PVD_VDDIO2_IRQHandler (void) ;
	void PVD_VDDIO2_IRQHandler (void)
#else
	void PVD_IRQHandler (void) ;
	void PVD_IRQHandler (void)
#endif
{
	uint32_t	reg ;
	(void) reg ;

	aaIntEnter () ;

	#if (defined EXTI_RPR1_RPIF16)
		reg = EXTI->RPR1 | EXTI->FPR1 ;
		if ((reg & EXTI_RPR1_RPIF16) != 0u)
		{
			// PVD
			extiIrq_0_31_Handler (16, extiCallback16) ;
		}
	#endif

	#if (defined EXTI_RPR2_RPIF34)
		reg = EXTI->RPR2 | EXTI->FPR2 ;
		if ((reg & EXTI_RPR2_RPIF34) != 0u)
		{
			// PVM VDDIO
			extiIrq_0_31_Handler (34, extiDefaultCallback) ;
		}
	#endif

	aaIntExit () ;
}
#endif

//----------------------------------------------------------------------
// EXTI 17, 18, 20

#if (WITH_EXTI_COMP_1_3 == 1)
#if (defined COMP1)
	void ADC1_COMP_IRQHandler (void) ;
	void ADC1_COMP_IRQHandler (void)
#else
	void ADC1_IRQHandler (void) ;
	void ADC1_IRQHandler (void)
#endif
{
	uint32_t	reg ;

	(void) reg ;
	aaIntEnter () ;

	#if (defined COMP1)
	reg = EXTI->RPR1 | EXTI->FPR1 ;
		if ((reg & EXTI_RPR1_RPIF17) != 0u)
		{
			// COMP1
			extiIrq_0_31_Handler (17, extiCallback17) ;
		}
	#endif
	#if (defined COMP2)
		if ((reg & EXTI_RPR1_RPIF18) != 0u)
		{
			// COMP2
			extiIrq_0_31_Handler (18, extiCallback18) ;
		}
	#endif
	#if (defined COMP3)
		if ((reg & EXTI_RPR1_RPIF20) != 0u)
		{
			// COMP3
			extiIrq_0_31_Handler (20, extiCallback20) ;
		}
	#endif

	aaIntExit () ;
}
#endif 	// WITH_EXTI_COMP_1_3

//----------------------------------------------------------------------
//	Default callback for all lines

void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback0		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback1		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback2		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback3		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback4		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback5		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback6		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback7		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback8		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback9		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback10		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback11		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback12		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback13		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback14		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback15		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback16		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback17		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback18		(uint32_t flags) ;
void __attribute__ ((weak, alias ("extiDefaultCallback")))	extiCallback20		(uint32_t flags) ;

static	void extiDefaultCallback (uint32_t flags)
{
	(void) flags ;
}

//----------------------------------------------------------------------
