/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	gpiobasic.c	Some routine to manage basic GPIO and RCC features

	When		Who	What
	17/03/18	ac	Creation for STM32F4xx
	20/03/20	ac	port to STM32L4+ (manage PWR_CR2_IOSV)
	29/04/20	ac	Add GPIOx_ASCR for L47x and L48xx
	14/01/20	ac	Port to STM32G071
	04/03/21	ac	Add rccGetXxxClockFreq()functions

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2021 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"
#include	"gpiobasic.h"

#include	"stm32g0xx_ll_bus.h"
#include	"stm32g0xx_ll_rcc.h"

//----------------------------------------------------------------------

#define	OFFSET_MASK		0x0000FFFFu

//----------------------------------------------------------------------
//	Some utility functions to get clocks frequencies

uint32_t	rccGetSystemClockFreq	(void)
{
	// SYSTEM clock frequency
	return SystemCoreClock ;
}

uint32_t	rccGetHCLKClockFreq		(void)
{
	// HCLK clock frequency
	return __LL_RCC_CALC_HCLK_FREQ (SystemCoreClock, LL_RCC_GetAHBPrescaler());
}

uint32_t	rccGetPCLK1ClockFreq	(void)
{
	// PCLK1 clock frequency
	return __LL_RCC_CALC_PCLK1_FREQ	(rccGetHCLKClockFreq (), LL_RCC_GetAPB1Prescaler());
}

#if defined __LL_RCC_CALC_PCLK2_FREQ
uint32_t	rccGetPCLK2ClockFreq	(void)
{
	// PCLK2 clock frequency
	return __LL_RCC_CALC_PCLK2_FREQ (rccGetHCLKClockFreq (), LL_RCC_GetAPB2Prescaler());
}
#endif

//----------------------------------------------------------------------
// Peripheral clock enable
// Example:
// rccEnableClock (gpioPortFromName (pPinDesc->port), 1u << (pPinDesc->port - AA_GPIO_PORTA)) ;
// rccResetPeriph (SPI2, RCC_APBENR1_SPI2EN) ;
// rccResetPeriph (SPI1, RCC_APBENR2_SPI1EN) ;

// For timers enableBitMask is ignored

void	rccEnableClock (void * deviceAddress, uint32_t enableBitMask)
{
	// Only one AHB and one APB on G0
	// BUT: the syntax use APB1 and APB2 !!!! Thanks...
	if (((uintptr_t) deviceAddress & ~OFFSET_MASK) == AHBPERIPH_BASE)
	{
		/*	DMA1
			DMA2
			FLASH
			CRC
			AES
			RNG
		*/
		AA_ASSERT (enableBitMask != 0) ;
		LL_AHB1_GRP1_EnableClock (enableBitMask) ;
		return ;
	}

	if (((uintptr_t) deviceAddress ) >= IOPORT_BASE)
	{
		AA_ASSERT (enableBitMask != 0) ;
		LL_IOP_GRP1_EnableClock (enableBitMask) ;
		return ;
	}

	switch ((uintptr_t) deviceAddress)
	{
			//------------------ Timers in register APBRSTR1
#if (defined LPTIM1_BASE)
			case LPTIM1_BASE:
				LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_LPTIM1) ;
				break ;
#endif

#if (defined LPTIM2_BASE)
			case LPTIM2_BASE:
				LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_LPTIM2) ;
				break ;
#endif

#if (defined TIM2_BASE)
			case TIM2_BASE:
				LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_TIM2) ;
				break ;
#endif

#if (defined TIM3_BASE)
			case TIM3_BASE:
				LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_TIM3) ;
				break ;
#endif

#if (defined TIM4_BASE)
			case TIM4_BASE:
				LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_TIM4) ;
				break ;
#endif

#if (defined TIM6_BASE)
			case TIM6_BASE:
				LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_TIM6) ;
				break ;
#endif

#if (defined TIM7_BASE)
			case TIM7_BASE:
				LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_TIM7) ;
				break ;
#endif

		//------------------ Timers in register APBRSTR2
		case TIM1_BASE:
			LL_APB2_GRP1_EnableClock (LL_APB2_GRP1_PERIPH_TIM1) ;
			break ;

		case TIM14_BASE:
			LL_APB2_GRP1_EnableClock (LL_APB2_GRP1_PERIPH_TIM14) ;
			break ;

#if (defined TIM15_BASE)
		case TIM15_BASE:
			LL_APB2_GRP1_EnableClock (LL_APB2_GRP1_PERIPH_TIM15) ;
			break ;
#endif

		case TIM16_BASE:
			LL_APB2_GRP1_EnableClock (LL_APB2_GRP1_PERIPH_TIM16) ;
			break ;

		case TIM17_BASE:
			LL_APB2_GRP1_EnableClock (LL_APB2_GRP1_PERIPH_TIM17) ;
			break ;

		//------------------ USART LPUART
		case USART1_BASE:
			LL_APB2_GRP1_EnableClock (LL_APB2_GRP1_PERIPH_USART1) ;
			break ;

		case USART2_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_USART2) ;
			break ;

#if (defined USART3_BASE)
		case USART3_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_USART3) ;
			break ;
#endif

#if (defined USART4_BASE)
		case USART4_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_USART4) ;
			break ;
#endif

#if (defined USART5_BASE)
		case USART5_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_USART5) ;
			break ;
#endif

#if (defined USART6_BASE)
		case USART6_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_USART6) ;
			break ;
#endif

#if (defined LPUART1_BASE)
		case LPUART1_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_LPUART1) ;
			break ;
#endif

#if (defined LPUART2_BASE)
		case LPUART2_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_LPUART2) ;
			break ;
#endif

		//------------------ APB2

		case ADC1_BASE:
			LL_APB2_GRP1_EnableClock (LL_APB2_GRP1_PERIPH_ADC) ;
			break ;

		case SPI1_BASE:
			LL_APB2_GRP1_EnableClock (LL_APB2_GRP1_PERIPH_SPI1) ;
			break ;

		case SYSCFG_BASE:
			LL_APB2_GRP1_EnableClock (LL_APB2_GRP1_PERIPH_SYSCFG) ;
			break ;

		//------------------ APB1
		case RTC_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_RTC) ;
			break ;

		case WWDG_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_WWDG) ;
			break ;

#if defined (FDCAN1) || defined (FDCAN2)
		case FDCAN1_BASE:
		case FDCAN2_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_FDCAN) ;
			break ;
#endif

#if defined (USB_DRD_FS)
		case USB_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_USB) ;
			break ;
#endif

		case SPI2_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_SPI2) ;
			break ;

#if defined (SPI3_BASE)
		case SPI3_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_SPI3) ;
			break ;
#endif

#if defined (CRS_BASE)
		case CRS_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_CRS) ;
			break ;
#endif

		case I2C1_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_I2C1) ;
			break ;

		case I2C2_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_I2C2) ;
			break ;

#if defined (I2C3_BASE)
		case I2C3_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_I2C3) ;
			break ;
#endif

#if defined (DAC1_BASE)
		case DAC1_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_DAC1) ;
			break ;
#endif

		case PWR_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_PWR) ;
			break ;

		case DBG_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_DBGMCU) ;
			break ;

		//------------------ not managed
		default:
			AA_ASSERT (0) ;
			break ;
	}
}

//----------------------------------------------------------------------

void	rccDisableClock (void * deviceAddress, uint32_t enableBitMask)
{
	// Only one AHB and one APB on G0
	// BUT: the syntax use APB1 and APB2 !!!! Thanks...
	if (((uintptr_t) deviceAddress & ~OFFSET_MASK) == AHBPERIPH_BASE)
	{
		/*	DMA1
			DMA2
			FLASH
			CRC
			AES
			RNG
		*/
		AA_ASSERT (enableBitMask != 0) ;
		LL_AHB1_GRP1_DisableClock (enableBitMask) ;
		return ;
	}

	if (((uintptr_t) deviceAddress ) >= IOPORT_BASE)
	{
		AA_ASSERT (enableBitMask != 0) ;
		LL_IOP_GRP1_DisableClock (enableBitMask) ;
		return ;
	}

	switch ((uintptr_t) deviceAddress)
	{
		//------------------ Timers in register APBRSTR1
#if (defined LPTIM1_BASE)
		case LPTIM1_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_LPTIM1) ;
			break ;
#endif

#if (defined LPTIM2_BASE)
		case LPTIM2_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_LPTIM2) ;
			break ;
#endif

#if (defined TIM2_BASE)
		case TIM2_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_TIM2) ;
			break ;
#endif

#if (defined TIM3_BASE)
		case TIM3_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_TIM3) ;
			break ;
#endif

#if (defined TIM4_BASE)
		case TIM4_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_TIM4) ;
			break ;
#endif

#if (defined TIM6_BASE)
		case TIM6_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_TIM6) ;
			break ;
#endif

#if (defined TIM7_BASE)
		case TIM7_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_TIM7) ;
			break ;
#endif
		//------------------ Timers in register APBRSTR2
		case TIM1_BASE:
			LL_APB2_GRP1_DisableClock (LL_APB2_GRP1_PERIPH_TIM1) ;
			break ;

		case TIM14_BASE:
			LL_APB2_GRP1_DisableClock (LL_APB2_GRP1_PERIPH_TIM14) ;
			break ;

#if (defined TIM15_BASE)
		case TIM15_BASE:
			LL_APB2_GRP1_DisableClock (LL_APB2_GRP1_PERIPH_TIM15) ;
			break ;
#endif

		case TIM16_BASE:
			LL_APB2_GRP1_DisableClock (LL_APB2_GRP1_PERIPH_TIM16) ;
			break ;

		case TIM17_BASE:
			LL_APB2_GRP1_DisableClock (LL_APB2_GRP1_PERIPH_TIM17) ;
			break ;

		//------------------ USART LPUART
		case USART1_BASE:
			LL_APB2_GRP1_DisableClock (LL_APB2_GRP1_PERIPH_USART1) ;
			break ;

		case USART2_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_USART2) ;
			break ;

#if (defined USART3_BASE)
		case USART3_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_USART3) ;
			break ;
#endif

#if (defined USART4_BASE)
		case USART4_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_USART4) ;
			break ;
#endif

#if (defined USART5_BASE)
		case USART5_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_USART5) ;
			break ;
#endif

#if (defined USART6_BASE)
		case USART6_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_USART6) ;
			break ;
#endif

#if (defined LPUART1_BASE)
		case LPUART1_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_LPUART1) ;
			break ;
#endif

#if (defined LPUART2_BASE)
		case LPUART2_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_LPUART2) ;
			break ;
#endif

		//------------------ APB2

		case ADC1_BASE:
			LL_APB2_GRP1_DisableClock (LL_APB2_GRP1_PERIPH_ADC) ;
			break ;

		case SPI1_BASE:
			LL_APB2_GRP1_DisableClock (LL_APB2_GRP1_PERIPH_SPI1) ;
			break ;

		case SYSCFG_BASE:
			LL_APB2_GRP1_DisableClock (LL_APB2_GRP1_PERIPH_SYSCFG) ;
			break ;

		//------------------ APB1
		case RTC_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_RTC) ;
			break ;

		case WWDG_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_WWDG) ;
			break ;

#if defined (FDCAN1) || defined (FDCAN2)
		case FDCAN1_BASE:
		case FDCAN2_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_FDCAN) ;
			break ;
#endif

#if defined (USB_DRD_FS)
		case USB_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_USB) ;
			break ;
#endif

		case SPI2_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_SPI2) ;
			break ;

#if defined (SPI3)
		case SPI3_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_SPI3) ;
			break ;
#endif

#if defined (CRS)
		case CRS_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_CRS) ;
			break ;
#endif

		case I2C1_BASE:
			LL_APB1_GRP1_EnableClock (LL_APB1_GRP1_PERIPH_I2C1) ;
			break ;

		case I2C2_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_I2C2) ;
			break ;

#if defined (I2C3_BASE)
		case I2C3_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_I2C3) ;
			break ;
#endif

#if defined (DAC1_BASE)
		case DAC1_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_DAC1) ;
			break ;
#endif

		case PWR_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_PWR) ;
			break ;

		case DBG_BASE:
			LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_DBGMCU) ;
			break ;

		//------------------ not managed
		default:
			AA_ASSERT (0) ;
			break ;
	}
}

//----------------------------------------------------------------------
// Example:
// rccResetPeriph (SPI2, RCC_APBENR1_SPI2EN) ;
// rccResetPeriph (SPI1, RCC_APBENR2_SPI1EN) ;

void	rccResetPeriph (void * deviceAddress, uint32_t enableBitMask)
{
	// Only one AHB and one APB on G0
	// BUT: the syntax use APB1 and APB2 !!!! Thanks...

	if (((uintptr_t) deviceAddress & ~OFFSET_MASK) == AHBPERIPH_BASE)
	{
		/*	DMA1
			DMA2
			FLASH
			CRC
			AES
			RNG
		*/
		AA_ASSERT (enableBitMask != 0) ;
		LL_AHB1_GRP1_ForceReset   (enableBitMask) ;
		LL_AHB1_GRP1_ReleaseReset (enableBitMask) ;
		return ;
	}

	if (((uintptr_t) deviceAddress ) >= IOPORT_BASE)
	{
		// IO ports reset: GPIOA, GPIOB...
		AA_ASSERT (enableBitMask != 0) ;
		LL_IOP_GRP1_ForceReset   (enableBitMask) ;
		LL_IOP_GRP1_ReleaseReset (enableBitMask) ;
		return ;
	}

	switch ((uintptr_t) deviceAddress)
	{
		//------------------ Timers in register APBRSTR1
#if (defined LPTIM1_BASE)
		case LPTIM1_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_LPTIM1) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_LPTIM1) ;
			break ;
#endif

#if (defined LPTIM2_BASE)
		case LPTIM2_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_LPTIM2) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_LPTIM2) ;
			break ;
#endif

#if (defined TIM2_BASE)
		case TIM2_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_TIM2) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_TIM2) ;
			break ;
#endif

#if (defined TIM3_BASE)
		case TIM3_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_TIM3) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_TIM3) ;
			break ;
#endif

#if (defined TIM4_BASE)
		case TIM4_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_TIM4) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_TIM4) ;
			break ;
#endif

#if (defined TIM6_BASE)
		case TIM6_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_TIM6) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_TIM6) ;
			break ;
#endif

#if (defined TIM7_BASE)
		case TIM7_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_TIM7) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_TIM7) ;
			break ;
#endif
	//------------------ Timers in register APBRSTR2
		case TIM1_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB2_GRP1_PERIPH_TIM1) ;
			LL_APB2_GRP1_ReleaseReset (LL_APB2_GRP1_PERIPH_TIM1) ;
			break ;

		case TIM14_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB2_GRP1_PERIPH_TIM14) ;
			LL_APB2_GRP1_ReleaseReset (LL_APB2_GRP1_PERIPH_TIM14) ;
			break ;

#if (defined TIM15_BASE)
		case TIM15_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB2_GRP1_PERIPH_TIM15) ;
			LL_APB2_GRP1_ReleaseReset (LL_APB2_GRP1_PERIPH_TIM15) ;
			break ;
#endif

		case TIM16_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB2_GRP1_PERIPH_TIM16) ;
			LL_APB2_GRP1_ReleaseReset (LL_APB2_GRP1_PERIPH_TIM16) ;
			break ;

		case TIM17_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB2_GRP1_PERIPH_TIM17) ;
			LL_APB2_GRP1_ReleaseReset (LL_APB2_GRP1_PERIPH_TIM17) ;
			break ;

			//------------------ USART LPUART
		case USART1_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB2_GRP1_PERIPH_USART1) ;
			LL_APB2_GRP1_ReleaseReset (LL_APB2_GRP1_PERIPH_USART1) ;
			break ;

		case USART2_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_USART2) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_USART2) ;
			break ;

#if (defined USART3_BASE)
		case USART3_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_USART3) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_USART3) ;
			break ;
#endif

#if (defined USART4_BASE)
		case USART4_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_USART4) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_USART4) ;
			break ;
#endif

#if (defined USART5_BASE)
		case USART5_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_USART5) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_USART5) ;
			break ;
#endif

#if (defined USART6_BASE)
		case USART6_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_USART6) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_USART6) ;
			break ;
#endif

#if (defined LPUART1_BASE)
		case LPUART1_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_LPUART1) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_LPUART1) ;
			break ;
#endif

#if (defined LPUART2_BASE)
		case LPUART2_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_LPUART2) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_LPUART2) ;
			break ;
#endif

		//------------------ APB2

		case ADC1_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB2_GRP1_PERIPH_ADC) ;
			LL_APB2_GRP1_ReleaseReset (LL_APB2_GRP1_PERIPH_ADC) ;
			break ;

		case SPI1_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB2_GRP1_PERIPH_SPI1) ;
			LL_APB2_GRP1_ReleaseReset (LL_APB2_GRP1_PERIPH_SPI1) ;
			break ;

		case SYSCFG_BASE:
			LL_APB2_GRP1_ForceReset   (LL_APB2_GRP1_PERIPH_SYSCFG) ;
			LL_APB2_GRP1_ReleaseReset (LL_APB2_GRP1_PERIPH_SYSCFG) ;
			break ;

		//------------------ APB1
		case RTC_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_RTC) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_RTC) ;
			break ;

		case WWDG_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_WWDG) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_WWDG) ;
			break ;

#if defined (FDCAN1) || defined (FDCAN2)
		case FDCAN1_BASE:
		case FDCAN2_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_FDCAN) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_FDCAN) ;
			break ;
#endif

#if defined (USB_DRD_FS)
		case USB_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_USB) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_USB) ;
			break ;
#endif

		case SPI2_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_SPI2) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_SPI2) ;
			break ;

#if defined (SPI3_BASE)
		case SPI3_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_SPI3) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_SPI3) ;
			break ;
#endif

#if defined (CRS_BASE)
		case CRS_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_CRS) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_CRS) ;
			break ;
#endif

		case I2C1_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_I2C1) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_I2C1) ;
			break ;

		case I2C2_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_I2C2) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_I2C2) ;
			break ;

#if defined (I2C3_BASE)
		case I2C3_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_I2C3) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_I2C3) ;
			break ;
#endif

#if defined (DAC1_BASE)
		case DAC1_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_DAC1) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_DAC1) ;
			break ;
#endif

		case PWR_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_PWR) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_PWR) ;
			break ;

		case DBG_BASE:
			LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_DBGMCU) ;
			LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_DBGMCU) ;
			break ;

		//------------------ not managed
		default:
			AA_ASSERT (0) ;
			break ;
	}
}

//----------------------------------------------------------------------
//	All devices except: timers, GPIO, I2S

// For timer use rccGetTimerClockFrequency()
// For I2S use  LL_RCC_GetI2SClockSource (LL_RCC_I2Sx_CLKSOURCE)

uint32_t	rccGetClockFrequency (void * deviceAddress)
{
	uint32_t		devClock ;

	switch ((uintptr_t) deviceAddress)
	{
		case SPI1_BASE:
		case SPI2_BASE:
#if (! defined LL_RCC_I2C2_CLKSOURCE)
		case I2C2_BASE:
#endif
#if (! defined RCC_CCIPR_USART2SEL)
		case USART2_BASE:
#endif
#if (! defined RCC_CCIPR_USART3SEL  && defined USART3_BASE)
		case USART3_BASE:
#endif
#if (defined USART4_BASE)
		case USART4_BASE:
#endif
#if (defined USART5_BASE)
		case USART5_BASE:
#endif
#if (defined USART6_BASE)
		case USART6_BASE:
#endif
#if (defined DAC_BASE)
		case DAC_BASE:
#endif
			// These devices are on on APB clock
			devClock = rccGetPCLK1ClockFreq () ;
			break ;

		case I2C1_BASE:
			devClock = LL_RCC_GetI2CClockFreq (LL_RCC_I2C1_CLKSOURCE) ;
			break ;

#if (defined LL_RCC_I2C2_CLKSOURCE)
		case I2C2_BASE:
			devClock = LL_RCC_GetI2CClockFreq (LL_RCC_I2C2_CLKSOURCE) ;
			break ;
#endif

		case USART1_BASE:
			devClock = LL_RCC_GetUSARTClockFreq (LL_RCC_USART1_CLKSOURCE) ;
			break ;

#if (defined RCC_CCIPR_USART2SEL)
		case USART2_BASE:
			devClock = LL_RCC_GetUSARTClockFreq (LL_RCC_USART2_CLKSOURCE) ;
			break ;
#endif

#if (defined RCC_CCIPR_USART3SEL)
		case USART3_BASE:
			devClock = LL_RCC_GetUSARTClockFreq (LL_RCC_USART3_CLKSOURCE) ;
			break ;
#endif

#if (defined LPUART1_BASE)
		case LPUART1_BASE:
			devClock = LL_RCC_GetLPUARTClockFreq (LL_RCC_LPUART1_CLKSOURCE) ;
			break ;
#endif

#if (defined LPUART2)
		case LPUART2_BASE:
			devClock = LL_RCC_GetLPUARTClockFreq (LL_RCC_LPUART2_CLKSOURCE) ;
			break ;
#endif

		case ADC1_BASE:
			devClock = LL_RCC_GetADCClockFreq (LL_RCC_ADC_CLKSOURCE) ;
			break ;

#if (defined (FDCAN1)  ||  defined (FDCAN2))
		case FDCAN1_BASE:
		case FDCAN2_BASE:
			devClock = LL_RCC_GetFDCANClockFreq (LL_RCC_FDCAN_CLKSOURCE) ;
			break ;
#endif

#if (defined CEC)
		case CEC_BASE:
			devClock = LL_RCC_GetCECClockFreq (LL_RCC_CEC_CLKSOURCE) ;
			break ;
#endif

#if (defined RNG)
		case RNG_BASE:
			devClock = LL_RCC_GetRNGClockFreq (LL_RCC_RNG_CLKSOURCE) ;
			break ;
#endif

		default:
			devClock = 0u ;
			break ;
	}

	AA_ASSERT (devClock != 0u) ;	// Clock source not started
	return devClock ;
}

//----------------------------------------------------------------------
//	Get TIMER and LPTIMER clock frequency
//	Timer clock depends on APB bus prescaler value
//	Timer clock may be APBx peripherals frequency x2

uint32_t	rccGetTimerClockFrequency (void * deviceAddress)
{
	uint32_t	prescaler ;
	uint32_t	timClock ;
	
	(void) deviceAddress ;	// Some MCUs have all timers on PCLK1

#if (defined LPTIM1)
	if (deviceAddress == LPTIM1)
	{
		timClock = LL_RCC_GetLPTIMClockFreq (LL_RCC_LPTIM1_CLKSOURCE) ;
	}
	else
#endif

#if (defined LPTIM2)
	if (deviceAddress == LPTIM2)
	{
		timClock = LL_RCC_GetLPTIMClockFreq (LL_RCC_LPTIM2_CLKSOURCE) ;
	}
	else
#endif

#if (defined RCC_CCIPR_TIM1SEL)
	if (deviceAddress == TIM1)
	{
		timClock = LL_RCC_GetTIMClockFreq (LL_RCC_TIM1_CLKSOURCE) ;
	}
	else
#endif

#if (defined RCC_CCIPR_TIM15SEL)
	if (deviceAddress == TIM15)
	{
		timClock = LL_RCC_GetTIMClockFreq (LL_RCC_TIM15_CLKSOURCE) ;
	}
	else
#endif

	{
		// All other timers are on PCLK1
		timClock = rccGetPCLK1ClockFreq () ;		// timClock is PCLK1
		prescaler = LL_RCC_GetAPB1Prescaler () ;
		if (prescaler != 0u)
		{
			timClock *= 2u ;
		}
	}

	AA_ASSERT (timClock != 0u) ;	// Clock source not started

	return timClock ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	GPIO handling
//----------------------------------------------------------------------
// Configure one pin using its pin descriptor
// Uses a 4-byte descriptor instead of 24 bytes in the stack with LL

void	gpioConfigureAndSetPin_ (const gpioPinDesc_t * pPinDesc, uint32_t value)
{
	GPIO_TypeDef		* pGpio ;
	uint32_t			temp ;

	if (pPinDesc->port == 0u || pPinDesc->port == '0')
	{
		// Pin not used
		return ;
	}

	//	On STM32L4+ the pins PG[15:2] are powered by VDDIO2, and need to be connected before use...
	#if (defined STM32L4  &&  defined(PWR_CR2_IOSV))
		if ((pPinDesc->port == 'G')  &&  (pPinDesc->pin > 1u))
		{
			volatile uint32_t	reg ;

			RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN ;
			reg = RCC->APB1ENR1 ;	// Delay after an RCC peripheral clock enabling
			(void) reg ;

			PWR->CR2 |= PWR_CR2_IOSV ;
		}
	#endif

	pGpio = gpioPortFromName (pPinDesc->port) ;

	// Enable port clock
	rccEnableClock (pGpio, 1u << (pPinDesc->port - AA_GPIO_PORTA)) ;

	// In case of Alternate function mode selection
	if ((pPinDesc->flags & AA_GPIO_MODE_MASK) == AA_GPIO_MODE_ALTERNATE)
	{
		// Check the Alternate function parameter: 0x0..0xF
		AA_ASSERT ((pPinDesc->af & 0xF0u) == 0u) ;
	}
	else
	{
		// AF must be 0 if not AA_GPIO_MODE_ALTERNATE
		AA_ASSERT ((pPinDesc->af & 0x0Fu) == 0u) ;
	}

	// In case of Output or Alternate function mode selection
	if ((pPinDesc->flags & AA_GPIO_MODE_MASK) == AA_GPIO_MODE_ALTERNATE ||
		(pPinDesc->flags & AA_GPIO_MODE_MASK) == AA_GPIO_MODE_OUTPUT)
	{
		// Configure the IO Speed
		temp = pGpio->OSPEEDR ;
		temp &= ~(GPIO_OSPEEDR_OSPEED0 << (pPinDesc->pin * 2u)) ;
		temp |= (((pPinDesc->flags & AA_GPIO_SPEED_MASK) >> 2u) << (pPinDesc->pin * 2u)) ;
		pGpio->OSPEEDR = temp ;

		// Configure the IO Output Type
		temp = pGpio->OTYPER ;
		temp &= ~(GPIO_OTYPER_OT0 << pPinDesc->pin) ;
		temp |= (((pPinDesc->flags & AA_GPIO_OPEN_DRAIN) >> 4u) << pPinDesc->pin) ;
		pGpio->OTYPER = temp ;
	}

	// Activate the Pull-up or Pull down resistor for the current IO
	temp = pGpio->PUPDR ;
	temp &= ~(GPIO_PUPDR_PUPD0 << (pPinDesc->pin * 2u)) ;
	temp |= (((pPinDesc->flags >> 6u) & GPIO_PUPDR_PUPD0) << (pPinDesc->pin * 2u)) ;
	pGpio->PUPDR = temp ;

	// Configure Alternate function mapped with the current IO
	temp = pGpio->AFR[pPinDesc->pin >> 3u] ;
	temp &= ~((uint32_t) 0x0Fu        << ((pPinDesc->pin & 0x07u) * 4u)) ;
	temp |=  ((uint32_t) pPinDesc->af << ((pPinDesc->pin & 0x07u) * 4u)) ;
	pGpio->AFR[pPinDesc->pin >> 3u] = temp ;

	if (value != 0u)
	{
		// Set output value before direction mode
		if ((value & 1u) == 0u)
		{
			gpioPortPinReset (pGpio, 1u << pPinDesc->pin) ;
		}
		else
		{
			gpioPortPinSet (pGpio, 1u << pPinDesc->pin) ;
		}
	}

	// Configure IO Direction mode (Input, Output, Alternate or Analog)
	temp = pGpio->MODER ;
	temp &= ~(GPIO_MODER_MODE0 << (pPinDesc->pin * 2u)) ;
	temp |= ((pPinDesc->flags & AA_GPIO_MODE_MASK) << (pPinDesc->pin * 2u)) ;
	pGpio->MODER = temp ;

#if defined(GPIO_ASCR_ASC0)
	// Special case for L47x and L48xx : enable analog switch for ADC
	if (pPinDesc->flags & AA_GPIO_MODE_ADCSWITCH)
	{
		pGpio->ASCR |= 1 << pPinDesc->pin ;
	}
#endif
}

//----------------------------------------------------------------------
//	Set the value of the pin using its descriptor
//	Atomic

void	gpioPinSet		(const gpioPinDesc_t * pPinDesc, uint32_t bValue)
{
	GPIO_TypeDef	* pPort ;
	uint32_t		pinMask ;

	AA_ASSERT (pPinDesc->port != 0u  &&  pPinDesc->port != '0') ;

	pPort = gpioPortFromName (pPinDesc->port) ;
	pinMask = 1u << pPinDesc->pin ;

	if (bValue == 0u)
	{
		#if (BSP_HAS_BRR != 0)
			pPort->BRR = pinMask ;			// Reset bit
		#else
			pPort->BSRR = pinMask << 16u ;	// Reset bit
		#endif
	}
	else
	{
		pPort->BSRR = pinMask ;				// Set bit
	}
}

//----------------------------------------------------------------------
//	Get the value of the pin using its descriptor

uint32_t	gpioPinGet		(const gpioPinDesc_t * pPinDesc)
{
	GPIO_TypeDef	* pPort ;

	AA_ASSERT (pPinDesc->port != 0u  &&  pPinDesc->port != '0') ;

	pPort = gpioPortFromName (pPinDesc->port) ;
	return (pPort->IDR & (1u << pPinDesc->pin)) == 0u ? 0u : 1u ;
}

//----------------------------------------------------------------------
//	Toggle the value of the pin using its descriptor (atomic)

void	gpioPinToggle		(const gpioPinDesc_t * pPinDesc)
{
	GPIO_TypeDef	* pPort ;
	uint32_t		pinMask ;
	uint32_t		odr ;

	AA_ASSERT (pPinDesc->port != 0u  &&  pPinDesc->port != '0') ;

	pPort = gpioPortFromName (pPinDesc->port) ;
	pinMask	= 1u << pPinDesc->pin ;
	odr = pPort->ODR ;

	pPort->BSRR = ((odr & pinMask) << 16u) | (~odr & pinMask) ;
}

//----------------------------------------------------------------------
