/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	spi25xx.h	Driver for SPI EEPROM 25AA128/25LC128
									  25AA256/25LC256
									  25AA512/25LC512

	When		Who	What
	10/05/19	ac	Creation

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2021 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"spi25xx.h"

//----------------------------------------------------------------------
// EEPROM commands
#define		EE_READ			0x03u		// Read data from memory array beginning at selected address
#define		EE_WRITE		0x02u		// Write data to memory array beginning at selected address
#define		EE_WRDI			0x04u		// Reset the write enable latch (disable write operations)
#define		EE_WREN			0x06u		// Set the write enable latch (enable write operations)
#define		EE_RDSR			0x05u		// Read STATUS register
#define		EE_WRSR			0x01u		// Write STATUS register

// Bits of status register
#define		SR_WIP			0x01u		// Write In Progress
#define		SR_WEL			0x02u		// Write Enable Latch
#define		SR_BP0			0x04u		// Block protection 0
#define		SR_BP1			0x08u		// Block protection 1
#define		SR_WPEN			0x80u		// Write Protect Enable, related to hardware WP pin

static	spiStatus_t	writePage_ (spi25xxDesc_t * p25xxDesc, uint32_t address, void * pBuffer, uint32_t size) ;

//----------------------------------------------------------------------
//	Default SPI slave enable.
//	Use the pin defined as NSS in SPI descriptor

static	void	defSpiSlaveSelect (spiHandle_t spiHdl, uint32_t bSelect)
{
	spiSlaveSelect (spiHdl, bSelect) ;
}

//----------------------------------------------------------------------
//	Configure SPI port
//	Speed is a hint, because this SPI only allows a very limited number of speeds

spiStatus_t	spi25xxOpen (spi25xxDesc_t * p25xxDesc, spi25xxSize_t size, uint32_t speed)
{
	// Build EEPROM descriptor
	if (p25xxDesc->pSlaveSelect == NULL)
	{
		p25xxDesc->pSlaveSelect = defSpiSlaveSelect ;
	}
	switch (size)
	{
		case SPI25XX128:
			p25xxDesc->pageSize = 64 ;
			p25xxDesc->addrMask = 0x3F ;
			break ;

		case SPI25XX256:
			p25xxDesc->pageSize = 64 ;
			p25xxDesc->addrMask = 0x7F ;
			break ;

		case SPI25XX512:
			p25xxDesc->pageSize = 128 ;
			p25xxDesc->addrMask = 0xFF ;
			break ;

		default:
			return SPI_EFAIL ;
	}

	// Initialize SPI master
	(void) spiMasterConfigure (p25xxDesc->pSpiDesc, speed, SPI_NSS_USER | SPI_MODE_0, & p25xxDesc->spiHdl) ;

	// Deselect the slave
	(* p25xxDesc->pSlaveSelect) (p25xxDesc->spiHdl, 0u) ;

	return SPI_ENONE ;
}

//----------------------------------------------------------------------

void	spi25xxClose (spi25xxDesc_t * p25xxDesc)
{
	(void) p25xxDesc ;

}

//----------------------------------------------------------------------
//	Write a page or a partial page
//	This write must not cross page boundary

static	spiStatus_t	writePage_ (spi25xxDesc_t * p25xxDesc, uint32_t address, void * pBuffer, uint32_t size)
{
	spiHandle_t			spiHdl = p25xxDesc->spiHdl ;
	uint8_t				cmd [3] ;

	// Send WREN command
	cmd [0] = EE_WREN ;

	(* p25xxDesc->pSlaveSelect) (spiHdl, 1u) ;
	spiMasterExchange (spiHdl, cmd, NULL, 1u) ;
	(* p25xxDesc->pSlaveSelect) (spiHdl, 0u) ;

	// Send WRITE command and address, then data
	cmd [0] = EE_WRITE ;
	cmd [1] = (uint8_t) ((address >> 8u) & p25xxDesc->addrMask) ;
	cmd [2] = (uint8_t) (address & 0x00FFu) ;

	(* p25xxDesc->pSlaveSelect) (spiHdl, 1u) ;
	spiMasterExchange (spiHdl, cmd, NULL, 3u) ;
	spiMasterExchange (spiHdl, pBuffer, NULL, size) ;
	(* p25xxDesc->pSlaveSelect) (spiHdl, 0u) ;

	// Wait for the end of writing by reading the status register
	cmd [0] = EE_RDSR ;
	do
	{
		aaTaskDelay (1u) ;		// Write delay is ~5ms
		(* p25xxDesc->pSlaveSelect) (spiHdl, 1u) ;
		spiMasterExchange (spiHdl, & cmd[0], NULL, 1u) ;
		spiMasterExchange (spiHdl, NULL, & cmd[1], 1u) ;
		(* p25xxDesc->pSlaveSelect) (spiHdl, 0u) ;

	} while ((cmd[1] & SR_WIP) != 0u) ;

	return SPI_ENONE ;
}

//----------------------------------------------------------------------
//	Write data, managing EEPROM pages

spiStatus_t	spi25xxWrite (spi25xxDesc_t * p25xxDesc, uint32_t address, void * pBuffer, uint32_t size)
{
	uint32_t	addr    = address ;
	uint32_t	remain  = size ;
	uint8_t		* pData = (uint8_t *) pBuffer ;
	uint32_t	nn;

	// Write 1st page
	nn = p25xxDesc->pageSize - (addr % p25xxDesc->pageSize) ;		// Count of bytes to write to this page
	if (nn > remain)
	{
		nn = remain ;
	}
	writePage_ (p25xxDesc, addr, pData, nn) ;
	remain -= nn ;
	addr   += nn ;
	pData  += nn ;

	// Write full pages, except perhaps the last one
	while (remain != 0u)
	{
		nn = p25xxDesc->pageSize ;		// Count of bytes to write to this page
		if (nn > remain)
		{
			nn = remain ;
		}
		writePage_ (p25xxDesc, addr, pData, nn) ;
		remain -= nn ;
		addr   += nn ;
		pData  += nn ;
	}

	return SPI_ENONE ;
}

//----------------------------------------------------------------------
//	Read data

spiStatus_t	spi25xxRead (spi25xxDesc_t * p25xxDesc, uint32_t address, void * pBuffer, uint32_t size)
{
	spiHandle_t			spiHdl = p25xxDesc->spiHdl ;
	uint8_t				cmd [3] ;

	cmd [0] = EE_READ ;
	cmd [1] = (uint8_t) ((address >> 8u) & p25xxDesc->addrMask) ;
	cmd [2] = (address & 0x00FFu) ;

	(* p25xxDesc->pSlaveSelect) (spiHdl, 1u) ;			// Select SPI slave

	spiMasterExchange (spiHdl, cmd, NULL, 3u) ;			// Send Read command and address

	spiMasterExchange (spiHdl, NULL, pBuffer, size) ;	// Receive DATA

	(* p25xxDesc->pSlaveSelect) (spiHdl, 0u) ;			// Unselect SPI slave

	return SPI_ENONE ;
}

//----------------------------------------------------------------------
