/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	spibasic.h	SPI for G0 and L4+
				Master, polling or DMA exchanges, software or hardware SS management, no CRC
				Slave, asynchronous DMA only, software NSS management, no CRC

	When		Who	What
	27/02/20	ac	Creation for STM32H7
	31/03/20	ac	Port to STM32L4+
	13/04/21	ac	Port to STM32G0/L4+
					Add SPI slave management: spiSlaveConfigure() spiSlaveExchangeDma() spiSlaveAbort()
					Add spiConfigAgain()
	10/05/21	ac	remove speed from spiSlaveConfigure()

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2021 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#if ! defined SPIBASIC_H_
#define SPIBASIC_H_
//----------------------------------------------------------------------
#include	"gpiobasic.h"
#include	"dmabasic.h"

//	Set to 1 for supported SPI devices.
//	This allows to insert proper SPI descriptors from spiDescArray in spiXXxx.h

#define	WITH_SPI_1		1
#define	WITH_SPI_2		0
#define	WITH_SPI_3		0	// Not for all MCUs

// If one WITH_SPI_x is 1 then spiDescArray in spiH7xx.h is not empty and must be defined
#if (WITH_SPI_1 == 1  ||  WITH_SPI_2 == 1  ||  WITH_SPI_3 == 1)
	#define	WITH_SPI_X		1
#endif

//----------------------------------------------------------------------
// Flags for spiMasterConfigure()
// Choose one in each group

// Word size: flag value 4 to 16. A value of 0 is converted to 8.
#define	SPI_BITS_MASK			0x0000003F		// Internal use
#define	SPI_BITS_8				0x00000008		// 8 bits data word (default)
#define	SPI_BITS_12				0x0000000C		// 12 bits data word
#define	SPI_BITS_16				0x00000010		// 16 bits data word

#define	SPI_MODE_MASK			0x00000300		// Internal use
#define	SPI_MODE_0				0x00000000		// CPOL:0	CPHA:0 (Default)
#define	SPI_MODE_1				0x00000100		// CPOL:0	CPHA:1
#define	SPI_MODE_2				0x00000200		// CPOL:1	CPHA:0
#define	SPI_MODE_3				0x00000300		// CPOL:1	CPHA:1

#define	SPI_NSS_MASK			0x00001000		// Internal use
#define	SPI_NSS_AUTO			0x00000000		// NSS pin managed automatically by hardware (default)
												// BEWARE: Automatic NSS output need a pullup, even if the pin is flagged push-pull
#define	SPI_NSS_USER			0x00001000		// NSS pin is user managed (ignored by the hardware)

#define	SPI_MSB_MASK			0x00100000		// Internal use
#define	SPI_MSB_FIRST			0x00000000		// Data is transfered with the MSB first (default)
#define	SPI_LSB_FIRST			0x00100000		// Data is transfered with the LSB first

#define	SPI_DIR_MASK			0x03000000		// Internal use
#define	SPI_DIR_FULL_DUPLEX		0x00000000		// Full-Duplex mode or TX only. Rx and Tx transfer on 2 lines (default)
#define	SPI_DIR_RX_ONLY			0x01000000		// Receive only transfer on MOSI/MOSI (slave receive) or MISO/MISO (master receive)
#define	SPI_DIR_HALF_RX			0x02000000		// Half-Duplex Rx mode. Rx transfer on master MOSI <-> slave MISO
#define	SPI_DIR_HALF_TX			0x03000000		// Half-Duplex Tx mode. Tx transfer on master MOSI <-> slave MISO

// Library internal use only
#define	SPI_MASTER_				0x80000000		// Master mode

//----------------------------------------------------------------------
//	Options to use with spiMasterSetOption()
//	Some not for STM32G0 / F4

#define	SPI_OPT_SWAPIO			0x01			// Swap MOSI and MISO, if value is 0 no swap, else swap
#define	SPI_OPT_GAP				0x02			// Delay between words, the value is a number of bit clock
#define	SPI_OPT_TIMEOUT			0x03			// Timeout of exchange, the value is a number of tick

//----------------------------------------------------------------------
//	Default priority for DMA interrupt, used in SPI descriptors

#define	SPI_DMA_INT_PRIORITY		(BSP_MAX_INT_PRIO+2)

typedef struct spiDesc_t
{
	uint16_t		id ;			// Descriptor id
	SPI_TypeDef		* pSpi ;		// Pointer to SPI registers
	gpioPinDesc_t	pinMosi ;		// MOSI pin I/O descriptor
	gpioPinDesc_t	pinMiso ;		// MISO pin I/O descriptor
	gpioPinDesc_t	pinSck ;		// Clock pin I/O descriptor
	gpioPinDesc_t	pinNss ;		// Slave select pin I/O descriptor
	dmaDesc_t		dmaTx ;
	dmaDesc_t		dmaRx ;

} spiDesc_t ;

//----------------------------------------------------------------------
//	Returned errors

typedef enum
{
	SPI_ENONE		= 0,		// No error
	SPI_EARG		= 1,
	SPI_EFAIL		= 2,
	SPI_ENOTINIT	= 3,
	SPI_ETIMEOUT	= 4,
	SPI_EDMA		= 5

} spiStatus_t ;

// The SPI handle returned by spiMasterConfigure()
typedef void * spiHandle_t ;

#ifdef __cplusplus
extern "C" {
#endif

// Library internal use only
spiStatus_t			spiConfigure_			(const spiDesc_t * pSpiDesc, uint32_t speed, uint32_t flags, spiHandle_t * pHdl) ;


spiStatus_t			spiGetDesc				(uint32_t id, const spiDesc_t ** ppSpiDesc) ;
void				spiSlaveSelect			(spiHandle_t hdl, uint32_t bSelect) ;
uint32_t			spiSpeedToBr			(const spiDesc_t * pSpiDesc, uint32_t speed) ;
uint32_t			spiBrToSpeed			(const spiDesc_t * pSpiDesc, uint32_t br) ;
void				spiConfigAgain			(const spiHandle_t hdl) ;

__ALWAYS_STATIC_INLINE spiStatus_t			spiMasterConfigure		(const spiDesc_t * pSpiDesc, uint32_t speed, uint32_t flags, spiHandle_t * pHdl)
{
	return spiConfigure_ (pSpiDesc, speed, flags | SPI_MASTER_, pHdl) ;
}
spiStatus_t			spiMasterSetOption		(spiHandle_t hdl, uint32_t option, uintptr_t value) ;

spiStatus_t 		spiMasterExchange		(spiHandle_t hdl, void * pDataOut, void * pDataIn, uint32_t count) ;

// The function spiMasterExchangeDma() use the AA_SIGNAL_IO task signal
spiStatus_t			spiMasterExchangeDma 	(spiHandle_t hdl, void * pDataOut, void * pDataIn, uint32_t count) ;


__ALWAYS_STATIC_INLINE spiStatus_t			spiSlaveConfigure		(const spiDesc_t * pSpiDesc, uint32_t flags, spiHandle_t * pHdl)
{
	return spiConfigure_ (pSpiDesc, 0, flags & ~SPI_MASTER_, pHdl) ;
}

spiStatus_t			spiSlaveExchangeDma 	(spiHandle_t hdl, void * pDataOut, void * pDataIn, uint32_t count) ;
spiStatus_t			spiSlaveAbort			(spiHandle_t hdl) ;

#ifdef __cplusplus
}
#endif

//----------------------------------------------------------------------
#endif	// SPIBASIC_H_
