/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	timbasic.c	Basic routines for STM32G0 timers

	When		Who	What
	08/03/18	ac	Creation
	22/04/20	ac	Remove bug in interrupt handler
	28/06/20	ac	Add TIM_ENC_BOTH

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2021 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"
#include	"timbasic.h"
#include	"gpiobasic.h"

#if defined (DAC1)
	#include	"stm32G0xx_ll_dac.h"	// Common IRQ handler for TIM6 and DAC
#endif

//------------------------------------------------------------------------------
//	The DAC IRQ is common with some TIM IRQ
//	Allows to get DAC IRQ handler:
//	It is mandatory to include DAC management (WITH_DAC1 or WITH_DAC2 in dacbasic.h)
#define		WITH_DAC		0		// Include DAC IRQ management

#if (WITH_DAC == 1)
	extern	void	dacBasicIrqHandler (void) ;
#endif

//------------------------------------------------------------------------------
//	Descriptor array for this processor

// Include the proper timer descriptors
#if (WITH_TIM_X == 1)
	#if (defined STM32F405xx || defined STM32F415xx || defined STM32F407xx || defined STM32F417xx ||	\
		 defined STM32F427xx || defined STM32F437xx || defined STM32F429xx || defined STM32F439xx ||	\
		 defined STM32F469xx || defined STM32F479xx || defined STM32F446xx || defined STM32F411xE ||	\
		 defined STM32F401xC || defined STM32F401xE)
		#include "timF4xx.h"
	#endif

#if (defined STM32G0)
	#include "timG0xx.h"
#endif

	// Count of timers in timDescArray
	STATIC const	uint32_t	timAarrayCount = sizeof (timDescArray) / sizeof (timDesc_t) ;
#endif

//------------------------------------------------------------------------------
//	Every timer have a descriptor for interrupt callback management
//	Only 1 timer interrupt descriptor in use at any time for each timer

// Number of the highest timers in the MCU		THIS MUST BE UPDATED IF YOU CHANGE MCU
#define	MAX_TIMER	17

typedef struct
{
	const timDesc_t	* pTimDesc ;
	fnCallback_t	pCallback ;
	uintptr_t		arg ;

} timCallback_t ;

timCallback_t	timCallbackArray [MAX_TIMER] =
{
	{ NULL, NULL, 0u },		// TIM1
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },
	{ NULL, NULL, 0u },		// TIM14
	{ NULL, NULL, 0u },		// TIM15
	{ NULL, NULL, 0u },		// TIM16
	{ NULL, NULL, 0u },		// TIM17
} ;

//------------------------------------------------------------------------------
//	Convert channel number 1..4 to LL number LL_TIM_CHANNEL_CH1..LL_TIM_CHANNEL_CH4
//	LL_TIM_CHANNEL_CH1 == TIM_CCER_CC1E == (0x1U << TIM_CCER_CC1E_Pos)

static	uint32_t	timGetLlChannel (const timDesc_t * pTimDesc, uint32_t channel)
{
	if (channel == 1u  &&  (pTimDesc->capabilities & TIM_CAP_CHAN_1) != 0u)
	{
		return LL_TIM_CHANNEL_CH1 ;
	}

	if (channel == 2u  &&  (pTimDesc->capabilities & TIM_CAP_CHAN_2) != 0u)
	{
		return LL_TIM_CHANNEL_CH2 ;
	}

	if (channel == 3u  &&  (pTimDesc->capabilities & TIM_CAP_CHAN_3) != 0u)
	{
		return LL_TIM_CHANNEL_CH3 ;
	}

	if (channel == 4u  &&  (pTimDesc->capabilities & TIM_CAP_CHAN_4) != 0u)
	{
		return LL_TIM_CHANNEL_CH4 ;
	}
	return 0u ;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//	Return timer descriptor address from timer descriptor id

timError_t		timGetDesc (uint32_t id, const timDesc_t ** ppTimDesc)
{
	#if (WITH_TIM_X == 1)
		uint32_t	ii ;
		const timDesc_t	* pTimDesc = & timDescArray [0] ;

		for (ii = 0u ; ii < timAarrayCount ; ii++)
		{
			if (id == pTimDesc->id)
			{
				* ppTimDesc = pTimDesc ;
				return TIM_ENONE ;
			}
			pTimDesc++ ;
		}
	#else
		(void) id ;
		(void) ppTimDesc ;
	#endif

	return TIM_EFAIL ;	// Id not found
}

//------------------------------------------------------------------------------
//	Configure a timer for external clock mode 1 and 2

// TODO  Allows to configure ETR : prescaler, filter

timError_t	timExtClockInit (const timDesc_t * pTimDesc, uint32_t input, uint32_t prescaler, uint32_t flags)
{
	TIM_TypeDef		* pTim = pTimDesc->pTim ;
	uint32_t		llChannel ;

	// Enable timer clock
	rccEnableClock (pTim, 0) ;

	// Reset timer
	rccResetPeriph (pTim, 0) ;

	// Configure input pin
	if (input == LL_TIM_TS_TI1FP1  ||  input == LL_TIM_TS_TI1F_ED)
	{
		llChannel = LL_TIM_CHANNEL_CH1 ;
		gpioConfigurePin (& pTimDesc->pins [TIM_PIN_CH1]) ;
	}

	if (input == LL_TIM_TS_TI2FP2)
	{
		llChannel = LL_TIM_CHANNEL_CH2 ;
		gpioConfigurePin (& pTimDesc->pins [TIM_PIN_CH2]) ;
	}

	if (input == LL_TIM_TS_ETRF)
	{
		gpioConfigurePin (& pTimDesc->pins [TIM_PIN_ETR]) ;
	}

	// Configure input channel
	if (input == LL_TIM_TS_TI1FP1  ||  input == LL_TIM_TS_TI2FP2)
	{
		// Set CCMR1_CCxS : DIRECTTI(01) / INDIRECTTI(10) / TRC(11)
		LL_TIM_IC_SetActiveInput (pTim, llChannel, LL_TIM_ACTIVEINPUT_DIRECTTI) ;

		// CCER_CCxP : Reset value is RISING
		#if defined (LL_TIM_IC_POLARITY_BOTHEDGE)
			if ((flags & (TIM_ECLK_EDGE_RISING | TIM_ECLK_EDGE_FALLING)) == (TIM_ECLK_EDGE_RISING | TIM_ECLK_EDGE_FALLING))
			{
				LL_TIM_IC_SetPolarity (pTim, llChannel, LL_TIM_IC_POLARITY_BOTHEDGE) ;
			}
			else
		#endif
		if ((flags & TIM_ECLK_EDGE_FALLING) != 0u)
		{
			LL_TIM_IC_SetPolarity (pTim, llChannel, LL_TIM_IC_POLARITY_FALLING) ;
		}
		else
		{
			// Nothing, Reset value is RISING
		}
	}

	// CCMR1_1C1F : Input capture filter. Reset value is LL_TIM_IC_FILTER_FDIV1
	//LL_TIM_IC_SetFilter (pTim, llChannel, LL_TIM_IC_FILTER_FDIV1) ;

	// Select SMCR_TS : ITR0 ITR1 ITR2 ITR3 TI1F_ED TI1FP1 TI2FP2 ETRF
	// To be done when SMCR_SMS is 000
	LL_TIM_SetTriggerInput (pTim, input) ;

	if (input == LL_TIM_TS_ETRF)
	{
		// This is reset configuration
		LL_TIM_ConfigETR (pTim, LL_TIM_ETR_POLARITY_NONINVERTED, LL_TIM_ETR_PRESCALER_DIV1, LL_TIM_ETR_FILTER_FDIV1);

		// Select SMCR_SMS : internal / mode 1 / mode 2
		LL_TIM_SetClockSource (pTim, LL_TIM_CLOCKSOURCE_EXT_MODE2) ;
	}
	else
	{
		// Select SMCR_SMS : internal / mode 1 / mode 2
		LL_TIM_SetClockSource (pTim, LL_TIM_CLOCKSOURCE_EXT_MODE1) ;

		// Disable external clock mode 2. Reset value is disabled
		//LL_TIM_DisableExternalClock (pTim) ;
	}

	LL_TIM_SetPrescaler (pTim, prescaler - 1u) ;

	// DIER_TIE : reset value is Disabled trigger interrupt (TIE)
	//LL_TIM_DisableIT_TRIG (pTim) ;

	// Reset value is Disabled trigger DMA request (TDE)
	//LL_TIM_DisableDMAReq_TRIG (pTim) ;

	// Reset value is LL_TIM_TRGO_RESET
	//LL_TIM_SetTriggerOutput (pTim, LL_TIM_TRGO_RESET) ;

	// Reset value is Disabled Master/Slave mode
	//LL_TIM_DisableMasterSlaveMode (pTim) ;

	return TIM_ENONE ;
}

//------------------------------------------------------------------------------
//	Initialize a timer for PWM
//	prescaler is the prescaler value to get timer clock frequency. if 0 then timTimeBaseInit() compute the prescaler value.
//	if TIM_INIT_FORCE_ARR is specified
//		timeBaseFrequency is exact ARR value (the value is not decremented before writing to register)
//	else
//		timeBaseFrequency is the desired time base or PWM frequency (in Hz), which is used to compute ARR value

void  timTimeBaseInit (const timDesc_t * pTimDesc, uint32_t prescaler, uint32_t timeBaseFrequency, uint32_t flags)
{
	TIM_TypeDef		* pTim = pTimDesc->pTim ;
	uint32_t		clock = rccGetTimerClockFrequency (pTim) ;

	// Enable timer clock
	rccEnableClock (pTimDesc->pTim, 0) ;
	
	// Reset timer
	rccResetPeriph (pTimDesc->pTim, 0) ;

	// Time base configuration
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Set counter mode. Reset value is LL_TIM_COUNTERMODE_UP
	if ((flags & (TIM_CR1_CMS | TIM_CR1_DIR)) != 0u)
	{
		LL_TIM_SetCounterMode (pTim, flags & (TIM_CR1_CMS | TIM_CR1_DIR)) ;
	}

	// Set ARR Auto reload, default value is ARR not buffered
	if ((flags & TIM_INIT_ARR_PRELOADEN) != 0u)
	{
		LL_TIM_EnableARRPreload (pTim) ;
	}

	// Set UDIS bit: Update event disabled
	if ((flags & TIM_INIT_UPDATEDIS) != 0u)
	{
		LL_TIM_EnableUpdateEvent (pTim) ;
	}

	// Set Division ratio for digital filter of ETR or TIx, default is DIV1
	LL_TIM_SetClockDivision (pTim, flags & (TIM_CR1_CKD_1 | TIM_CR1_CKD_0)) ;

	LL_TIM_SetOnePulseMode (pTim, flags & LL_TIM_ONEPULSEMODE_SINGLE) ;
	LL_TIM_SetUpdateSource (pTim, flags & LL_TIM_UPDATESOURCE_COUNTER) ;

	// Compute and set prescaler value
	if (prescaler != 0u)
	{
		LL_TIM_SetPrescaler (pTim, prescaler - 1u) ;
	}
	else
	{
		// Compute prescaler value
		uint32_t	pre = clock / timeBaseFrequency ;
		if ((pTimDesc->capabilities & TIM_CAP_32_BITS) != 0u)
		{
			// 32 bit CNT: No prescaler
			// TODO : compute prescaler with 64 bits values ?
			LL_TIM_SetPrescaler (pTim, 0u) ;	// Divide by 1
		}
		else
		{
			// 16 bit CNT
			if (pre > 65535)
			{
				// Need prescaler > 1
				pre = rccGetTimerClockFrequency (pTim) / (timeBaseFrequency * 32768);
				LL_TIM_SetPrescaler (pTim, pre - 1u) ;
			}
			else
			{
				LL_TIM_SetPrescaler (pTim, 0u) ;	// Divide by 1
			}
		}
	}

	// Set ARR value
	if ((flags & TIM_INIT_FORCE_ARR) == 0u)
	{
		// Set ARR register to get timeBaseFrequency timer frequency
		LL_TIM_SetAutoReload (pTim, __LL_TIM_CALC_ARR (clock, LL_TIM_GetPrescaler (pTim), timeBaseFrequency)) ;
	}
	else
	{
		// TIM_INIT_FORCE_ARR specify: timeBaseFrequency is the ARR value
		LL_TIM_SetAutoReload (pTim, timeBaseFrequency) ;
	}
}

//------------------------------------------------------------------------------
//	Set master timer output TRGO source
//	flags is one of LL_TIM_TRGO_xxx

void	timSetTriggerOutput		(const timDesc_t * pTimDesc, uint32_t flags)
{
	TIM_TypeDef		* pTim = pTimDesc->pTim ;

	LL_TIM_SetTriggerOutput (pTim, flags);
}

//------------------------------------------------------------------------------
//	Initialize timer as encoder mode

// Flags can contain a filter value : LL_TIM_IC_FILTER_FDIVx_Ny defined by xxx_ll_tim.h
// Default value is no filter

void		timEncoderInit (const timDesc_t * pTimDesc, uint32_t flags)
{
	TIM_TypeDef 	* pTim = pTimDesc->pTim ;
	uint32_t		temp ;

	// Check if encoder is available on this timer
	if (! IS_TIM_ENCODER_INTERFACE_INSTANCE (pTim))
	{
		AA_ASSERT (0)	;
		return ;
	}

	// Enable timer clock
	rccEnableClock (pTim, 0) ;

	// Reset timer
	rccResetPeriph (pTim, 0) ;

	// Enable pins
	gpioConfigurePin (& pTimDesc->pins [TIM_PIN_CH1]) ;
	gpioConfigurePin (& pTimDesc->pins [TIM_PIN_CH2]) ;

	// Select encoder interface mode
	if ((flags & TIM_ENC_BOTH) != 0)
	{
		// Count on TI1FP1 and TI2FP2 edge
		LL_TIM_SetEncoderMode (pTim, LL_TIM_ENCODERMODE_X4_TI12);
	}
	else if ((flags & TIM_ENC_SWITCH) == 0u)
	{
		// Count on TI1FP1 edge
		LL_TIM_SetEncoderMode (pTim, LL_TIM_ENCODERMODE_X2_TI1);
	}
	else
	{
		// Count on TI2FP2 edge
		LL_TIM_SetEncoderMode (pTim, LL_TIM_ENCODERMODE_X2_TI2);
	}

	// Set edge polarity
	if ((flags & TIM_ENC_INVCH1) != 0u)
	{
		LL_TIM_IC_SetPolarity (pTim, LL_TIM_CHANNEL_CH1, LL_TIM_IC_POLARITY_FALLING);
	}
	if ((flags & TIM_ENC_INVCH2) != 0u)
	{
		LL_TIM_IC_SetPolarity (pTim, LL_TIM_CHANNEL_CH2, LL_TIM_IC_POLARITY_FALLING);
	}

	// Activate inputs, e.g. IC1 on TI1, IC2 on TI2
	temp = LL_TIM_ACTIVEINPUT_DIRECTTI ;
	if ((flags & TIM_ENC_IC1_INDIRECTTI) != 0u)
	{
		temp = LL_TIM_ACTIVEINPUT_INDIRECTTI ;
	}
	LL_TIM_IC_SetActiveInput (pTim, LL_TIM_CHANNEL_CH1, temp) ;

	temp = LL_TIM_ACTIVEINPUT_DIRECTTI ;
	if ((flags & TIM_ENC_IC2_INDIRECTTI) != 0u)
	{
		temp = LL_TIM_ACTIVEINPUT_INDIRECTTI ;
	}
	LL_TIM_IC_SetActiveInput (pTim, LL_TIM_CHANNEL_CH2, temp) ;

	// Set filter on IC1 and IC2
	LL_TIM_IC_SetFilter (pTim, LL_TIM_CHANNEL_CH1, flags & 0x000F0000);
	LL_TIM_IC_SetFilter (pTim, LL_TIM_CHANNEL_CH2, flags & 0x000F0000);

	// Set ARR to simulate free running counter
	if ((pTimDesc->capabilities & TIM_CAP_32_BITS) == 0u)
	{
		pTim->ARR = 0xFFFFu ;
	}
	else
	{
		pTim->ARR = 0xFFFFFFFFu ;
	}
}

//------------------------------------------------------------------------------
//	Returns the native pointer to timer registers. Allows application to configure the timer

void		timGetRegPtr	(const timDesc_t * pTimDesc, TIM_TypeDef ** ppTimPtr)
{
	* ppTimPtr = pTimDesc->pTim ;
}

//------------------------------------------------------------------------------
//	Set the timer counter value

void		timSetCounter	(const timDesc_t * pTimDesc, uint32_t value)
{
	LL_TIM_SetCounter (pTimDesc->pTim, value) ;
}

//------------------------------------------------------------------------------
//	Get the current timer counter value

uint32_t	timGetCounter	(const timDesc_t * pTimDesc)
{
	return LL_TIM_GetCounter (pTimDesc->pTim) ;
}

//------------------------------------------------------------------------------
//	Set timer reload counter value. This value is directly written in the register.

void		timSetReload	(const timDesc_t * pTimDesc, uint32_t reload)
{
	pTimDesc->pTim->ARR = reload ;
}

//------------------------------------------------------------------------------
//	Get timer reload counter value. This value is directly read from the register

uint32_t	timGetReload (const timDesc_t * pTimDesc)
{
	 return (uint32_t)(pTimDesc->pTim->ARR) ;
}

//------------------------------------------------------------------------------
//	Configure timer interrupt
//	Flags contain the interrupt flag mask of the interrupt to enable

// Warning: some timers (as TIM1 and TIM8) can use more than 1 interrupt (up to 4)
// TODO : this function manage only 1 IRQn per timer

void	timConfigureIntr (const timDesc_t * pTimDesc, uint32_t priority, fnCallback_t pCallback, uintptr_t arg, uint32_t flags)
{
	TIM_TypeDef * pTim = pTimDesc->pTim ;

	if ((flags & TIM_INTR_DISABLE) != 0u)
	{
		// Disable all interrupts of this timer
		pTim->DIER &= ~(LL_TIM_DIER_UIE | LL_TIM_DIER_CC1IE | LL_TIM_DIER_CC2IE | LL_TIM_DIER_CC3IE | LL_TIM_DIER_CC4IE | LL_TIM_DIER_COMIE | LL_TIM_DIER_TIE | LL_TIM_DIER_BIE) ;
		NVIC_DisableIRQ (pTimDesc->timIrqn) ;
		return ;
	}

	// Configure the NVIC to handle the TIM interrupt
	NVIC_SetPriority (pTimDesc->timIrqn, priority) ;
	NVIC_EnableIRQ (pTimDesc->timIrqn) ;

	timCallbackArray [pTimDesc->timNumber-1u].pTimDesc	= pTimDesc ;
	timCallbackArray [pTimDesc->timNumber-1u].pCallback	= pCallback ;
	timCallbackArray [pTimDesc->timNumber-1u].arg		= arg ;

	if ((flags & TIM_INTR_UPDATE) != 0u)
	{
		LL_TIM_EnableIT_UPDATE (pTim) ;
	}

	if ((flags & LL_TIM_DIER_COMIE) != 0u)
	{
		LL_TIM_EnableIT_COM (pTim) ;
	}

	if ((flags & LL_TIM_DIER_TIE) != 0u)
	{
		LL_TIM_EnableIT_TRIG (pTim) ;
	}

	if ((flags & LL_TIM_DIER_BIE) != 0u)
	{
		LL_TIM_EnableIT_BRK (pTim) ;
	}

	// Enable the capture/compare interrupt for channel
	if ((flags & LL_TIM_DIER_CC1IE) != 0u)
	{
		LL_TIM_EnableIT_CC1 (pTim) ;
	}

	if ((flags & LL_TIM_DIER_CC2IE) != 0u)
	{
		LL_TIM_EnableIT_CC2 (pTim) ;
	}

	if ((flags & LL_TIM_DIER_CC3IE) != 0u)
	{
		LL_TIM_EnableIT_CC3 (pTim) ;
	}

	if ((flags & LL_TIM_DIER_CC4IE) != 0u)
	{
		LL_TIM_EnableIT_CC4 (pTim) ;
	}
}

//------------------------------------------------------------------------------
//	Start timer

void	timStart (const timDesc_t * pTimDesc)
{
	TIM_TypeDef * pTim = pTimDesc->pTim ;
	
	// Force update generation: Clear CNT and prescaler
	LL_TIM_GenerateEvent_UPDATE (pTim) ;

	// Enable counter
	LL_TIM_EnableCounter (pTim) ;
}

//------------------------------------------------------------------------------
//	Stop timer

void	timStop (const timDesc_t * pTimDesc)
{
	// Disable counter
	LL_TIM_DisableCounter (pTimDesc->pTim) ;
}

//------------------------------------------------------------------------------
//	Disable the timer: end PWM generation on all channels

void	timPwmReset (const timDesc_t * pTimDesc)
{
	TIM_TypeDef * pTim = pTimDesc->pTim ;
	
	pTim->CR1  = 0u ;		// Timer disabled
	pTim->DIER = 0u ;		// All IRQs disabled
	pTim->SR   = 0u ;		// Clear eventual pending IRQs

	// Disable all channels
	if ((pTimDesc->capabilities & TIM_CAP_CHAN_1) != 0u)
	{
		timOutputChannelDisable (pTimDesc, 1u) ;
	}
	if ((pTimDesc->capabilities & TIM_CAP_CHAN_2) != 0u)
	{
		timOutputChannelDisable (pTimDesc, 2u) ;
	}
	if ((pTimDesc->capabilities & TIM_CAP_CHAN_3) != 0u)
	{
		timOutputChannelDisable (pTimDesc, 3u) ;
	}
	if ((pTimDesc->capabilities & TIM_CAP_CHAN_4) != 0u)
	{
		timOutputChannelDisable (pTimDesc, 4u) ;
	}
}

//------------------------------------------------------------------------------
//	Set channel PWM ratio. Channel 1..4

void	timOutputChannelSet (const timDesc_t * pTimDesc, uint32_t channel, uint32_t value)
{
	volatile uint32_t	* pReg ;
	
	if (channel > 0u  &&  channel < 5u)
	{
		pReg = & pTimDesc->pTim->CCR1 ;
		pReg [channel-1u] = value ;
		//LL_TIM_OC_SetCompareCH1 (pTimDesc->pTim, value);
	}
}

//------------------------------------------------------------------------------
//	Configure and enable an output channel (e.g. PWM).
//	Does not manage complementary channels of TIM1 or TIM8
//	Channel		1..4
//	Flags		Combination of bits of TIMx_CCMRx and TIMx_CCER bits in output mode


void	timOutputChannelEnable (const timDesc_t * pTimDesc, uint32_t channel, uint32_t flags)
{
	TIM_TypeDef		* pTim = pTimDesc->pTim ;
	uint32_t		llChannel ;

	llChannel = timGetLlChannel (pTimDesc, channel) ;
	if (llChannel == 0u)
	{
		return ;
	}
	
	// Configure GPIO pin
	if (pTimDesc->pins [channel - 1u].port != 0u)
	{
		gpioConfigurePin (& pTimDesc->pins [channel - 1u]) ;
	}

	// Disable channel output
	LL_TIM_CC_DisableChannel (pTim, llChannel) ;


	// Set output mode (OCxM and CCxS). Reset value is LL_TIM_OCMODE_FROZEN
	LL_TIM_OC_SetMode (pTim, llChannel, flags & TIM_CCMR1_OC1M_Msk) ;

	// Set TIMx_CCRx register preload (OCxPE).
	if (flags & TIM_OCMODE_PRELOADEN)
	{
		// TIM_CCRx preload value is loaded in the shadow register at each update event.
		LL_TIM_OC_EnablePreload (pTim, llChannel) ;
	}
	else
	{
		// TIM_CCRx preload value is loaded immediately in the shadow register.
		LL_TIM_OC_DisablePreload (pTim, llChannel) ;
	}

	// Set output compare fast
	if (flags & TIM_OCMODE_FASTEN)
	{
		// TIM_CCRx preload value is loaded in the shadow register at each update event.
		LL_TIM_OC_EnableFast (pTim, llChannel) ;
	}
	else
	{
		// TIM_CCRx preload value is loaded immediately in shadow register.
		LL_TIM_OC_DisableFast (pTim, llChannel) ;
	}

	// Set output compare clear on ETRF input high
	if (flags & TIM_OCMODE_CLEAREN)
	{
		// TIM_CCRx preload value is loaded in the shadow register at each update event.
		LL_TIM_OC_EnableClear (pTim, llChannel) ;
	}
	else
	{
		// TIM_CCRx preload value is loaded immediately in shadow register.
		LL_TIM_OC_DisableClear (pTim, llChannel) ;
	}

	// Set output channel polarity. Reset value is LL_TIM_OCPOLARITY_HIGH
	LL_TIM_OC_SetPolarity (pTim, llChannel, (flags >> 16u) & TIM_CCER_CC1P_Msk) ;

	// Set compare value to default: half of the counter period (50% duty cycle)
	// This is useful to configure frequency output without managing PWM
	timOutputChannelSet (pTimDesc, channel, ((LL_TIM_GetAutoReload (pTim) + 1u ) / 2u)) ;

	// For some timers (e.g. TIM1 TIM8) need to set MOE
	if (IS_TIM_BREAK_INSTANCE (pTim))
	{
		pTim->BDTR |= TIM_BDTR_MOE ;
	}

	// Enable channel output
	LL_TIM_CC_EnableChannel (pTim, llChannel) ;
}

//------------------------------------------------------------------------------
//	Reset channel PWM configuration, and disable its output.
//	Channel		1..4

void	timChannelDisable (const timDesc_t * pTimDesc, uint32_t channel)
{
	TIM_TypeDef		* pTim = pTimDesc->pTim ;
	uint32_t		llChannel ;

	llChannel = timGetLlChannel (pTimDesc, channel) ;
	if (llChannel == 0u)
	{
		return ;
	}

	// Disable output
	LL_TIM_CC_DisableChannel (pTim, llChannel) ;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//	Input capture management
//	The prescaler is the real prescaler value (not the value -1)

void			timInputInit			(const timDesc_t * pTimDesc, uint32_t prescaler)
{
	// Enable timer clock
	rccEnableClock (pTimDesc->pTim, 0) ;

	// Reset timer
	rccResetPeriph (pTimDesc->pTim, 0) ;

	pTimDesc->pTim->PSC = prescaler - 1 ;
}

//------------------------------------------------------------------------------
//	Configure and enable an input capture channel

void	timInputChannelInit	(const timDesc_t * pTimDesc, uint32_t channel, uint32_t flags)
{
	TIM_TypeDef		* pTim = pTimDesc->pTim ;
	uint32_t		ccmr, ccer ;

	if (channel == 0u  ||  channel > 4u)
	{
		AA_ASSERT (0) ;
		return ;
	}

	// Configure GPIO pin
	if (pTimDesc->pins [channel - 1u].port != 0u)
	{
		gpioConfigurePin (& pTimDesc->pins [channel - 1u]) ;
	}

	// Set channel in input mode
	// Set active polarity
	// Enable channel capture
	ccmr = flags & 0x000Fu ;				// IC1PSC + CC1S
	ccmr |= ((flags >> 16) & 0x00F0u) ;		// + IC1F
	ccer = ((flags >> 24) & (TIM_CCER_CC1P | TIM_CCER_CC1NP)) | TIM_CCER_CC1E ;	// IC1P + CC1NP + CC1E
	switch (channel)
	{
		case 1:
			pTim->CCER  &= ~TIM_CCER_CC1E ;		// Disable channel
			pTim->CCMR1 = (pTim->CCMR1 & ~0x00FFu) | ccmr ;
			pTim->CCER  = (pTim->CCER & 0x000Fu)   | ccer ;
			break ;

		case 2:
			pTim->CCER  &= ~TIM_CCER_CC2E ;		// Disable channel
			pTim->CCMR1 = (pTim->CCMR1 & ~0xFF00u) | (ccmr << 8u) ;
			pTim->CCER  = (pTim->CCER & 0x00F0u)   | (ccer << 4u) ;
			break ;

		case 3:
			pTim->CCER  &= ~TIM_CCER_CC3E ;		// Disable channel
			pTim->CCMR2 = (pTim->CCMR2 & ~0x00FFu) | ccmr ;
			pTim->CCER  = (pTim->CCER & 0x0F00u)   | (ccer << 8u) ;
			break ;

		case 4:
			pTim->CCER  &= ~TIM_CCER_CC4E ;		// Disable channel
			pTim->CCMR2 = (pTim->CCMR2 & ~0xFF00u) |(ccmr << 8u) ;
			pTim->CCER  = (pTim->CCER & 0xF000u)   | (ccer << 12u) ;
			break ;

		default:
			break ;
	}
}

//------------------------------------------------------------------------------
//	Enable a combination of multiple channels
//	channelMask is not a channel number but a bit mask of the channels to enable
//	Example: to enable channel 3 and channel 4 at the same time use the mask:
//		(1 << 3) | (1 << 4)

void	timInputChannelEnable	(const timDesc_t * pTimDesc, uint32_t channelMask)
{
	uint32_t		ccer ;

	ccer  = (channelMask & 0x02u) >> 1u ;
	ccer |= (channelMask & 0x04u) << 3u ;
	ccer |= (channelMask & 0x08u) << 7u ;
	ccer |= (channelMask & 0x10u) << 11u ;

	pTimDesc->pTim->CCER |= ccer ;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

#define	TIM_SR_IMASK	(TIM_SR_TIF | TIM_SR_CC4IF |  TIM_SR_CC3IF |  TIM_SR_CC2IF |  TIM_SR_CC1IF |  TIM_SR_UIF)

__STATIC_INLINE void callCallback (TIM_TypeDef * pTim, timCallback_t * pCbDesc)
{
	uint32_t	sr = pTim->SR ;

	if (pCbDesc->pCallback != NULL)
	{
		(* pCbDesc->pCallback) (pCbDesc->pTimDesc, pCbDesc->arg, sr) ;
	}
	if ((sr & TIM_SR_UIF) != 0u)
	{
		pTim->SR = ~TIM_SR_UIF ;	// clear UIF flag
	}
}

//------------------------------------------------------------------------------

#if (WITH_TIM_1 == 1)
void TIM1_BRK_UP_TRG_COM_IRQHandler (void) ;
void TIM1_BRK_UP_TRG_COM_IRQHandler (void)
{
	aaIntEnter () ;

	// User must clear interrupt flags other than UIF
	callCallback (TIM1, & timCallbackArray [1u-1u]) ;

	aaIntExit () ;
}
#endif	// WITH_TIM1 || WITH_TIM10

//------------------------------------------------------------------------------

#if (WITH_TIM_1 == 1)
void	TIM1_CC_IRQHandler (void) ;
void	TIM1_CC_IRQHandler (void)
{
	aaIntEnter () ;

	// User must clear interrupt flags other than UIF
	callCallback (TIM1, & timCallbackArray [1u-1u]) ;

	aaIntExit () ;
}
#endif	// WITH_TIM1

//------------------------------------------------------------------------------

#if (WITH_TIM_2 == 1)
void	TIM2_IRQHandler(void) ;
void	TIM2_IRQHandler(void)
{
	aaIntEnter () ;

	// User must clear interrupt flags other than UIF
	callCallback (TIM2, & timCallbackArray [2u-1u]) ;

	aaIntExit () ;
}
#endif	// WITH_TIM2

//------------------------------------------------------------------------------

#if (WITH_TIM_3 == 1)
void	TIM3_IRQHandler(void) ;
void	TIM3_IRQHandler(void)
{
	aaIntEnter () ;

	// User must clear interrupt flags other than UIF
	callCallback (TIM3, & timCallbackArray [3u-1u]) ;

	aaIntExit () ;
}
#endif	// WITH_TIM3

//------------------------------------------------------------------------------
//	Common interrupt handler for TIM6 and DAC
//	Manage DAC1

#if ((WITH_TIM_6 == 1)  ||  (WITH_DAC == 1))
void TIM6_DAC_LPTIM1_IRQHandler (void) ;
void TIM6_DAC_LPTIM1_IRQHandler (void)
{
	aaIntEnter () ;

	#if (WITH_TIM_6 == 1)
	{
		if (((TIM6->SR & TIM_SR_IMASK) & TIM6->DIER) != 0u)
		{
			// User must clear interrupt flags other than UIF
			callCallback (TIM6, & timCallbackArray [6u-1u]) ;
		}
	}
	#endif	// WITH_TIM_6

	#if (WITH_DAC == 1)
	{
		//	Check if this is a DMA under run interrupt
		if ((DAC1->SR & (DAC_SR_DMAUDR1 | DAC_SR_DMAUDR2)) != 0u)
		{
			dacBasicIrqHandler () ;		// In dacbasic.c
		}
	}
	#endif	// WITH_DAC

	// TODO : Manage LPTIM1 interrupts

	aaIntExit () ;
}
#endif	// WITH_TIM6

//------------------------------------------------------------------------------

#if (WITH_TIM_7 == 1)
void	TIM7_LPTIM2_IRQHandler (void) ;
void	TIM7_LPTIM2_IRQHandler (void)
{
	aaIntEnter () ;

	#if (WITH_TIM_7 == 1)
	{
		if (((TIM7->SR & TIM_SR_IMASK) & TIM7->DIER) != 0u)
		{
			// User must clear interrupt flags other than UIF
			callCallback (TIM7, & timCallbackArray [7u-1u]) ;
		}
	}
	#endif	// WITH_TIM8

// TODO : Manage LPTIM2 interrupts

	aaIntExit () ;
}
#endif	// WITH_TIM7

//------------------------------------------------------------------------------

#if (WITH_TIM_14 == 1)
void	TIM14_IRQHandler (void) ;
void	TIM14_IRQHandler (void)
{
	aaIntEnter () ;

	// User must clear interrupt flags other than UIF
	callCallback (TIM14, & timCallbackArray [14u-1u]) ;

	aaIntExit () ;
}
#endif	// WITH_TIM14

//------------------------------------------------------------------------------

#if (WITH_TIM_15 == 1)
void	TIM15_IRQHandler (void) ;
void	TIM15_IRQHandler (void)
{
	aaIntEnter () ;

	// User must clear interrupt flags other than UIF
	callCallback (TIM15, & timCallbackArray [15u-1u]) ;

	aaIntExit () ;
}
#endif	// WITH_TIM15

//------------------------------------------------------------------------------

#if (WITH_TIM_16 == 1)
void	TIM16_IRQHandler (void) ;
void	TIM16_IRQHandler (void)
{
	aaIntEnter () ;

	// User must clear interrupt flags other than UIF
	callCallback (TIM16, & timCallbackArray [16u-1u]) ;

	aaIntExit () ;
}
#endif	// WITH_TIM16

//------------------------------------------------------------------------------

#if (WITH_TIM_17 == 1)
void	TIM17_IRQHandler (void) ;
void	TIM17_IRQHandler (void)
{
	aaIntEnter () ;

	// User must clear interrupt flags other than UIF
	callCallback (TIM17, & timCallbackArray [17u-1u]) ;

	aaIntExit () ;
}
#endif	// WITH_TIM17

//------------------------------------------------------------------------------
