/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	timbasic.h	Basic routines for STM32G0xx timers

	When		Who	What
	06/02/21	ac	Creation
	12/07/21	ac	Add input capture management

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2021 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#ifndef TIMBASIC_H_
#define TIMBASIC_H_
//------------------------------------------------------------------------------
#include	"gpiobasic.h"
#include	"stm32g0xx_ll_tim.h"

//------------------------------------------------------------------------------
//	Define which timer are managed by timbasic
//	Those who are at 0 can have a user-managed interrupt handlers and descriptor
//	Adjust these values to your MCU

#define		WITH_TIM_1		0		// Include TIM1
#define		WITH_TIM_2		0		// Include TIM2
#define		WITH_TIM_3		0		// Include TIM3

#define		WITH_TIM_4		0		// Include TIM4
#define		WITH_TIM_5		0		// Include TIM5		Not in G0

#define		WITH_TIM_6		0		// Include TIM6
#define		WITH_TIM_7		1		// Include TIM7

#define		WITH_TIM_8		0		// Include TIM8		Not in G0
#define		WITH_TIM_9		0		// Include TIM9		Not in G0
#define		WITH_TIM_10		0		// Include TIM10	Not in G0
#define		WITH_TIM_11		0		// Include TIM11	Not in G0
#define		WITH_TIM_12		0		// Include TIM12	Not in G0
#define		WITH_TIM_13		0		// Include TIM13	Not in G0

#define		WITH_TIM_14		0		// Include TIM14
#define		WITH_TIM_15		0		// Include TIM15
#define		WITH_TIM_16		0		// Include TIM16
#define		WITH_TIM_17		0		// Include TIM17

// If one WITH_TIM_x is 1 then timDescArray in timG0xx.h is not empty and must be defined
#if (WITH_TIM_1 == 1   ||  WITH_TIM_2 == 1  ||  WITH_TIM_3 == 1  ||  WITH_TIM_4 == 1   ||  WITH_TIM_5 == 1   ||  WITH_TIM_6 == 1   || \
	 WITH_TIM_7 == 1   ||  WITH_TIM_8 == 1  ||  WITH_TIM_9 == 1  ||  WITH_TIM_10 == 1  ||  WITH_TIM_11 == 1  ||  WITH_TIM_12 == 1  || \
	 WITH_TIM_13 == 1  ||  WITH_TIM_14 == 1 ||  WITH_TIM_15 == 1 ||  WITH_TIM_16 == 1  ||  WITH_TIM_17 == 1)
	#define	WITH_TIM_X		1
#endif

// Check if selected TIM are defined for this MCU
#if (WITH_TIM_1 == 1   &&  ! defined TIM1)	||	\
	(WITH_TIM_2 == 1   &&  ! defined TIM2)	||	\
	(WITH_TIM_3 == 1   &&  ! defined TIM3)	||	\
	(WITH_TIM_4 == 1   &&  ! defined TIM4)	||	\
	(WITH_TIM_5 == 1   &&  ! defined TIM5)	||	\
	(WITH_TIM_6 == 1   &&  ! defined TIM6)	||	\
	(WITH_TIM_7 == 1   &&  ! defined TIM7)	||	\
	(WITH_TIM_8 == 1   &&  ! defined TIM8)	||	\
	(WITH_TIM_9 == 1   &&  ! defined TIM9)	||	\
	(WITH_TIM_10 == 1  &&  ! defined TIM10)	||	\
	(WITH_TIM_11 == 1  &&  ! defined TIM11)	||	\
	(WITH_TIM_12 == 1  &&  ! defined TIM12)	||	\
	(WITH_TIM_13 == 1  &&  ! defined TIM13)	||	\
	(WITH_TIM_14 == 1  &&  ! defined TIM14)	||	\
	(WITH_TIM_15 == 1  &&  ! defined TIM15)	||	\
	(WITH_TIM_16 == 1  &&  ! defined TIM16)	||	\
	(WITH_TIM_17 == 1  &&  ! defined TIM17)
	#error	"TIM does not exist in this MCU"
#endif

//------------------------------------------------------------------------------
// Timer descriptor for timer basic library

typedef struct timDesc_t
{
	uint8_t			id ;				// Timer descriptor identifier
	uint8_t			timNumber ;			// Timer number: 1..
//	uint8_t			timRstPos ;			// RCC_APBxRSTR_TIMnRST_Pos
//	uint8_t			timEnPos ;			// RCC_APBxRSTR_TIMnEN_Pos
	uint32_t		capabilities ;		// Capabilities flags
	TIM_TypeDef		* pTim ;			// A pointer to TIM registers: e.g. TIM1
	IRQn_Type		timIrqn ;			// The interrupt vector for this timer: e.g. TIM1_IRQn
	gpioPinDesc_t	pins [9] ;

} timDesc_t ;

// Index of pins in timDesc_t.pins
#define	TIM_PIN_CH1		0
#define	TIM_PIN_CH2		1
#define	TIM_PIN_CH3		2
#define	TIM_PIN_CH4		3
#define	TIM_PIN_CHN1	4
#define	TIM_PIN_CHN2	5
#define	TIM_PIN_CHN3	6
#define	TIM_PIN_ETR		7
#define	TIM_PIN_BKIN	8

//------------------------------------------------------------------------------
// Values for timDesc_t.capabilities bits

#define		TIM_CAP_TYPE_ADVT	0x00000001		// Advanced Timer, TIM1, TIM8
#define		TIM_CAP_TYPE_GPT1	0x00000002		// General Purpose Timer, TIM2 to TIM5
#define		TIM_CAP_TYPE_GPT2	0x00000004		// General Purpose Timer, TIM9 to TIM17
#define		TIM_CAP_TYPE_BASIC	0x00000008		// Basic Timer, TIM6, TIM7

#define		TIM_CAP_CHAN_1		0x00000010
#define		TIM_CAP_CHAN_2		0x00000020
#define		TIM_CAP_CHAN_3		0x00000040
#define		TIM_CAP_CHAN_4		0x00000080
#define		TIM_CAP_CHAN_1_2	(TIM_CAP_CHAN_1 | TIM_CAP_CHAN_2)
#define		TIM_CAP_CHAN_1_4	(TIM_CAP_CHAN_1 | TIM_CAP_CHAN_2 | TIM_CAP_CHAN_3 | TIM_CAP_CHAN_4)

#define		TIM_CAP_32_BITS		0x00000100		// Have 32 bits counter
#define		TIM_CAP_UP_DOWN		0x00000200		// Have UP, DOWN and UP/DOWN counter type, else UP only
#define		TIM_CAP_DMA			0x00000400		// Have DMA request generation
#define		TIM_CAP_COMP_OUT	0x00000800		// Have complementary output
#define		TIM_CAP_ETR			0x00001000		// Have External Trigger Input
#define		TIM_CAP_BREAK		0x00002000		// Have Break Input

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Values for flag parameter of timTimeBaseInit()
// This function allows to configure CR1 timer register (except CEN)
// LL values are redefined because the LL driver does not define all necessary values

#define		TIM_INIT_FORCE_ARR				0x80000000			// timeBaseFrequency parameter is the ARR value

#define		TIM_INIT_ARR_PRELOADEN			TIM_CR1_ARPE						// Auto-reload preload enable: TIMx-ARR register is buffered
#define		TIM_INIT_ONEPULSE_MODE			LL_TIM_ONEPULSEMODE_SINGLE			// Counter stops counting at the next update event
#define		TIM_INIT_OVRONLY_UPDATE			LL_TIM_UPDATESOURCE_COUNTER			// Only counter overflow/underflow generates an update request
#define		TIM_INIT_UPDATEDIS				TIM_CR1_UDIS						// Update event disabled

// Select only one of:
#define		TIM_INIT_COUNTUP				0									// Default
#define		TIM_INIT_COUNTDOWN				TIM_CR1_DIR
#define		TIM_INIT_MODE_CENTER_UP			LL_TIM_COUNTERMODE_CENTER_UP		/*!< The counter counts up and down alternatively. Output compare interrupt flags of output channels  are set only when the counter is counting down. */
#define		TIM_INIT_MODE_CENTER_DOWN		LL_TIM_COUNTERMODE_CENTER_DOWN		/*!< The counter counts up and down alternatively. Output compare interrupt flags of output channels  are set only when the counter is counting up */
#define		TIM_INIT_MODE_CENTER_UP_DOWN	LL_TIM_COUNTERMODE_CENTER_UP_DOWN	/*!< The counter counts up and down alternatively. Output compare interrupt flags of output channels  are set only when the counter is counting up or down. */

// Select only one of:
#define		TIM_INIT_CLOCKDIVISION_DIV1		LL_TIM_CLOCKDIVISION_DIV1			/*!< tDTS=tCK_INT, default */
#define		TIM_INIT_CLOCKDIVISION_DIV2		LL_TIM_CLOCKDIVISION_DIV2			/*!< tDTS=2*tCK_INT */
#define		TIM_INIT_CLOCKDIVISION_DIV4		LL_TIM_CLOCKDIVISION_DIV4			/*!< tDTS=4*tCK_INT */

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Values for flag parameter of timSetTriggerOutput, master timer output cause
// flags is one of:
//		LL_TIM_TRGO_RESET
//		LL_TIM_TRGO_ENABLE
//		LL_TIM_TRGO_UPDATE
//		LL_TIM_TRGO_CC1IF
//		LL_TIM_TRGO_OC1REF
//		LL_TIM_TRGO_OC2REF
//		LL_TIM_TRGO_OC3REF
//		LL_TIM_TRGO_OC4REF

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Values for flag parameter of timConfigureIntr()
// All these interrupts are not allowed by all timers
// flags is one of:

#define		TIM_INTR_DISABLE	0x80000000			// To disable all timer interrupts
#define		TIM_INTR_UPDATE		TIM_DIER_UIE		// Interrupt at every timer overflow/underflow
//	TIM_DIER_UIE	/*!< Update interrupt enable */
//	TIM_DIER_CC1IE	/*!< Capture/compare 1 interrupt enable */
//	TIM_DIER_CC2IE	/*!< Capture/compare 2 interrupt enable */
//	TIM_DIER_CC3IE	/*!< Capture/compare 3 interrupt enable */
//	TIM_DIER_CC4IE	/*!< Capture/compare 4 interrupt enable */
//	TIM_DIER_COMIE	/*!< COM interrupt enable */
//	TIM_DIER_TIE	/*!< Trigger interrupt enable */
//	TIM_DIER_BIE	/*!< Break interrupt enable */

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Values for input parameter of timExtClockInit(): external clock source

#define	TIM_ECLK_IN0			LL_TIM_TS_ITR0
#define	TIM_ECLK_IN1			LL_TIM_TS_ITR1
#define	TIM_ECLK_IN2			LL_TIM_TS_ITR2
#define	TIM_ECLK_IN3			LL_TIM_TS_ITR3
#define	TIM_ECLK_IN4			LL_TIM_TS_TI1F_ED
#define	TIM_ECLK_IN5			LL_TIM_TS_TI1FP1
#define	TIM_ECLK_IN6			LL_TIM_TS_TI2FP2
#define	TIM_ECLK_IN7			LL_TIM_TS_ETRF

// Values for flags parameter of timExtClockInit()
#define	TIM_ECLK_EDGE_RISING	1
#define	TIM_ECLK_EDGE_FALLING	2
#if defined (LL_TIM_IC_POLARITY_BOTHEDGE)
	#define	TIM_ECLK_EDGE_BOTH	3		// Not allowed in STM32F1
#endif

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Values for flags parameter of timOutputChannelEnable()

#define TIM_OCMODE_PRELOADEN	TIM_CCMR1_OC1PE						// Preload register on TIMx_CCRx enabled
#define TIM_OCMODE_FASTEN		TIM_CCMR1_OC1FE						// Output compare fast enabled
#define TIM_OCMODE_CLEAREN		TIM_CCMR1_OC1CE						// Output compare clear on ETRF input high enabled
#define TIM_OCMODE_POLLOW		(TIM_CCER_CC1P << 16)				// OCx active low. Default is OCx active high

// Select only one of:
#define TIM_OCMODE_FROZEN		LL_TIM_OCMODE_FROZEN				/*!<The comparison between the output compare register TIMx_CCRy and the counter TIMx_CNT has no effect on the output channel level */
#if (! defined USE_HAL_DRIVER)
// HAL also define these constants...
#define TIM_OCMODE_ACTIVE		LL_TIM_OCMODE_ACTIVE				/*!<OCyREF is forced high on compare match*/
#define TIM_OCMODE_INACTIVE		LL_TIM_OCMODE_INACTIVE				/*!<OCyREF is forced low on compare match*/
#define TIM_OCMODE_TOGGLE		LL_TIM_OCMODE_TOGGLE				/*!<OCyREF toggles on compare match*/
#define TIM_OCMODE_FORCED_INACTIVE	LL_TIM_OCMODE_FORCED_INACTIVE	/*!<OCyREF is forced low*/
#define TIM_OCMODE_FORCED_ACTIVE	LL_TIM_OCMODE_FORCED_ACTIVE		/*!<OCyREF is forced high*/
#define TIM_OCMODE_PWM1			LL_TIM_OCMODE_PWM1					/*!<In upcounting, channel y is active as long as TIMx_CNT<TIMx_CCRy else inactive.  In downcounting, channel y is inactive as long as TIMx_CNT>TIMx_CCRy else active.*/
#define TIM_OCMODE_PWM2			LL_TIM_OCMODE_PWM2					/*!<In upcounting, channel y is inactive as long as TIMx_CNT<TIMx_CCRy else active.  In downcounting, channel y is active as long as TIMx_CNT>TIMx_CCRy else inactive*/
#endif

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Values for flags parameter of timInputChannelEnable()

// Flags can contain one filter value : LL_TIM_IC_FILTER_FDIVx_Ny defined by xxx_ll_tim.h
// Default value is no filter

#define TIM_ICMODE_TISTD	0x0001			// Channel input is mapped to standard TI (TI1 for chan1, TI2 for chan2...)
#define TIM_ICMODE_TISWAP	0x0002			// Channel input is mapped to other  TI   (TI2 for chan1, TI1 for chan2...)
#define TIM_ICMODE_TRC		0x0003			// Channel input is mapped to TRC

#define TIM_ICMODE_DIV2		0x0004			// Input prescaler 2
#define TIM_ICMODE_DIV4		0x0008			// Input prescaler 4
#define TIM_ICMODE_DIV8		0x000C			// Input prescaler 8

#define TIM_ICMODE_RISING	0x0000										// Sensitive to rising edge
#define TIM_ICMODE_FALLING	((TIM_CCER_CC1P) << 24)						// Sensitive to falling edge
#define TIM_ICMODE_BOTH		((TIM_CCER_CC1P | TIM_CCER_CC1NP) << 24)	// Sensitive to rising edge

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Values for flags parameter of timEncoderInit()

#define	TIM_ENC_SWITCH			0x0010u	// Default count on TI1FP1. if TIM_ENC_SWITCH then count on TI2FP2.
#define	TIM_ENC_INVCH1			0x0020u	// Default sensitive to rising edge on CH1. If TIM_ENC_INVCH1 then invert CH1.
#define	TIM_ENC_INVCH2			0x0040u	// Default sensitive to rising edge on CH2. If TIM_ENC_INVCH2 then invert CH2.
#define	TIM_ENC_BOTH			0x0080u	// Count on TI1FP1 and TI2FP2

// Flags can contain one filter value : LL_TIM_IC_FILTER_FDIVx_Ny defined by xxx_ll_tim.h
// Default value is no filter

// Select only one of:
#define TIM_ENC_IC1_DIRECTTI	0x0000u		/*!< IC1 is mapped on TI1, default */
#define TIM_ENC_IC1_INDIRECTTI	0x0100u		/*!< IC1 is mapped on TI2 */

// Select only one of:
#define TIM_ENC_IC2_DIRECTTI	0x0000u		/*!< IC2 is mapped on TI2, default */
#define TIM_ENC_IC2_INDIRECTTI	0x0200u		/*!< IC2 is mapped on TI1 */

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Definition of returned values

typedef enum
{
	TIM_ENONE		= 0,
	TIM_EFAIL		= 1

} timError_t ;

//------------------------------------------------------------------------------
// Type of interrupt callback
typedef void (* fnCallback_t) (const timDesc_t * pTimDesc, uintptr_t arg, uint32_t cause) ;

#ifdef __cplusplus
extern "C" {
#endif


timError_t		timGetDesc 				(uint32_t index, const timDesc_t ** ppTimDesc) ;
void			timGetRegPtr			(const timDesc_t * pTimDesc, TIM_TypeDef ** ppTimPtr) ;
void			timChannelDisable		(const timDesc_t * pTimDesc, uint32_t channel) ;

void  			timTimeBaseInit			(const timDesc_t * pTimDesc, uint32_t prescaler, uint32_t timeBaseFrequency, uint32_t flags) ;
void			timConfigureIntr		(const timDesc_t * pTimDesc, uint32_t priority, fnCallback_t pCallback, uintptr_t arg, uint32_t flags) ;
void			timSetTriggerOutput		(const timDesc_t * pTimDesc, uint32_t flags) ;
timError_t		timExtClockInit			(const timDesc_t * pTimDesc, uint32_t input, uint32_t prescaler, uint32_t flags) ;
void			timEncoderInit			(const timDesc_t * pTimDesc, uint32_t flags) ;

void			timSetCounter			(const timDesc_t * pTimDesc, uint32_t value) ;
uint32_t		timGetCounter			(const timDesc_t * pTimDesc) ;
void			timSetReload	 		(const timDesc_t * pTimDesc, uint32_t reload) ;
uint32_t		timGetReload	 		(const timDesc_t * pTimDesc) ;
void			timStart				(const timDesc_t * pTimDesc) ;
void			timStop					(const timDesc_t * pTimDesc) ;

void			timPwmReset				(const timDesc_t * pTimDesc) ;
void			timOutputChannelEnable	(const timDesc_t * pTimDesc, uint32_t channel, uint32_t flags) ;
void			timOutputChannelSet		(const timDesc_t * pTimDesc, uint32_t channel, uint32_t value) ;
#define			timOutputChannelDisable	timChannelDisable

void			timInputInit			(const timDesc_t * pTimDesc, uint32_t prescaler) ;
void			timInputChannelInit		(const timDesc_t * pTimDesc, uint32_t channel, uint32_t flags) ;
void			timInputChannelEnable	(const timDesc_t * pTimDesc, uint32_t channelMask) ;
#define			timInputChannelDisable	timChannelDisable

#ifdef __cplusplus
}
#endif

//------------------------------------------------------------------------------

#endif // TIMBASIC_H_
