/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	isrpipe.c		Communication channel for Zero Latency Interrupts

	When		Who	What
	15/09/19	ac	Creation
	01/06/20	ac	aaPipe32Create(): add AA_WITH_MALLOC condition

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2021 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"	// For aaCriticalEnter() aaCriticalExit()
#include	"aalist.h"
#include	"aalogmes.h"
#include	"aaprintf.h"

#include	"isrpipe.h"

//--------------------------------------------------------------------------------
//	Pipe for communication between ISR and tasks
//	No protection or blocking: only 1 writer on one side, and one reader on the other side.

typedef struct
{
	uintptr_t		* pBuffer ;		// The ring buffer. Mandatory first item of the struct (used as list pointer in free list)
	uint16_t		count ;			// Count of items in pBuffer
	uint16_t		mask ;
	uint16_t		wIndex ;		// Index of next data to write in pBuffer
	uint16_t		rIndex ;		// Index of next data to read  in pBuffer
	uint16_t		flags ;
	uint8_t			ovfl ;			// 1 when the writer signaled an overflow

} aaPipe32_t ;

// Flags, internal use only
#define		AA_PIPE_DYNAMICBUFFER_	0x4000		// The data buffer was dynamically allocated by pipe32
#define		AA_PIPE_DELETED_		0x8000		// This pipe is not used

// The pipes array
STATIC	aaPipe32_t			aaPipe32 [AA_PIPE32_MAX] ;

// The list of free pipes
STATIC	aaStackNode_t		aaPipe32List ;


//--------------------------------------------------------------------------------
//	Initialize the pipe32 package. Call only once.

void		aaPipe32Init	(void)
{
	uint32_t	ii ;

	// Do it only once
	if (aaPipe32List.pNext == (struct aaStackNode_t *) NULL)
	{
		aaStackInitHead (& aaPipe32List) ;
		for (ii = 0 ; ii < AA_PIPE32_MAX ; ii++)
		{
			aaPipe32 [ii].flags = AA_PIPE_DELETED_ ;
			aaStackPush (& aaPipe32List, (aaStackNode_t *) & aaPipe32 [ii].pBuffer) ;
		}
	}
}

//--------------------------------------------------------------------------------
//	Initialize a pipe32
//	If pBuffer is NULL, the data buffer is dynamically allocated
//	count must be a power of 2
//	Flags is AA_PIPE_ISR_WRITER or AA_PIPE_TASK_WRITER

aaError_t	aaPipe32Create		(aaPipe32Id_t * pId, uintptr_t * pBuffer, uint32_t count, uint32_t flags)
{
	aaPipe32_t	* pPipe ;

	* pId = AA_INVALID_PIPE32 ;

	// Check if the pipe32 package is initialized
	AA_ASSERT (aaPipe32List.pNext != 0) ;

	// Check if count is a power of 2
	if (((count) & (count - 1)) != 0)
	{
		return AA_EARG ;
	}

	#if (AA_WITH_MALLOC != 1)
	{
		if (pBuffer == NULL)
		{
			return AA_EARG ;	// Dynamic memory management not allowed
		}
	}
	#endif

	// Find a pipe descriptor block in the free list
	aaCriticalEnter () ;
	if (aaStackIsEmpty (& aaPipe32List) == 1)
	{
		aaCriticalExit ();
		return AA_EDEPLETED ;
	}

	pPipe = (aaPipe32_t *) aaStackPop (& aaPipe32List) ;
	aaCriticalExit () ;

	pPipe->count  = (uint16_t) count ;
	pPipe->mask   = (uint16_t) (count - 1u) ;
	pPipe->flags  = flags & AA_PIPE_ISR_WRITER ;
	pPipe->rIndex = 0 ;
	pPipe->wIndex = 0 ;
	pPipe->ovfl   = 0 ;

	// Allocate buffer
	pPipe->pBuffer = pBuffer ;
	if (pBuffer == NULL)
	{
		// Dynamically allocate the data buffer
		#if (AA_WITH_MALLOC == 1)
		{
			pPipe->pBuffer = (uintptr_t *) aaMalloc (count * sizeof (uintptr_t)) ;
			if (pPipe->pBuffer == NULL)
			{
				// Memory error, return the pipe to the free list
				aaCriticalEnter () ;
				aaStackPush (& aaPipe32List, (aaStackNode_t *) & pPipe->pBuffer) ;
				aaCriticalExit () ;
				return AA_EMEMORY ;
			}
			pPipe->flags |= AA_PIPE_DYNAMICBUFFER_ ;		// Pipe buffer allocated by the kernel
		}
		#else
		{
			// Dynamic memory management not allowed
			AA_ERRORNOTIFY (AA_ERROR_ISRPIPE) ;
		}
		#endif		// AA_WITH_MALLOC
	}

	* pId = pPipe ;
	return AA_ENONE ;
}

//--------------------------------------------------------------------------------
// Give back a pipe32 to the list of free pipes

aaError_t	aaPipe32Delete		(aaPipe32Id_t id)
{
	aaPipe32_t	* pPipe = (aaPipe32_t *) id ;

	if (pPipe->flags == AA_PIPE_DELETED_)
	{
		return AA_EARG ;	// Already deleted
	}

	#if (AA_WITH_MALLOC == 1)
	{
		if ((pPipe->flags & AA_PIPE_DYNAMICBUFFER_) != 0)
		{
			aaFree (pPipe->pBuffer) ;
		}
	}
	#endif		// AA_WITH_MALLOC

	pPipe->pBuffer = NULL ;
	pPipe->flags   = AA_PIPE_DELETED_ ;

	// Put pipe in free list
	aaCriticalEnter () ;
	aaStackPush (& aaPipe32List, (aaStackNode_t *) & pPipe->pBuffer) ;
	aaCriticalExit () ;

	return AA_ENONE ;
}

//--------------------------------------------------------------------------------
// Initialize a pipe as after creation

aaError_t	aaPipe32Reset		(aaPipe32Id_t id)
{
	aaPipe32_t	* pPipe = (aaPipe32_t *) id ;

	if (pPipe->flags == AA_PIPE_DELETED_)
	{
		return AA_EARG ;	// Deleted
	}

	pPipe->rIndex = 0 ;
	pPipe->wIndex = 0 ;
	pPipe->ovfl   = 0 ;

	return AA_ENONE ;
}

//--------------------------------------------------------------------------------
//	Get the number of data that can be written to the pipe

aaError_t	aaPipe32ToGive		(aaPipe32Id_t id, uint32_t * pCount)
{
	aaPipe32_t	* pPipe = (aaPipe32_t *) id ;

	* pCount = (pPipe->count - (((uint32_t) pPipe->wIndex - pPipe->rIndex)  & pPipe->mask)) - 1u ;
	return AA_ENONE ;
}

//--------------------------------------------------------------------------------
//	Get the number of data that can be read from the pipe

aaError_t	aaPipe32ToTake		(aaPipe32Id_t id, uint32_t * pCount)
{
	aaPipe32_t	* pPipe = (aaPipe32_t *) id ;

	* pCount = ((uint32_t) pPipe->wIndex - pPipe->rIndex) & pPipe->mask ;
	return AA_ENONE ;
}

//--------------------------------------------------------------------------------
//	Put data to the pipe

aaError_t	aaPipe32Give	(aaPipe32Id_t id, uintptr_t data)
{
	aaPipe32_t	* pPipe = (aaPipe32_t *) id ;

	// Check if enough room
	if ((pPipe->count - (pPipe->wIndex - pPipe->rIndex)) != 1u)
	{
		// Write data
		pPipe->pBuffer [pPipe->wIndex] = data;
		pPipe->wIndex = (uint16_t) ((pPipe->wIndex + 1u) & pPipe->mask) ;
		return AA_ENONE ;
	}

	pPipe->ovfl = 1 ;
	return AA_EFAIL ;
}

//--------------------------------------------------------------------------------
//	Get data from the pipe

aaError_t	aaPipe32Take		(aaPipe32Id_t id, uintptr_t * pData)
{
	aaPipe32_t	* pPipe = (aaPipe32_t *) id ;

	if (pPipe->rIndex != pPipe->wIndex)
	{
		* pData = pPipe->pBuffer [pPipe->rIndex] ;
		pPipe->rIndex = (uint16_t) ((pPipe->rIndex + 1u) & pPipe->mask) ;
		return AA_ENONE ;
	}
	return AA_EFAIL ;
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	Fast block allocator, suitable for ZLI use.
//--------------------------------------------------------------------------------

#define		POOL_MAGIC		0xB001AC1C

//--------------------------------------------------------------------------------
//	Count must be a power of 2
//	There is no dynamic memory allocation, so:
//	All data (pPool, pBuffer, pIndex) must be allocated before the call of this function

aaError_t	aaBlockPoolCreate		(aaBlockPool_t * pPool, uint8_t * pBuffer, uint8_t * pIndex, uint32_t count, uint32_t size)
{
	// Check if count is a power of 2, and <= 256
	if ((((count) & (count - 1u)) != 0)  ||  count > 256u)
	{
		return AA_EARG ;
	}

	pPool->pBuffer		= pBuffer ;
	pPool->pIndex		= pIndex ;
	pPool->count		= (uint16_t) count ;
	pPool->mask			= (uint16_t) (count - 1u) ;
	pPool->size			= (uint16_t) size ;
	pPool->underflow	= 0u ;
	pPool->magic		= POOL_MAGIC ;		// The pool is initialized and valid

	aaBlockPoolReset (pPool) ;

	return AA_ENONE ;
}


//--------------------------------------------------------------------------------

aaError_t	aaBlockPoolDelete		(aaBlockPool_t * pPool)
{
	if (pPool->magic == POOL_MAGIC)
	{
		pPool->magic		= 0 ;
		pPool->pBuffer		= NULL ;
		pPool->pIndex		= NULL ;
		return AA_ENONE ;
	}
	return AA_EFAIL ;
}

//--------------------------------------------------------------------------------

aaError_t	aaBlockPoolReset		(aaBlockPool_t * pPool)
{
	uint32_t	ii ;

	if (pPool->magic == POOL_MAGIC)
	{
		pPool->giveIndex	= (uint16_t) (pPool->count - 1u) ;
		pPool->takeIndex	= 0u ;
		pPool->underflow	= 0u ;

		// Fill pIndex array
		for (ii = 0 ; ii < pPool->count ; ii++)
		{
			pPool->pIndex [ii] = (uint8_t) ii ;
		}
		return AA_ENONE ;
	}
	return AA_EFAIL ;
}

//--------------------------------------------------------------------------------
// Useless for common use but used for testing

aaError_t	aaBlockPoolToGive		(aaBlockPool_t * pPool, uint32_t * pCount)
{
	if (pPool->magic == POOL_MAGIC)
	{
		* pCount = pPool->count - (((uint32_t) pPool->giveIndex - pPool->takeIndex) & pPool->mask) - 1u ;
		return AA_ENONE ;
	}
	return AA_EFAIL ;
}

//--------------------------------------------------------------------------------
//	Returns the count of available blocks in the pool

aaError_t	aaBlockPoolToTake		(aaBlockPool_t * pPool, uint32_t * pCount)
{
	if (pPool->magic == POOL_MAGIC)
	{
		* pCount = ((uint32_t) pPool->giveIndex - pPool->takeIndex) & pPool->mask ;
		return AA_ENONE ;
	}
	return AA_EFAIL ;
}

//--------------------------------------------------------------------------------
//	Extract a block from the pool

aaError_t	aaBlockPoolTake			(aaBlockPool_t * pPool, void ** ppBlock)
{
	uint32_t	ix ;

	if (pPool->magic == POOL_MAGIC)
	{
		if (pPool->takeIndex != pPool->giveIndex)
		{
			ix = pPool->pIndex [pPool->takeIndex] ;
			* ppBlock = & pPool->pBuffer [ix * pPool->size] ;
			pPool->takeIndex = (uint16_t) (pPool->takeIndex + 1u) & pPool->mask ;
			return AA_ENONE ;
		}
		pPool->underflow = 1 ;
	}
	return AA_EFAIL ;
}

//--------------------------------------------------------------------------------
//	Returns a block to the pool
//	__NEVER_INLINE to avoid compiler warning with some optimization level (as -Ofast)

__NEVER_INLINE aaError_t	aaBlockPoolGive			(aaBlockPool_t * pPool, void * pBlock)
{
	uint32_t	ix ;

	if (pPool->magic == POOL_MAGIC)
	{
		// Check if enough room: normally yes
		if ((pPool->count - (pPool->giveIndex - pPool->takeIndex)) != 1)
		{
			// Put the bloc in free buffer
			ix = (uint32_t) ((((uint8_t *) pBlock) - pPool->pBuffer) / pPool->size) ;
			pPool->pIndex [pPool->giveIndex] = (uint8_t) ix ;
			pPool->giveIndex = (uint16_t) ((pPool->giveIndex + 1u) & pPool->mask) ;
			return AA_ENONE ;
		}
		AA_ASSERT (0) ;	// It is abnormal not to be able to give
	}
	return AA_EFAIL ;
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	Pipe32 test

#define		PIPE_COUNT	8

static void	pipeCheckCount (aaPipe32Id_t pipeId, uint32_t w, uint32_t r)
{
	uint32_t		count ;

	aaPipe32ToGive (pipeId, & count);
	aaPrintf ("To write: %u-%u ", count, w) ;
	AA_ASSERT (count == w) ;
	aaPipe32ToTake (pipeId, & count);
	aaPrintf ("To Read: %u-%u \n", count, r) ;
	AA_ASSERT (count == r) ;
}

void	testPipe32 (void)
{
	aaPipe32Id_t	pipeId ;
	aaError_t		res ;
	uintptr_t		data = 0 ;		// To avoid: 'data' may be used uninitialized

	(void) res ;

	aaPrintf ("\nTest pipe32\n") ;

	aaPipe32Init () ;

	res = aaPipe32Create (& pipeId, NULL, PIPE_COUNT, AA_PIPE_ISR_WRITER) ;
	AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Reset (pipeId) ;
	AA_ASSERT (res == AA_ENONE) ;

	// Check count
	pipeCheckCount (pipeId, PIPE_COUNT-1, 0) ;

	res = aaPipe32Give (pipeId, 1) ;			AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Give (pipeId, 2) ;			AA_ASSERT (res == AA_ENONE) ;
	pipeCheckCount (pipeId, PIPE_COUNT-3, 2) ;

	res = aaPipe32Give (pipeId, 3) ;			AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Give (pipeId, 4) ;			AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Give (pipeId, 5) ;			AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Give (pipeId, 6) ;			AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Give (pipeId, 7) ;			AA_ASSERT (res == AA_ENONE) ;
	pipeCheckCount (pipeId, 0, PIPE_COUNT-1) ;

	res = aaPipe32Take (pipeId, & data) ;	aaPrintf ("Data %u-1\n", data) ;	AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Take (pipeId, & data) ;	aaPrintf ("Data %u-2\n", data) ;	AA_ASSERT (res == AA_ENONE) ;
	pipeCheckCount (pipeId, 2, PIPE_COUNT-3) ;

	res = aaPipe32Give (pipeId, 8) ;			AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Give (pipeId, 9) ;			AA_ASSERT (res == AA_ENONE) ;
	pipeCheckCount (pipeId, 0, PIPE_COUNT-1) ;

	res = aaPipe32Take (pipeId, & data) ;	aaPrintf ("Data %u-3\n", data) ;	AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Take (pipeId, & data) ;	aaPrintf ("Data %u-4\n", data) ;	AA_ASSERT (res == AA_ENONE) ;
	pipeCheckCount (pipeId, 2, PIPE_COUNT-3) ;

	res = aaPipe32Take (pipeId, & data) ;	aaPrintf ("Data %u-5\n", data) ;	AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Take (pipeId, & data) ;	aaPrintf ("Data %u-6\n", data) ;	AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Take (pipeId, & data) ;	aaPrintf ("Data %u-7\n", data) ;	AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Take (pipeId, & data) ;	aaPrintf ("Data %u-8\n", data) ;	AA_ASSERT (res == AA_ENONE) ;
	res = aaPipe32Take (pipeId, & data) ;	aaPrintf ("Data %u-9\n", data) ;	AA_ASSERT (res == AA_ENONE) ;
	pipeCheckCount (pipeId, PIPE_COUNT-1, 0) ;

	res = aaPipe32Delete (pipeId) ;				AA_ASSERT (res == AA_ENONE) ;

	aaPrintf ("\nTest pipe32 ended\n\n") ;
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	BlocPool test


// testBlockPool uses AA_ASSERT, so it must be defined
#if (AA_WITH_DEBUG != 1)

void	testBlockPool (void)
{
	aaPrintf ("\nCannot run testBlockPool if AA_WITH_DEBUG is not 1\n") ;
}
#else

// - - - - - - - - - - - - - Define the block pool
#define		POOL_COUNT	8
#define		POOL_SIZE	16

STATIC	aaBlockPool_t	pool ;

STATIC	uint8_t		poolBuffer [POOL_COUNT * POOL_SIZE] ;
STATIC	uint8_t		poolIndex  [POOL_COUNT] ;

// - - - - - - - - - - - - -
//	__NEVER_INLINE to avoid compiler warning with some optimization level (as -Ofast)

__NEVER_INLINE static void	poolCheckCount (aaBlockPool_t * pPool, uint32_t give, uint32_t take)
{
	uint32_t		count = 0 ;	// Initialized to avois compiler warning (may be used uninitialized)

	(void) give ;		// To avoid compiler warning when AA_ASSERT is removed
	(void) take ;

	aaBlockPoolToGive (pPool, & count) ;
	aaPrintf ("To give: %u-%u ", count, give) ;
	AA_ASSERT (count == give) ;

	aaBlockPoolToTake (pPool, & count);
	aaPrintf ("To take: %u-%u \n", count, take) ;
	AA_ASSERT (count == take) ;
}

void	testBlockPool (void)
{
	uint32_t		ii ;
	aaError_t		res ;
	uint8_t			* pBlock = NULL ; 	// Initialized to avois compiler warning (may be used uninitialized)

	(void) res ;

	aaPrintf ("\nTest BlockPool, count = %u\n", POOL_COUNT) ;

	res = aaBlockPoolCreate (& pool, poolBuffer, poolIndex, POOL_COUNT, POOL_SIZE) ;
	AA_ASSERT (res == AA_ENONE) ;

	// Initialize each block content with its index
	for (ii = 0 ; ii < POOL_COUNT ; ii++)
	{
		poolBuffer [ii * POOL_SIZE] = (uint8_t) ii ;
	}

	poolCheckCount (& pool, 0, POOL_COUNT-1) ;

	aaPrintf ("Take 2\n") ;
	res = aaBlockPoolTake (& pool, (void **) & pBlock) ;		AA_ASSERT (res == AA_ENONE) ; AA_ASSERT (* pBlock == 0) ;
	res = aaBlockPoolTake (& pool, (void **) & pBlock) ;		AA_ASSERT (res == AA_ENONE) ; AA_ASSERT (* pBlock == 1) ;
	poolCheckCount (& pool, 2, POOL_COUNT-3) ;

	aaPrintf ("Give 2\n") ;
	res = aaBlockPoolGive (& pool, pBlock) ;					AA_ASSERT (res == AA_ENONE) ;
	res = aaBlockPoolGive (& pool, pBlock) ;					AA_ASSERT (res == AA_ENONE) ;
	poolCheckCount (& pool, 0, POOL_COUNT-1) ;

	aaPrintf ("Take 7\n") ;
	res = aaBlockPoolTake (& pool, (void **) & pBlock) ;		AA_ASSERT (res == AA_ENONE) ; AA_ASSERT (* pBlock == 2) ;
	res = aaBlockPoolTake (& pool, (void **) & pBlock) ;		AA_ASSERT (res == AA_ENONE) ; AA_ASSERT (* pBlock == 3) ;
	res = aaBlockPoolTake (& pool, (void **) & pBlock) ;		AA_ASSERT (res == AA_ENONE) ; AA_ASSERT (* pBlock == 4) ;
	res = aaBlockPoolTake (& pool, (void **) & pBlock) ;		AA_ASSERT (res == AA_ENONE) ; AA_ASSERT (* pBlock == 5) ;
	res = aaBlockPoolTake (& pool, (void **) & pBlock) ;		AA_ASSERT (res == AA_ENONE) ; AA_ASSERT (* pBlock == 6) ;
	res = aaBlockPoolTake (& pool, (void **) & pBlock) ;		AA_ASSERT (res == AA_ENONE) ; AA_ASSERT (* pBlock == 1) ;
	res = aaBlockPoolTake (& pool, (void **) & pBlock) ;		AA_ASSERT (res == AA_ENONE) ; AA_ASSERT (* pBlock == 1) ;
	poolCheckCount (& pool, POOL_COUNT-1, 0) ;

	res = aaBlockPoolTake (& pool, (void **) & pBlock) ;		AA_ASSERT (res == AA_EFAIL) ;
	aaPrintf ("Take 1, EFAIL expected : OK\n") ;

	aaPrintf ("Give 7\n") ;
	res = aaBlockPoolGive (& pool, pBlock) ;					AA_ASSERT (res == AA_ENONE) ;
	res = aaBlockPoolGive (& pool, pBlock) ;					AA_ASSERT (res == AA_ENONE) ;
	res = aaBlockPoolGive (& pool, pBlock) ;					AA_ASSERT (res == AA_ENONE) ;
	res = aaBlockPoolGive (& pool, pBlock) ;					AA_ASSERT (res == AA_ENONE) ;
	res = aaBlockPoolGive (& pool, pBlock) ;					AA_ASSERT (res == AA_ENONE) ;
	res = aaBlockPoolGive (& pool, pBlock) ;					AA_ASSERT (res == AA_ENONE) ;
	res = aaBlockPoolGive (& pool, pBlock) ;					AA_ASSERT (res == AA_ENONE) ;
	poolCheckCount (& pool, 0, POOL_COUNT-1) ;


	res = aaBlockPoolReset (& pool) ;
	AA_ASSERT (res == AA_ENONE) ;

	res = aaBlockPoolDelete (& pool) ;
	AA_ASSERT (res == AA_ENONE) ;

	res = aaBlockPoolReset (& pool) ;
	AA_ASSERT (res == AA_EFAIL) ;

	aaPrintf ("\nTest BlockPool ended\n\n") ;
}

#endif

//--------------------------------------------------------------------------------
